@@ -21,6 +21,9 @@
 #include "advection.h"
 #include "source.h"
 
+
+#include "solid.h"
+
 #if 1
 static gdouble transverse_term (FttCell * cell,
 				gdouble size,
@@ -74,10 +77,10 @@ void gfs_cell_advected_face_values (FttCell * cell,
       MIN ((1. - unorm)/2., 0.5)*g;
     gdouble vr = GFS_VARIABLE (cell, par->v->i) + 
       MAX ((- 1. - unorm)/2., -0.5)*g;
-    gdouble src = par->dt*gfs_variable_source (par->v, cell)/2.;
+    gdouble src = par->dt*gfs_variable_mac_source (par->v, cell)/2.;
     gdouble dv;
 
-#ifdef FTT_2D
+#if FTT_2D
     dv = transverse_term (cell, size, FTT_ORTHOGONAL_COMPONENT (c), par);
 #else  /* FTT_3D */
     static FttComponent orthogonal[FTT_DIMENSION][2] = {
@@ -122,7 +125,7 @@ void gfs_cell_advected_face_values (FttCell * cell,
     gdouble vr = GFS_VARIABLE (cell, par->v) + MAX ((-1. - u)/2., -0.5)*g[c];
     gdouble dv;
     FttComponent cp;
-#ifdef FTT_2D
+#if FTT_2D
 
     cp = FTT_ORTHOGONAL_COMPONENT (c);
     dv = par->dt*GFS_VARIABLE (cell, GFS_VELOCITY (cp))*g[cp]/(2.*size);
@@ -143,7 +146,37 @@ void gfs_cell_advected_face_values (FttCell * cell,
 }
 #endif
 
-#ifdef FTT_2D
+/**
+ * gfs_cell_non_advected_face_values:
+ * @cell: a #FttCell.
+ * @par: the (non)advection parameters.
+ *
+ * Fills the face variable (@v field of #GfsFaceStateVector) of all the
+ * faces of @cell with the non-advected value of variable @par->v at time
+ * t + dt/2.
+ */
+void gfs_cell_non_advected_face_values (FttCell * cell,
+					const GfsAdvectionParams * par)
+{
+  FttComponent c;
+  GfsStateVector * s;
+
+  g_return_if_fail (cell != NULL);
+  g_return_if_fail (par != NULL);
+
+  s = GFS_STATE (cell);
+  for (c = 0; c < FTT_DIMENSION; c++) {
+    gdouble g = (* par->gradient) (cell, c, par->v->i);
+    gdouble vl = GFS_VARIABLE (cell, par->v->i) + g/2.;
+    gdouble vr = GFS_VARIABLE (cell, par->v->i) - g/2.;
+    gdouble src = par->dt*gfs_variable_mac_source (par->v, cell)/2.;
+
+    s->f[2*c].v     = vl + src;
+    s->f[2*c + 1].v = vr + src;
+  }
+}
+
+#if (FTT_2D || FTT_2D3)
 
 static gdouble interpolate_1D1 (const FttCell * cell,
 				FttDirection dright,
@@ -155,31 +188,36 @@ static gdouble interpolate_1D1 (const FttCell * cell,
   GfsStateVector * s;
 
   g_return_val_if_fail (cell != NULL, 0.);
-  g_return_val_if_fail (!GFS_IS_MIXED (cell), 0.);
 
   dleft = FTT_OPPOSITE_DIRECTION (dright);
   n = ftt_cell_neighbor (cell, dup);
   s = GFS_STATE (cell);
   if (n && !GFS_CELL_IS_BOUNDARY (n)) {
-    FttCell * c;
+    gdouble s1 = s->solid ? s->solid->s[dleft] : 1., s2;
+    gdouble v1 = s->f[dleft].v, v2;
 
+    g_assert (s1 > 0.);
     /* check for corner refinement violation (topology.fig) */
     g_assert (ftt_cell_level (n) == ftt_cell_level (cell));
 
-    if (FTT_CELL_IS_LEAF (n))
-      c = n;
+    if (FTT_CELL_IS_LEAF (n)) {
+      v2 = GFS_STATE (n)->f[dleft].v;
+      s2 = GFS_IS_MIXED (n) ? GFS_STATE (n)->solid->s[dleft] : 1.;
+    }
     else {
       FttDirection d[FTT_DIMENSION];
 
       d[0] = dleft;
       d[1] = FTT_OPPOSITE_DIRECTION (dup);
-      c = ftt_cell_child_corner (n, d);
-      /* check for mixed cell refinement violation */
-      g_assert (c);
-
-      x *= 4./3.;
+      n = ftt_cell_child_corner (n, d);
+      if (n) {
+	v2 = GFS_STATE (n)->f[dleft].v;
+	s2 = GFS_IS_MIXED (n) ? GFS_STATE (n)->solid->s[dleft]/2. : 0.5;
+      }
+      else
+	s2 = v2 = 0.;
     }
-    return (1. - x)*s->f[dleft].v + x*GFS_STATE (c)->f[dleft].v;
+    return s2 > 0. ? (v2*(s1 - 1. + 2.*x) + v1*(s2 + 1. - 2.*x))/(s1 + s2) : v1;
   }
   return s->f[dleft].v;
 }
@@ -256,8 +294,7 @@ static gdouble interpolate_2D1 (const FttCell * cell,
 /**
  * gfs_face_upwinded_value:
  * @face: a #FttCellFace.
- * @centered_upwinding: if %TRUE uses centered upwinding else uses
- * face upwinding.
+ * @upwinding: type of upwinding.
  *
  * This function assumes that the face variable has been previously
  * defined using gfs_cell_advected_face_values().
@@ -265,16 +302,23 @@ static gdouble interpolate_2D1 (const FttCell * cell,
  * Returns: the upwinded value of the face variable.  
  */
 gdouble gfs_face_upwinded_value (const FttCellFace * face,
-				 gboolean centered_upwinding)
+				 GfsUpwinding upwinding)
 {
-  gdouble un;
+  gdouble un = 0.;
 
   g_return_val_if_fail (face != NULL, 0.);
 
-  if (centered_upwinding)
-    un = gfs_face_interpolated_value (face, GFS_VELOCITY_INDEX (face->d/2));
-  else /* face upwinding */
-    un = GFS_FACE_NORMAL_VELOCITY (face);
+  if (GFS_FACE_FRACTION (face) == 0.)
+    return 0.;
+
+  switch (upwinding) {
+  case GFS_CENTERED_UPWINDING:
+    un = gfs_face_interpolated_value (face, GFS_VELOCITY_INDEX (face->d/2)); break;
+  case GFS_FACE_UPWINDING:
+    un = GFS_FACE_NORMAL_VELOCITY (face); break;
+  case GFS_NO_UPWINDING:
+    break;
+  }
   if (!FTT_FACE_DIRECT (face))
     un = - un;
 
@@ -290,9 +334,9 @@ gdouble gfs_face_upwinded_value (const FttCellFace * face,
       return GFS_STATE (face->cell)->f[face->d].v;
     else {
       gdouble vcoarse;
-#ifdef FTT_2D
+#if (FTT_2D || FTT_2D3)
       gint dp;
-      static gint perpendicular[FTT_NEIGHBORS][FTT_CELLS] = 
+      static gint perpendicular[FTT_NEIGHBORS_2D][FTT_CELLS] = 
       {{-1,  2, -1,  3},
        { 2, -1,  3, -1},
        { 1,  0, -1, -1},
@@ -308,8 +352,12 @@ gdouble gfs_face_upwinded_value (const FttCellFace * face,
        {{-1,-1},{-1,-1},{-1,-1},{-1,-1},{1,2},{0,2},{1,3},{0,3}}};
 #endif /* FTT_3D */
 
+#if FTT_2D3
+      g_assert (face->d < FTT_NEIGHBORS_2D);
+#endif
+
       dp = perpendicular[face->d][FTT_CELL_ID (face->cell)];
-#ifdef FTT_2D
+#if (FTT_2D || FTT_2D3)
       g_assert (dp >= 0);
       vcoarse = interpolate_1D1 (face->neighbor, face->d, dp, 1./4.);
 #else  /* FTT_3D */
@@ -400,9 +448,9 @@ void gfs_face_velocity_advection_flux (const FttCellFace * face,
     flux *= GFS_FACE_NORMAL_VELOCITY (face);
   else /* tangential component */
 #else
-  flux *= gfs_face_upwinded_value (face, par->use_centered_upwinding)
-    /* pressure correction */
-    - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c));
+    flux *= gfs_face_upwinded_value (face, par->upwinding)
+      /* pressure correction */
+      - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c))*par->dt/2.;
 #endif
   if (!FTT_FACE_DIRECT (face))
     flux = - flux;
@@ -452,13 +500,13 @@ void gfs_face_velocity_convective_flux (const FttCellFace * face,
   if (c == face->d/2) /* normal component */
     u = GFS_FACE_NORMAL_VELOCITY (face);
   else /* tangential component */
-    u = gfs_face_upwinded_value (face, par->use_centered_upwinding)
+    u = gfs_face_upwinded_value (face, par->upwinding)
       /* pressure correction */
-      - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c));
+      - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c))*par->dt/2.;
 #else
-  u = gfs_face_upwinded_value (face, par->use_centered_upwinding)
+  u = gfs_face_upwinded_value (face, par->upwinding)
     /* pressure correction */
-    - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c));
+    - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c))*par->dt/2.;
 #endif
   u *= par->dt/(2.*ftt_cell_size (face->cell));
   if (!FTT_FACE_DIRECT (face))
@@ -487,6 +535,7 @@ void gfs_face_velocity_convective_flux (const FttCellFace * face,
 /**
  * gfs_face_advected_normal_velocity:
  * @face: a #FttCellFace.
+ * @upwinding: the type of upwinding.
  *
  * Fills the normal component of the velocity at @face with the value
  * advected (to time t + dt/2) from the centered velocities.
@@ -495,21 +544,26 @@ void gfs_face_velocity_convective_flux (const FttCellFace * face,
  * defined for the correct component of the velocity using
  * gfs_cell_advected_face_values().  
  */
-void gfs_face_advected_normal_velocity (const FttCellFace * face)
+void gfs_face_advected_normal_velocity (const FttCellFace * face,
+					GfsUpwinding * upwinding)
 {
   gdouble u;
 
   g_return_if_fail (face != NULL);
 
-  GFS_FACE_NORMAL_VELOCITY_LEFT (face) = u = 
-    gfs_face_upwinded_value (face, TRUE);
+  if (GFS_FACE_FRACTION (face) == 0.)
+    return;
+
+  GFS_FACE_NORMAL_VELOCITY_LEFT (face) = u = gfs_face_upwinded_value (face, *upwinding);
 
   switch (ftt_face_type (face)) {
   case FTT_FINE_FINE:
     GFS_FACE_NORMAL_VELOCITY_RIGHT (face) = u;
     break;
   case FTT_FINE_COARSE:
-    GFS_FACE_NORMAL_VELOCITY_RIGHT (face) += u/(FTT_CELLS/2);
+    GFS_FACE_NORMAL_VELOCITY_RIGHT (face) += 
+      u*GFS_FACE_FRACTION_LEFT (face)/(GFS_FACE_FRACTION_RIGHT (face)*
+				       FTT_CELLS_DIRECTION (face->d));
     break;
   default:
     g_assert_not_reached ();
@@ -529,6 +583,9 @@ void gfs_face_interpolated_normal_velocity (const FttCellFace * face)
 
   g_return_if_fail (face != NULL);
 
+  if (GFS_FACE_FRACTION (face) == 0.)
+    return;
+
   GFS_FACE_NORMAL_VELOCITY_LEFT (face) = u = 
     gfs_face_interpolated_value (face, GFS_VELOCITY_INDEX (face->d/2));
 
@@ -537,7 +594,9 @@ void gfs_face_interpolated_normal_velocity (const FttCellFace * face)
     GFS_FACE_NORMAL_VELOCITY_RIGHT (face) = u;
     break;
   case FTT_FINE_COARSE:
-    GFS_FACE_NORMAL_VELOCITY_RIGHT (face) += u/(FTT_CELLS/2);
+    GFS_FACE_NORMAL_VELOCITY_RIGHT (face) += 
+      u*GFS_FACE_FRACTION_LEFT (face)/(GFS_FACE_FRACTION_RIGHT (face)*
+				       FTT_CELLS_DIRECTION (face->d));
     break;
   default:
     g_assert_not_reached ();
@@ -571,9 +630,8 @@ static void set_merged (FttCell * cell)
 
     ftt_cell_neighbors (cell, &neighbor);
     for (i = 0; i < FTT_NEIGHBORS && abest < 1.; i++)
-      if (neighbor.c[i] && !GFS_CELL_IS_BOUNDARY (neighbor.c[i])) {
-	gdouble a = GFS_IS_MIXED (neighbor.c[i]) ? 
-	  GFS_STATE (neighbor.c[i])->solid->a : 1.;
+      if (neighbor.c[i] && !GFS_CELL_IS_BOUNDARY (neighbor.c[i]) && solid->s[i] > 0.) {
+	gdouble a = GFS_IS_MIXED (neighbor.c[i]) ? GFS_STATE (neighbor.c[i])->solid->a : 1.;
 	
 	if (a > abest) {
 	  abest = a;
@@ -598,8 +656,8 @@ void gfs_set_merged (GfsDomain * domain)
 {
   g_return_if_fail (domain != NULL);
 
-  gfs_domain_traverse_mixed (domain, 
-			    (FttCellTraverseFunc) set_merged, NULL);
+  gfs_domain_traverse_mixed (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS,
+			     (FttCellTraverseFunc) set_merged, NULL);
 }
 
 static void add_merged (GSList ** merged, FttCell * cell)
@@ -621,9 +679,10 @@ static void add_merged (GSList ** merged, FttCell * cell)
 	if (!FTT_CELL_IS_LEAF (neighbor.c[i])) {
 	  FttCellChildren child;
 	  guint j;
-	  
-	  ftt_cell_children_direction (neighbor.c[i], 
-				       FTT_OPPOSITE_DIRECTION (i), &child);
+#if FTT_2D3
+	  g_assert_not_implemented ();
+#endif
+	  ftt_cell_children_direction (neighbor.c[i], FTT_OPPOSITE_DIRECTION (i), &child);
 	  for (j = 0; j < FTT_CELLS/2; j++)
 	    if (GFS_IS_MIXED (child.c[j]) &&
 		GFS_STATE (child.c[j])->solid->merged == cell)
@@ -684,7 +743,7 @@ void gfs_domain_traverse_merged (GfsDomain * domain,
 
   datum[0] = func;
   datum[1] = data;
-  gfs_domain_traverse_mixed (domain, 
+  gfs_domain_traverse_mixed (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS,
 			    (FttCellTraverseFunc) traverse_merged, datum);
   gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
 			   (FttCellTraverseFunc) traverse_non_merged, datum);
@@ -717,21 +776,22 @@ void gfs_advection_update (GSList * merged, const GfsAdvectionParams * par)
 #else /* D. Calhoun approach */
       FttDirection d;
       gdouble mins = G_MAXDOUBLE;
+      GfsSolidVector * solid = GFS_STATE (cell)->solid;
 
       for (d = 0; d < FTT_NEIGHBORS; d++)
-	if (cell->s.solid->s[d] > 0. && 1./cell->s.solid->s[d] < mins)
-	  mins = 1./cell->s.solid->s[d];
+	if (solid->s[d] > 0. && 1./solid->s[d] < mins)
+	  mins = 1./solid->s[d];
 #if 0
 fprintf (stderr, "%g %g %g\n",
-	 cell->s.solid->a, mins, 
-	 GFS_VARIABLE (cell, par->fv)/(mins*cell->s.solid->a));
+	 solid->a, mins, 
+	 GFS_VARIABLE (cell, par->fv)/(mins*solid->a));
 #endif
-      if (mins*cell->s.solid->a > 0.01)
-	GFS_VARIABLE (cell, par->v) += 
-	  GFS_VARIABLE (cell, par->fv)/(mins*cell->s.solid->a);
+      if (mins*solid->a > 0.01)
+	GFS_VARIABLE (cell, par->v->i) += 
+	  GFS_VARIABLE (cell, par->fv->i)/(mins*solid->a);
       else
-	GFS_VARIABLE (cell, par->v) += 100.*GFS_VARIABLE (cell, par->fv);
-      g_assert (GFS_VARIABLE (cell, par->v) < 10.);
+	GFS_VARIABLE (cell, par->v->i) += 100.*GFS_VARIABLE (cell, par->fv->i);
+      g_assert (GFS_VARIABLE (cell, par->v->i) < 10.);
 #endif
     }
     else
@@ -744,22 +804,19 @@ fprintf (stderr, "%g %g %g\n",
 
     while (i) {
       FttCell * cell = i->data;
-      gdouble vol = ftt_cell_volume (cell);
-      gdouble a = GFS_IS_MIXED (cell) ? cell->s.solid->a : 1.;
+      gdouble a = GFS_IS_MIXED (cell) ? GFS_STATE (cell)->solid->a : 1.;
       
-      total_vol += vol*a;
-      w += vol*GFS_VARIABLE (cell, par->fv);
+      total_vol += a;
+      w += GFS_VARIABLE (cell, par->fv->i);
       i = i->next;
     }
-    w /= total_vol*total_vol;
+    w /= total_vol;
 
     i = merged;
     while (i) {
       FttCell * cell = i->data;
-      gdouble vol = ftt_cell_volume (cell);
-      gdouble a = GFS_IS_MIXED (cell) ? cell->s.solid->a : 1.;
 
-      GFS_VARIABLE (cell, par->v) += a*vol*w;
+      GFS_VARIABLE (cell, par->v->i) += w;
       i = i->next;
     }
 #else /* average value */
@@ -798,8 +855,7 @@ void gfs_advection_params_write (GfsAdvectionParams * par, FILE * fp)
            "{\n"
 	   "  cfl      = %g\n"
 	   "  gradient = %s\n"
-	   "  flux     = %s\n"
-	   "}",
+	   "  flux     = %s\n",
 	   par->cfl,
 	   par->gradient == gfs_center_gradient ? 
 	   "gfs_center_gradient" :
@@ -810,6 +866,12 @@ void gfs_advection_params_write (GfsAdvectionParams * par, FILE * fp)
 	   "gfs_face_velocity_advection_flux" :
 	   par->flux == gfs_face_velocity_convective_flux ?
 	   "gfs_face_velocity_convective_flux" : "NULL");
+  switch (par->scheme) {
+  case GFS_GODUNOV: fputs ("  scheme   = godunov\n", fp); break;
+  case GFS_VOF:     fputs ("  scheme   = vof\n", fp); break;
+  case GFS_NONE:    fputs ("  scheme   = none\n", fp); break;
+  }
+  fputc ('}', fp);
 }
 
 void gfs_advection_params_init (GfsAdvectionParams * par)
@@ -817,9 +879,11 @@ void gfs_advection_params_init (GfsAdvectionParams * par)
   g_return_if_fail (par != NULL);
 
   par->cfl = 0.8;
+  par->dt = 1.;
   par->gradient = gfs_center_gradient;
-  par->use_centered_upwinding = FALSE;
+  par->upwinding = GFS_FACE_UPWINDING;
   par->use_centered_velocity = TRUE;
+  par->scheme = GFS_GODUNOV;
   par->rho = 1.;
   par->c = NULL;
 }
@@ -830,9 +894,10 @@ void gfs_advection_params_read (GfsAdvectionParams * par, GtsFile * fp)
     {GTS_DOUBLE, "cfl",      TRUE},
     {GTS_STRING, "gradient", TRUE},
     {GTS_STRING, "flux",     TRUE},
+    {GTS_STRING, "scheme",   TRUE},
     {GTS_NONE}
   };
-  gchar * gradient = NULL, * flux = NULL;
+  gchar * gradient = NULL, * flux = NULL, * scheme = NULL;
 
   g_return_if_fail (par != NULL);
   g_return_if_fail (fp != NULL);
@@ -840,6 +905,7 @@ void gfs_advection_params_read (GfsAdvectionParams * par, GtsFile * fp)
   var[0].data = &par->cfl;
   var[1].data = &gradient;
   var[2].data = &flux;
+  var[3].data = &scheme;
 
   gfs_advection_params_init (par);
   gts_file_assign_variables (fp, var);
@@ -871,4 +937,22 @@ void gfs_advection_params_read (GfsAdvectionParams * par, GtsFile * fp)
 			       "unknown flux parameter `%s'", flux);
     g_free (flux);
   }
+
+  if (scheme) {
+    if (!strcmp (scheme, "godunov"))
+      par->scheme = GFS_GODUNOV;
+    else if (!strcmp (scheme, "none"))
+      par->scheme = GFS_NONE;
+    else if (!strcmp (scheme, "vof")) {
+      par->scheme = GFS_VOF;
+      if (fp->type != GTS_ERROR && (par->cfl <= 0. || par->cfl > 0.5))
+	gts_file_variable_error (fp, var, "cfl", 
+				 "cfl `%g' is out of range `]0,0.5]'", 
+				 par->cfl);
+    }
+    else if (fp->type != GTS_ERROR)
+      gts_file_variable_error (fp, var, "scheme",
+			       "unknown scheme parameter `%s'", scheme);
+    g_free (scheme);
+  }
 }
