@@ -23,6 +23,7 @@
 
 #include "simulation.h"
 #include "init.h"
+#include "solid.h"
 
 static void ftt_cell_neighbors_not_periodic (FttCell * cell,
 					     FttCellNeighbors * neighbor)
@@ -60,7 +61,7 @@ static FttCell * ftt_cell_neighbor_not_periodic (FttCell * cell,
   return neighbor;
 }
 
-#ifdef FTT_2D
+#if FTT_2D
 
 typedef struct {
   FttCell * c1, * c2;
@@ -74,6 +75,10 @@ static guint cell_edges (FttCell * cell,
   FttCell * n1, * n2, * d1, * d2;
   guint level = ftt_cell_level (cell);
 
+#if FTT_2D3
+  g_assert_not_implemented ();
+#endif
+
   ftt_cell_neighbors_not_periodic (cell, &neighbor);
 
   d1 = NULL;
@@ -852,7 +857,7 @@ static gint cell_index (FttCell * cell,
 	g_assert_not_reached ();
       gfs_array_add (a, val);
     }
-    ftt_cell_pos (cell, &p);
+    gfs_cell_cm (cell, &p);
     fp[0] = p.x/lambda->x;
     fp[1] = p.y/lambda->y;
     fp[2] = p.z/lambda->z;
@@ -862,7 +867,7 @@ static gint cell_index (FttCell * cell,
   return GFS_STATE (cell)->dp - 1.;
 }
 
-#ifdef FTT_2D
+#if FTT_2D
 
 static void add_cell (FttCell * cell, gpointer * par)
 {
@@ -932,7 +937,7 @@ static void add_solid_face (GtsTriangle * t, GfsArray * con)
   gfs_array_add (con, index);
 }
 
-#ifdef FTT_2D
+#if FTT_2D
 Error m_ImportGfs2D (Object * in, Object * out);
 
 Error m_ImportGfs2D (Object * in, Object * out)
@@ -996,7 +1001,7 @@ Error m_ImportGfs3D (Object * in, Object * out)
   gfs_domain_match (domain);
   v = domain->variables;
   while (v) {
-    gfs_domain_center_bc (domain, v);
+    gfs_domain_bc (domain, FTT_TRAVERSE_LEAFS, -1, v);
     v = v->next;
   }
 
@@ -1016,13 +1021,13 @@ Error m_ImportGfs3D (Object * in, Object * out)
 
     if (!strcmp (v->name, "U") && v->next &&
 	!strcmp (v->next->name, "V")
-#ifndef FTT_2D
+#if (!FTT_2D)
 	&& v->next->next && !strcmp (v->next->next->name, "W")
 #endif /* not FTT_2D */
 	) {
       vv[0] = v; v = v->next;
       vv[1] = v; 
-#ifndef FTT_2D
+#if (!FTT_2D)
       v = v->next;
       vv[2] = v;
 #endif /* not FTT_2D */
@@ -1033,7 +1038,7 @@ Error m_ImportGfs3D (Object * in, Object * out)
 	     (vv[0] = v) && (v1 = v->next) &&
 	     v1->name[strlen (v1->name) - 1] == 'y' &&
 	     (vv[1] = v1) 
-#ifndef FTT_2D
+#if (!FTT_2D)
 	     && (v1 = v1->next) &&
 	     v1->name[strlen (v1->name) - 1] == 'z' &&
 	     (vv[2] = v1)
