@@ -18,6 +18,8 @@
  */
 
 #include <stdlib.h>
+#include <sys/wait.h>
+#include <unistd.h>
 #include "event.h"
 #include "solid.h"
 
@@ -28,20 +30,30 @@ static gboolean gfs_event_event (GfsEvent * event, GfsSimulation * sim)
       sim->time.t > event->end || 
       sim->time.i > event->iend) {
     gts_object_destroy (GTS_OBJECT (event));
-    return FALSE;
+    return (event->realised = FALSE);
+  }
+  if (event->end_event) {
+    if (event->n == 0 &&
+	(sim->time.t >= sim->time.end ||
+	 sim->time.i >= sim->time.iend)) {
+      event->n = 1;
+      return (event->realised = TRUE);
+    }
+    else
+      return (event->realised = FALSE);
   }
   if (sim->time.t >= event->t) {
     if (event->istep < G_MAXINT) {
       if (event->n == 0) {
 	event->i = sim->time.i + event->istep;
 	event->n++;
-	return TRUE;
+	return (event->realised = TRUE);
       }
     }
     else {
       event->n++;
       event->t = event->start + event->n*event->step;
-      return TRUE;
+      return (event->realised = TRUE);
     }
   }
   if (sim->time.i >= event->i) {
@@ -50,16 +62,16 @@ static gboolean gfs_event_event (GfsEvent * event, GfsSimulation * sim)
 	event->start = sim->time.t;
 	event->t = event->start + event->step;
 	event->n = 1;
-	return TRUE;
+	return (event->realised = TRUE);
       }
     }
     else {
       event->n++;
       event->i += event->istep;
-      return TRUE;
+      return (event->realised = TRUE);
     }
   }
-  return FALSE;
+  return (event->realised = FALSE);
 }
 
 static void gfs_event_write (GtsObject * object, FILE * fp)
@@ -122,7 +134,7 @@ static void gfs_event_read (GtsObject ** o, GtsFile * fp)
     gts_file_error (fp, "expecting a string (GfsEventClass)");
     return;
   }
-  klass = gts_object_class_from_name (fp->token->str);
+  klass = gfs_object_class_from_name (fp->token->str);
   if (klass == NULL) {
     gts_file_error (fp, "unknown class `%s'", fp->token->str);
     return;
@@ -317,108 +329,212 @@ void gfs_event_set (GfsEvent * e,
     e->i = e->istart = G_MAXINT/2;
 }
 
-/* GfsInit: Object */
+/* GfsGenericInit: Object */
 
-static void gfs_init_init (GfsEvent * event)
+static void gfs_generic_init_init (GfsEvent * event)
 {
   event->start = -1;
 }
 
-GfsEventClass * gfs_init_class (void)
+GfsEventClass * gfs_generic_init_class (void)
 {
   static GfsEventClass * klass = NULL;
 
   if (klass == NULL) {
-    GtsObjectClassInfo gfs_init_info = {
-      "GfsInit",
+    GtsObjectClassInfo gfs_generic_init_info = {
+      "GfsGenericInit",
       sizeof (GfsEvent),
       sizeof (GfsEventClass),
       (GtsObjectClassInitFunc) NULL,
-      (GtsObjectInitFunc) gfs_init_init,
+      (GtsObjectInitFunc) gfs_generic_init_init,
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
     klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_event_class ()),
+				  &gfs_generic_init_info);
+  }
+
+  return klass;
+}
+
+/* GfsInit: Object */
+
+static void gfs_init_read (GtsObject ** o, GtsFile * fp)
+{
+  if (GTS_OBJECT_CLASS (gfs_init_class ())->parent_class->read)
+    (* GTS_OBJECT_CLASS (gfs_init_class ())->parent_class->read) 
+      (o, fp);
+  if (fp->type == GTS_ERROR)
+    return;
+  
+  if (fp->type != '{') {
+    gts_file_error (fp, "expecting an opening brace");
+    return;
+  }
+  fp->scope_max++;
+  gts_file_next_token (fp);
+  while (fp->type != GTS_ERROR && fp->type != '}') {
+    if (fp->type == '\n') {
+      gts_file_next_token (fp);
+      continue;
+    }
+    if (fp->type != GTS_STRING) {
+      gts_file_error (fp, "expecting a variable name");
+      return;
+    }
+    else {
+      GfsVariable * v = gfs_variable_from_name (GFS_DOMAIN (gfs_object_simulation (*o))->variables, fp->token->str);
+      GfsFunction * f;
+
+      if (!v) {
+	gts_file_error (fp, "unknown variable `%s'\n", fp->token->str);
+	return;
+      }
+      gts_file_next_token (fp);
+
+      if (fp->type != '=') {
+	 gts_file_error (fp, "expecting `=`");
+	 return;
+      }
+      gts_file_next_token (fp);
+
+      f = gfs_function_new (gfs_function_class (), 0.);
+      gfs_function_read (f, fp);
+      if (fp->type == GTS_ERROR) {
+	gts_object_destroy (GTS_OBJECT (f));
+	return;
+      }
+      g_hash_table_insert (GFS_INIT (*o)->f, v, f);
+    }
+  }
+  if (fp->type != '}') {
+    gts_file_error (fp, "expecting a closing brace");
+    return;
+  }
+  fp->scope_max--;
+  gts_file_next_token (fp);
+}
+
+static void write_f (GfsVariable * v, GfsFunction * f, FILE * fp)
+{
+  fprintf (fp, "  %s =", v->name);
+  gfs_function_write (f, fp);
+  fputc ('\n', fp);
+}
+
+static void gfs_init_write (GtsObject * o, FILE * fp)
+{
+  GfsInit * i = GFS_INIT (o);
+
+  if (GTS_OBJECT_CLASS (gfs_init_class ())->parent_class->write)
+    (* GTS_OBJECT_CLASS (gfs_init_class ())->parent_class->write) 
+      (o, fp);
+
+  fputs (" {\n", fp);
+  g_hash_table_foreach (i->f, (GHFunc) write_f, fp);
+  fputc ('}', fp);
+}
+
+static void destroy_f (GfsVariable * v, GtsObject * o)
+{
+  gts_object_destroy (o);
+}
+
+static void gfs_init_destroy (GtsObject * object)
+{
+  GfsInit * i = GFS_INIT (object);
+
+  g_hash_table_foreach (i->f, (GHFunc) destroy_f, NULL);
+  g_hash_table_destroy (i->f);
+
+  (* GTS_OBJECT_CLASS (gfs_init_class ())->parent_class->destroy) 
+    (object);
+}
+
+static void init_vf (FttCell * cell, gpointer * data)
+{
+  GfsVariable * v = data[0];
+  GfsFunction * f = data[1];
+  GfsSimulation * sim = data[2];
+  FttVector p;
+
+  if (v->centered)
+    ftt_cell_pos (cell, &p);
+  else
+    gfs_cell_cm (cell, &p);
+  GFS_VARIABLE (cell, v->i) = gfs_function_value (f, &p, sim->time.t);
+}
+
+static void init_f (GfsVariable * v, GfsFunction * f, GfsDomain * domain)
+{
+  gpointer data[3];
+  
+  data[0] = v;
+  data[1] = f;
+  data[2] = domain;
+  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			    (FttCellTraverseFunc) init_vf, data);
+}
+
+static gboolean gfs_init_event (GfsEvent * event, GfsSimulation * sim)
+{
+  if ((* GFS_EVENT_CLASS (GTS_OBJECT_CLASS (gfs_init_class ())->parent_class)->event) (event, sim)) {
+    g_hash_table_foreach (GFS_INIT (event)->f, (GHFunc) init_f, sim);
+    return TRUE;
+  }
+  return FALSE;
+}
+
+static void gfs_init_class_init (GfsGenericInitClass * klass)
+{
+  GFS_EVENT_CLASS (klass)->event = gfs_init_event;
+  GTS_OBJECT_CLASS (klass)->read = gfs_init_read;
+  GTS_OBJECT_CLASS (klass)->write = gfs_init_write;
+  GTS_OBJECT_CLASS (klass)->destroy = gfs_init_destroy;
+}
+
+static void gfs_init_init (GfsInit * object)
+{
+  object->f = g_hash_table_new (NULL, NULL);
+}
+
+GfsGenericInitClass * gfs_init_class (void)
+{
+  static GfsGenericInitClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_init_info = {
+      "GfsInit",
+      sizeof (GfsInit),
+      sizeof (GfsGenericInitClass),
+      (GtsObjectClassInitFunc) gfs_init_class_init,
+      (GtsObjectInitFunc) gfs_init_init,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_generic_init_class ()),
 				  &gfs_init_info);
   }
 
   return klass;
 }
 
-/* GfsInitFlowConstant: Object */
-
-static void init_flow_constant (FttCell * cell, GfsInitFlowConstant * flow)
-{
-  FttVector pos;
-
-  GFS_STATE (cell)->u = flow->u.x;
-  GFS_STATE (cell)->v = flow->u.y;
-#ifndef FTT_2D
-  GFS_STATE (cell)->w = flow->u.z;
-#endif
-  ftt_cell_pos (cell, &pos);
-  GFS_STATE (cell)->c = pos.y == 0. ? 0.5 : pos.y > 0. ? 0. : 1.;
-}
-
-static gboolean gfs_init_flow_constant_event (GfsEvent * event, 
-					      GfsSimulation * sim)
-{
-  if ((* GFS_EVENT_CLASS (GTS_OBJECT_CLASS (gfs_init_flow_constant_class ())->parent_class)->event) (event, sim)) {
-    gfs_domain_cell_traverse (GFS_DOMAIN (sim), 
-			     FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1, 
-			     (FttCellTraverseFunc) init_flow_constant, event);
-    return TRUE;
-  }
-  return FALSE;
-}
-
-static void gfs_init_flow_constant_write (GtsObject * o, FILE * fp)
-{
-  GfsInitFlowConstant * flow = GFS_INIT_FLOW_CONSTANT (o);
-
-  if (GTS_OBJECT_CLASS (gfs_init_flow_constant_class ())->parent_class->write)
-    (* GTS_OBJECT_CLASS (gfs_init_flow_constant_class ())->parent_class->write)
-      (o, fp);
-
-#ifdef FTT_2D
-  fprintf (fp, " { U = %g V = %g }", flow->u.x, flow->u.y);
-#else
-  fprintf (fp, " { U = %g V = %g W = %g }", flow->u.x, flow->u.y, flow->u.z);
-#endif
-}
+/* GfsInitFlowConstant: Object: fixme: deprecated */
 
 static void gfs_init_flow_constant_read (GtsObject ** o, GtsFile * fp)
 {
-  GfsInitFlowConstant * flow = GFS_INIT_FLOW_CONSTANT (*o);
-  GtsFileVariable var[] = {
-    {GTS_DOUBLE, "U", TRUE},
-    {GTS_DOUBLE, "V", TRUE},
-    {GTS_DOUBLE, "W", TRUE},
-    {GTS_NONE}
-  };
-
   if (GTS_OBJECT_CLASS (gfs_init_flow_constant_class ())->parent_class->read)
     (* GTS_OBJECT_CLASS (gfs_init_flow_constant_class ())->parent_class->read) 
       (o, fp);
   if (fp->type == GTS_ERROR)
     return;
 
-  var[0].data = &flow->u.x;
-  var[1].data = &flow->u.y;
-  var[2].data = &flow->u.z;
-  gts_file_assign_variables (fp, var);
+  g_warning ("GfsInitFlowConstant is deprecated you should use GfsInit instead");
 }
 
 static void gfs_init_flow_constant_class_init (GfsEventClass * klass)
 {
-  GFS_EVENT_CLASS (klass)->event = gfs_init_flow_constant_event;
   GTS_OBJECT_CLASS (klass)->read = gfs_init_flow_constant_read;
-  GTS_OBJECT_CLASS (klass)->write = gfs_init_flow_constant_write;
-}
-
-static void gfs_init_flow_constant_init (GfsInitFlowConstant * object)
-{
-  object->u.x = object->u.y = object->u.z = 0.;
 }
 
 GfsEventClass * gfs_init_flow_constant_class (void)
@@ -428,10 +544,10 @@ GfsEventClass * gfs_init_flow_constant_class (void)
   if (klass == NULL) {
     GtsObjectClassInfo gfs_init_flow_constant_info = {
       "GfsInitFlowConstant",
-      sizeof (GfsInitFlowConstant),
+      sizeof (GfsInit),
       sizeof (GfsEventClass),
       (GtsObjectClassInitFunc) gfs_init_flow_constant_class_init,
-      (GtsObjectInitFunc) gfs_init_flow_constant_init,
+      (GtsObjectInitFunc) NULL,
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
@@ -442,7 +558,7 @@ GfsEventClass * gfs_init_flow_constant_class (void)
   return klass;
 }
 
-#ifdef FTT_2D
+#if FTT_2D
 
 /* GfsInitVorticity: Object */
 
@@ -506,12 +622,12 @@ static void stream_from_vorticity (GfsDomain * domain,
   gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
 			    (FttCellTraverseFunc) gfs_cell_reset, 
 			    stream);
-  gfs_residual (domain, FTT_TRAVERSE_LEAFS, -1, stream, vorticity, gfs_res);
-  norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1);
+  gfs_residual (domain, FTT_DIMENSION, FTT_TRAVERSE_LEAFS, -1, stream, vorticity, gfs_res);
+  norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1, 1.);
   maxlevel = gfs_domain_depth (domain);
   while (norm.infty > tolerance && maxit) {
-    gfs_poisson (domain, 0, maxlevel, 4, stream, vorticity);
-    norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1);
+    gfs_poisson_cycle (domain, FTT_DIMENSION, 0, maxlevel, 4, stream, vorticity);
+    norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1, 1.);
     maxit--;
   }
   if (maxit == 0)
@@ -561,7 +677,7 @@ GfsInitVorticityClass * gfs_init_vorticity_class (void)
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_init_class ()),
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_generic_init_class ()),
 				  &gfs_init_vorticity_info);
   }
 
@@ -586,7 +702,7 @@ static void gfs_event_sum_write (GtsObject * o, FILE * fp)
 static void gfs_event_sum_read (GtsObject ** o, GtsFile * fp)
 {
   GfsEventSum * s = GFS_EVENT_SUM (*o);
-  GfsDomain * domain = gfs_object_simulation (s);
+  GfsDomain * domain =  GFS_DOMAIN (gfs_object_simulation (s));
 
   if (GTS_OBJECT_CLASS (gfs_event_sum_class ())->parent_class->read)
     (* GTS_OBJECT_CLASS (gfs_event_sum_class ())->parent_class->read) 
@@ -711,6 +827,258 @@ GfsEventClass * gfs_event_sum2_class (void)
   return klass;
 }
 
+/* GfsEventStop: Object */
+
+static void gfs_event_stop_write (GtsObject * o, FILE * fp)
+{
+  GfsEventStop * s = GFS_EVENT_STOP (o);
+
+  if (GTS_OBJECT_CLASS (gfs_event_stop_class ())->parent_class->write)
+    (* GTS_OBJECT_CLASS (gfs_event_stop_class ())->parent_class->write)
+      (o, fp);
+
+  fprintf (fp, " %s %g", s->v->name, s->max);
+}
+
+static void gfs_event_stop_read (GtsObject ** o, GtsFile * fp)
+{
+  GfsEventStop * s = GFS_EVENT_STOP (*o);
+  GfsDomain * domain =  GFS_DOMAIN (gfs_object_simulation (s));
+
+  if (GTS_OBJECT_CLASS (gfs_event_stop_class ())->parent_class->read)
+    (* GTS_OBJECT_CLASS (gfs_event_stop_class ())->parent_class->read) 
+      (o, fp);
+  if (fp->type == GTS_ERROR)
+    return;
+
+  if (fp->type != GTS_STRING) {
+    gts_file_error (fp, "expecting a string (v)");
+    return;
+  }
+  if (!(s->v = gfs_variable_from_name (domain->variables, fp->token->str))) {
+    gts_file_error (fp, "unknown variable `%s'", fp->token->str);
+    return;
+  }
+  gts_file_next_token (fp);
+
+  if (fp->type != GTS_INT && fp->type != GTS_FLOAT) {
+    gts_file_error (fp, "expecting a number (max)");
+    return;
+  }
+  s->max = atof (fp->token->str);
+  gts_file_next_token (fp);
+
+  s->oldv = gfs_domain_add_variable (domain, NULL);
+  s->oldv->fine_coarse = s->v->fine_coarse;
+  gts_file_next_token (fp);
+}
+
+static void diff (FttCell * cell, GfsEventStop * s)
+{
+  GFS_VARIABLE (cell, s->oldv->i) -= GFS_VARIABLE (cell, s->v->i);
+}
+
+static void copy (FttCell * cell, GfsEventStop * s)
+{
+  GFS_VARIABLE (cell, s->oldv->i) = GFS_VARIABLE (cell, s->v->i);
+}
+
+static gboolean gfs_event_stop_event (GfsEvent * event, GfsSimulation * sim)
+{
+  if ((* GFS_EVENT_CLASS (GTS_OBJECT_CLASS (gfs_event_stop_class ())->parent_class)->event) (event, sim)) {
+    GfsEventStop * s = GFS_EVENT_STOP (event);
+
+    if (s->last >= 0.) {
+      GfsNorm n;
+
+      gfs_domain_cell_traverse (GFS_DOMAIN (sim),
+				FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+				(FttCellTraverseFunc) diff, s);
+      n = gfs_domain_norm_variable (GFS_DOMAIN (sim), s->oldv,
+				    FTT_TRAVERSE_LEAFS, -1);
+      if (n.infty <= s->max)
+	sim->time.end = sim->time.t;
+    }
+    gfs_domain_cell_traverse (GFS_DOMAIN (sim),
+			      FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			      (FttCellTraverseFunc) copy, s);
+    s->last = sim->time.t;
+    return TRUE;
+  }
+  return FALSE;
+}
+
+static void gfs_event_stop_class_init (GfsEventClass * klass)
+{
+  GTS_OBJECT_CLASS (klass)->read = gfs_event_stop_read;
+  GTS_OBJECT_CLASS (klass)->write = gfs_event_stop_write;
+  GFS_EVENT_CLASS (klass)->event = gfs_event_stop_event;
+}
+
+static void gfs_event_stop_init (GfsEventStop * object)
+{
+  object->last = -1.;
+}
+
+GfsEventClass * gfs_event_stop_class (void)
+{
+  static GfsEventClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_event_stop_info = {
+      "GfsEventStop",
+      sizeof (GfsEventStop),
+      sizeof (GfsEventClass),
+      (GtsObjectClassInitFunc) gfs_event_stop_class_init,
+      (GtsObjectInitFunc) gfs_event_stop_init,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_event_class ()),
+				  &gfs_event_stop_info);
+  }
+
+  return klass;
+}
+
+/* GfsEventScript: Object */
+
+static void gfs_event_script_destroy (GtsObject * o)
+{
+  GfsEventScript * s = GFS_EVENT_SCRIPT (o);
+
+  if (s->script) g_string_free (s->script, TRUE);
+
+  (* GTS_OBJECT_CLASS (gfs_event_script_class ())->parent_class->destroy) (o);
+} 
+
+static void gfs_event_script_write (GtsObject * o, FILE * fp)
+{
+  GfsEventScript * s = GFS_EVENT_SCRIPT (o);
+
+  if (GTS_OBJECT_CLASS (gfs_event_script_class ())->parent_class->write)
+    (* GTS_OBJECT_CLASS (gfs_event_script_class ())->parent_class->write)
+      (o, fp);
+
+  fputs (" {", fp);
+  if (s->script)
+    fputs (s->script->str, fp);
+  fputc ('}', fp);
+}
+
+static void gfs_event_script_read (GtsObject ** o, GtsFile * fp)
+{
+  GfsEventScript * s = GFS_EVENT_SCRIPT (*o);
+  guint scope;
+  gint c;
+
+  if (GTS_OBJECT_CLASS (gfs_event_script_class ())->parent_class->read)
+    (* GTS_OBJECT_CLASS (gfs_event_script_class ())->parent_class->read) 
+      (o, fp);
+  if (fp->type == GTS_ERROR)
+    return;
+
+  if (fp->type != '{') {
+    gts_file_error (fp, "expecting an opening brace");
+    return;
+  }
+  if (s->script)
+    g_string_free (s->script, TRUE);
+  s->script = g_string_new ("");
+  scope = fp->scope_max;
+  c = gts_file_getc (fp);
+  while (c != EOF && fp->scope > scope) {
+    g_string_append_c (s->script, c);
+    c = gts_file_getc (fp);
+  }
+  if (fp->scope != scope) {
+    gts_file_error (fp, "parse error");
+    return;
+  }
+  gts_file_next_token (fp);
+}
+
+static gboolean gfs_event_script_event (GfsEvent * event, GfsSimulation * sim)
+{
+  if ((* GFS_EVENT_CLASS (GTS_OBJECT_CLASS (gfs_event_script_class ())->parent_class)->event) 
+      (event, sim)) {
+    GfsEventScript * s = GFS_EVENT_SCRIPT (event);
+    if (s->script) {
+      gchar * scommand;
+      gchar sname[] = "/tmp/gfsXXXXXX";
+      gchar ename[] = "/tmp/gfsXXXXXX";
+      gint sf = mkstemp (sname);
+      gint ef = mkstemp (ename);
+      gint status;
+      FILE * f;
+
+      if (sf < 0 || ef < 0) {
+	g_warning ("GfsEventScript cannot create temporary files");
+	return TRUE;
+      }
+      f = fdopen (sf, "w");
+      fputs (s->script->str, f);
+      fclose (f);
+      scommand = g_strdup_printf ("GfsTime=%g GfsIter=%d GfsPid=%d "
+				  "GFS_STOP=%d sh %s",
+				  sim->time.t, sim->time.i, 
+				  GFS_DOMAIN (sim)->pid,
+				  GFS_EVENT_SCRIPT_STOP,
+				  sname);
+      fflush (stdout);
+      fflush (stderr);
+      status = system (scommand);
+      if (status != -1)
+	status = WEXITSTATUS (status);
+      g_free (scommand);
+      remove (sname);
+      if (status == GFS_EVENT_SCRIPT_STOP)
+	sim->time.end = sim->time.t;
+      else if (status == -1 || status != 0) {
+	FILE * ferr = fdopen (ef, "r");
+	gint c;
+
+	fputs ("Error while executing GfsEventScript:\n", stderr);
+	while ((c = fgetc (ferr)) != EOF)
+	  fputc (c, stderr);
+	fclose (ferr);
+      }
+      remove (ename);
+    }
+    return TRUE;
+  }
+  return FALSE;
+}
+
+static void gfs_event_script_class_init (GfsEventClass * klass)
+{
+  GTS_OBJECT_CLASS (klass)->destroy = gfs_event_script_destroy;
+  GTS_OBJECT_CLASS (klass)->read = gfs_event_script_read;
+  GTS_OBJECT_CLASS (klass)->write = gfs_event_script_write;
+  GFS_EVENT_CLASS (klass)->event = gfs_event_script_event;
+}
+
+GfsEventClass * gfs_event_script_class (void)
+{
+  static GfsEventClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_event_script_info = {
+      "GfsEventScript",
+      sizeof (GfsEventScript),
+      sizeof (GfsEventClass),
+      (GtsObjectClassInitFunc) gfs_event_script_class_init,
+      (GtsObjectInitFunc) NULL,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_event_class ()),
+				  &gfs_event_script_info);
+  }
+
+  return klass;
+}
+
 /* GfsInitFraction: Object */
 
 static void gfs_init_fraction_destroy (GtsObject * object)
@@ -739,7 +1107,7 @@ static void gfs_init_fraction_read (GtsObject ** o, GtsFile * fp)
     return;
 
   init = GFS_INIT_FRACTION (*o);
-  domain = gfs_object_simulation (init);
+  domain =  GFS_DOMAIN (gfs_object_simulation (init));
   if (fp->type != GTS_STRING) {
     gts_file_error (fp, "expecting a string (variable)");
     return;
@@ -870,9 +1238,88 @@ GfsInitFractionClass * gfs_init_fraction_class (void)
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_init_class ()),
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_generic_init_class ()),
 				  &gfs_init_fraction_info);
   }
 
   return klass;
 }
+
+/* GfsRemoveDroplets: Object */
+
+static gboolean gfs_remove_droplets_event (GfsEvent * event, GfsSimulation * sim)
+{
+  if ((* GFS_EVENT_CLASS (GTS_OBJECT_CLASS (gfs_remove_droplets_class ())->parent_class)->event) (event, sim)) {
+    gfs_domain_remove_droplets (GFS_DOMAIN (sim), 
+				GFS_REMOVE_DROPLETS (event)->c, 
+				GFS_REMOVE_DROPLETS (event)->min);
+    return TRUE;
+  }
+  return FALSE;
+}
+
+static void gfs_remove_droplets_read (GtsObject ** o, GtsFile * fp)
+{
+  GfsDomain * domain;
+
+  if (GTS_OBJECT_CLASS (gfs_remove_droplets_class ())->parent_class->read)
+    (* GTS_OBJECT_CLASS (gfs_remove_droplets_class ())->parent_class->read) 
+      (o, fp);
+  if (fp->type == GTS_ERROR)
+    return;
+
+  if (fp->type != GTS_STRING) {
+    gts_file_error (fp, "expecting a string (variable)");
+    return;
+  }
+  domain = GFS_DOMAIN (gfs_object_simulation (*o));
+  if ((GFS_REMOVE_DROPLETS (*o)->c = gfs_variable_from_name (domain->variables, fp->token->str))
+      == NULL) {
+    gts_file_error (fp, "unknown variable `%s'", fp->token->str);
+    return;
+  }
+  gts_file_next_token (fp);
+
+  if (fp->type != GTS_INT) {
+    gts_file_error (fp, "expecting an integer (min)");
+    return;
+  }
+  GFS_REMOVE_DROPLETS (*o)->min = atoi (fp->token->str);
+  gts_file_next_token (fp);
+}
+
+static void gfs_remove_droplets_write (GtsObject * o, FILE * fp)
+{
+  if (GTS_OBJECT_CLASS (gfs_remove_droplets_class ())->parent_class->write)
+    (* GTS_OBJECT_CLASS (gfs_remove_droplets_class ())->parent_class->write) 
+      (o, fp);
+  fprintf (fp, " %s %d", GFS_REMOVE_DROPLETS (o)->c->name, GFS_REMOVE_DROPLETS (o)->min);
+}
+
+static void gfs_remove_droplets_class_init (GfsEventClass * klass)
+{
+  GFS_EVENT_CLASS (klass)->event = gfs_remove_droplets_event;
+  GTS_OBJECT_CLASS (klass)->read = gfs_remove_droplets_read;
+  GTS_OBJECT_CLASS (klass)->write = gfs_remove_droplets_write;
+}
+
+GfsEventClass * gfs_remove_droplets_class (void)
+{
+  static GfsEventClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_remove_droplets_info = {
+      "GfsRemoveDroplets",
+      sizeof (GfsRemoveDroplets),
+      sizeof (GfsEventClass),
+      (GtsObjectClassInitFunc) gfs_remove_droplets_class_init,
+      (GtsObjectInitFunc) NULL,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_event_class ()),
+				  &gfs_remove_droplets_info);
+  }
+
+  return klass;
+}
