@@ -24,17 +24,17 @@
 
 gchar * ftt_direction_name[FTT_NEIGHBORS] = {
   "right", "left", "top", "bottom"
-#ifndef FTT_2D
+#if (!FTT_2D)
   , "front", "back"
-#endif /* FTT_3D */
+#endif /* FTT_3D || FTT_2D3 */
 };
 
 gint ftt_opposite_direction[FTT_NEIGHBORS] =
-#ifdef    FTT_2D
+#if      FTT_2D
   {1, 0, 3, 2};
-#else  /* FTT_3D */
+#else  /* FTT_3D || FTT_2D3 */
   {1, 0, 3, 2, 5, 4};
-#endif /* FTT_3D */
+#endif /* FTT_3D || FTT_2D3 */
 
 typedef struct _FttOct      FttOct;
 typedef struct _FttRootCell FttRootCell;
@@ -86,7 +86,7 @@ gdouble ftt_cell_volume (const FttCell * cell)
   g_return_val_if_fail (cell != NULL, 0.);
 
   size = ftt_level_size (ftt_cell_level (cell));
-#ifdef FTT_2D
+#if (FTT_2D || FTT_2D3)
   return size*size;
 #else  /* FTT_3D */
   return size*size*size;
@@ -124,24 +124,27 @@ void ftt_cell_children (const FttCell * cell,
  * @d: a direction.
  * @children: a #FttCellChildren.
  *
- * Fills @children with the FTT_CELLS/2 children (2 in 2D, 4 in 3D) of
- * @cell in direction @d.
+ * Fills @children with the children (2 in 2D, 4 in 3D, 2 or 4 in 2D3)
+ * of @cell in direction @d.
  * 
- * This function fails if @cell is a leaf.  
+ * This function fails if @cell is a leaf.
+ *
+ * Returns: the number of children in direction @d.
  */
-void ftt_cell_children_direction (const FttCell * cell,
-				  FttDirection d,
-				  FttCellChildren * children)
+guint ftt_cell_children_direction (const FttCell * cell,
+				   FttDirection d,
+				   FttCellChildren * children)
 {
   struct _FttOct * oct;
   guint i;
-  static gint index[FTT_NEIGHBORS][FTT_CELLS/2] =
-#ifdef    FTT_2D
+#if (FTT_2D || FTT_2D3)
+  static gint index[FTT_NEIGHBORS_2D][FTT_CELLS/2] =
   {{1, 3},
    {0, 2},
    {0, 1},
    {2, 3}};
 #else  /* FTT_3D */
+  static gint index[FTT_NEIGHBORS][FTT_CELLS/2] =
   {{1, 3, 5, 7},
    {0, 2, 4, 6},
    {0, 1, 4, 5},
@@ -150,15 +153,25 @@ void ftt_cell_children_direction (const FttCell * cell,
    {4, 5, 6, 7}};
 #endif /* FTT_3D */
 
-  g_return_if_fail (cell != NULL);
-  g_return_if_fail (!FTT_CELL_IS_LEAF (cell));
-  g_return_if_fail (d < FTT_NEIGHBORS);
-  g_return_if_fail (children != NULL);
+  g_return_if_fail (cell != NULL, 0);
+  g_return_if_fail (!FTT_CELL_IS_LEAF (cell), 0);
+  g_return_if_fail (d < FTT_NEIGHBORS, 0);
+  g_return_if_fail (children != NULL, 0);
 
   oct = cell->children;
+
+#if FTT_2D3
+  if (d >= FTT_NEIGHBORS_2D) {
+    for (i = 0; i < FTT_CELLS; i++)
+      children->c[i] = FTT_CELL_IS_DESTROYED (&(oct->cell[i])) ? NULL : &(oct->cell[i]);
+    return FTT_CELLS;
+  }
+#endif /* 2D3 */
+
   for (i = 0; i < FTT_CELLS/2; i++)
     children->c[i] = FTT_CELL_IS_DESTROYED (&(oct->cell[index[d][i]])) ? 
       NULL : &(oct->cell[index[d][i]]);
+  return FTT_CELLS/2;
 }
 
 /**
@@ -173,8 +186,8 @@ void ftt_cell_children_direction (const FttCell * cell,
 FttCell * ftt_cell_child_corner (const FttCell * cell,
 				 FttDirection d[FTT_DIMENSION])
 {
-#ifdef FTT_2D  
-  static gint index[FTT_NEIGHBORS][FTT_NEIGHBORS] = {
+#if (FTT_2D || FTT_2D3)
+  static gint index[FTT_NEIGHBORS_2D][FTT_NEIGHBORS_2D] = {
     {-1,-1,1,3},
     {-1,-1,0,2},
     {1,0,-1,-1},
@@ -186,6 +199,10 @@ FttCell * ftt_cell_child_corner (const FttCell * cell,
   g_return_val_if_fail (!FTT_CELL_IS_LEAF (cell), NULL);
   g_return_val_if_fail (d[0] < FTT_NEIGHBORS, NULL);
   g_return_val_if_fail (d[1] < FTT_NEIGHBORS, NULL);
+#  if FTT_2D3
+  if (d[0] >= FTT_NEIGHBORS_2D|| d[1] >= FTT_NEIGHBORS_2D)
+    g_assert_not_implemented ();
+#  endif
 
   i = index[d[0]][d[1]];
 #else  /* FTT_3D */
@@ -222,7 +239,8 @@ FttCell * ftt_cell_child_corner (const FttCell * cell,
 
   g_return_val_if_fail (i >= 0, NULL);
 
-  return &(cell->children->cell[i]);
+  return FTT_CELL_IS_DESTROYED (&(cell->children->cell[i])) ? NULL:
+    &(cell->children->cell[i]);
 }
 
 /**
@@ -237,11 +255,18 @@ void ftt_cell_neighbors_not_cached (const FttCell * cell,
 				    FttCellNeighbors * neighbors)
 {
   static gint neighbor_index[FTT_NEIGHBORS][FTT_CELLS]
-#ifdef FTT_2D
+#if FTT_2D
     = {{1,-1,3,-3},
        {-2,0,-4,2},
        {-3,-4,0,1},
        {2,3,-1,-2}};
+#elif FTT_2D3
+    = {{1,-1,3,-3},
+       {-2,0,-4,2},
+       {-3,-4,0,1},
+       {2,3,-1,-2},
+       {-1,-2,-3,-4},
+       {-1,-2,-3,-4}};
 #else  /* FTT_3D */
     = {{1,-1,3,-3,5,-5,7,-7},
        {-2,0,-4,2,-6,4,-8,6},
@@ -295,11 +320,18 @@ FttCell * ftt_cell_neighbor_not_cached (const FttCell * cell,
 					FttDirection d)
 {
   static gint neighbor_index[FTT_NEIGHBORS][FTT_CELLS]
-#ifdef FTT_2D
+#if FTT_2D
     = {{1,-1,3,-3},
        {-2,0,-4,2},
        {-3,-4,0,1},
        {2,3,-1,-2}};
+#elif FTT_2D3
+    = {{1,-1,3,-3},
+       {-2,0,-4,2},
+       {-3,-4,0,1},
+       {2,3,-1,-2},
+       {-1,-2,-3,-4},
+       {-1,-2,-3,-4}};
 #else  /* FTT_3D */
     = {{1,-1,3,-3,5,-5,7,-7},
        {-2,0,-4,2,-6,4,-8,6},
@@ -345,8 +377,7 @@ void ftt_cell_neighbors (const FttCell * cell,
   g_return_if_fail (neighbors != NULL);
 
   if (!FTT_CELL_IS_LEAF (cell) && neighbors != &cell->children->neighbors) {
-    memcpy (neighbors, &cell->children->neighbors, 
-	    sizeof (FttCellNeighbors));
+    memcpy (neighbors, &cell->children->neighbors, sizeof (FttCellNeighbors));
     return;
   }
 
@@ -411,6 +442,35 @@ FttFaceType ftt_face_type (const FttCellFace * face)
   g_assert (ftt_cell_level (face->cell) == ftt_cell_level (face->neighbor));
   return FTT_FINE_FINE;
 }
+
+/**
+ * ftt_cell_neighbor_is_brother:
+ * @cell: a #FttCell.
+ * @d: a #FttDirection.
+ *
+ * Returns: %TRUE if a (potential) neighbor of @cell in direction @d
+ * and @cell would have the same parent, %FALSE otherwise.
+ */
+gboolean ftt_cell_neighbor_is_brother (FttCell * cell, 
+				       FttDirection d)
+{
+  static gboolean b[FTT_CELLS][FTT_NEIGHBORS] = {
+#if FTT_2D
+    {1,0,0,1}, {0,1,0,1}, {1,0,1,0}, {0,1,1,0}
+#elif FTT_2D3
+    {1,0,0,1,0,0}, {0,1,0,1,0,0}, {1,0,1,0,0,0}, {0,1,1,0,0,0}
+#else  /* 3D */
+    {1,0,0,1,0,1}, {0,1,0,1,0,1}, {1,0,1,0,0,1}, {0,1,1,0,0,1},
+    {1,0,0,1,1,0}, {0,1,0,1,1,0}, {1,0,1,0,1,0}, {0,1,1,0,1,0}
+#endif /* 3D */
+  };
+
+  g_return_val_if_fail (cell != NULL, FALSE);
+  
+  if (FTT_CELL_IS_ROOT (cell))
+    return FALSE;
+  return b[FTT_CELL_ID (cell)][d];
+}
 #endif /* not G_CAN_INLINE */
 
 static void oct_new (FttCell * parent,
@@ -427,6 +487,9 @@ static void oct_new (FttCell * parent,
   oct = g_malloc0 (sizeof (FttOct));
   oct->level = ftt_cell_level (parent);
   oct->parent = parent;
+#ifdef FTT_2D3
+  oct->dz = ftt_cell_dz (parent);
+#endif
   ftt_cell_pos (parent, &(oct->pos));
   ftt_cell_neighbors (parent, &(oct->neighbors));
 
@@ -466,6 +529,9 @@ FttCell * ftt_cell_new (FttCellInitFunc init,
   FttCell * cell;
 
   cell = g_malloc0 (sizeof (FttRootCell));
+#ifdef FTT_2D3
+  FTT_ROOT_CELL (cell)->dz = 1.;
+#endif
   if (init)
     (* init) (cell, data);
 
@@ -491,8 +557,7 @@ gboolean ftt_cell_check (const FttCell * cell)
     if (neighbor.c[i] && 
 	!FTT_CELL_IS_LEAF (neighbor.c[i]) &&
 	ftt_cell_level (neighbor.c[i]) == level &&
-	neighbor.c[i]->children->neighbors.c[FTT_OPPOSITE_DIRECTION (i)]
-	!= cell) {
+	neighbor.c[i]->children->neighbors.c[FTT_OPPOSITE_DIRECTION (i)] != cell) {
       g_warning ("ftt_cell_check (%p): neighbor %d = %p: %d/%d",
 		 cell, 
 		 i,		 
@@ -618,14 +683,14 @@ void ftt_face_draw (const FttCellFace * face, FILE * fp)
 {
   gdouble size;
   FttVector p;
-#ifdef FTT_2D
+#if FTT_2D
   static FttVector dp[FTT_NEIGHBORS][2] = {
     {{1.,-1.,0.},{1.,1.,0.}},
     {{-1.,1.,0.},{-1.,-1,0.}},
     {{1.,1.,0.},{-1.,1.,0.}},
     {{-1.,-1.,0.},{1.,-1.,0.}}
   };
-#else  /* FTT_3D */
+#else  /* FTT_3D || FTT_2D3 */
   static FttVector dp[FTT_NEIGHBORS][4] = {
     {{1.,-1.,1.},{1.,-1.,-1.},{1.,1.,-1.},{1.,1.,1.}},
     {{-1.,-1.,1.},{-1.,-1.,-1.},{-1.,1.,-1.},{-1.,1.,1.}},
@@ -634,20 +699,20 @@ void ftt_face_draw (const FttCellFace * face, FILE * fp)
     {{1.,-1.,1.},{1.,1.,1.},{-1.,1.,1.},{-1.,-1.,1.}},
     {{1.,-1.,-1.},{1.,1.,-1.},{-1.,1.,-1.},{-1.,-1.,-1.}},
   };
-#endif /* FTT_3D */
+#endif /* FTT_3D ||  FTT_2D3 */
 
   g_return_if_fail (face != NULL);
   g_return_if_fail (fp != NULL);
 
   size = ftt_cell_size (face->cell)/2.;
   ftt_cell_pos (face->cell, &p);
-#ifdef FTT_2D
+#if FTT_2D
   fprintf (fp, "VECT 1 2 0 2 0 %g %g 0 %g %g 0\n",
 	   p.x + dp[face->d][0].x*size, 
 	   p.y + dp[face->d][0].y*size,
 	   p.x + dp[face->d][1].x*size, 
 	   p.y + dp[face->d][1].y*size);
-#else  /* FTT_3D */
+#else /* FTT_3D ||  FTT_2D3 */
   fprintf (fp, 
 	   "OFF 4 1 4 "
 	   "%g %g %g "
@@ -667,11 +732,11 @@ void ftt_face_draw (const FttCellFace * face, FILE * fp)
 	   p.x + dp[face->d][3].x*size,
 	   p.y + dp[face->d][3].y*size,
 	   p.z + dp[face->d][3].z*size);
-#endif /* FTT_3D */
+#endif /* FTT_3D ||  FTT_2D3 */
 }
 
 static gdouble coords[FTT_CELLS][3] =
-#ifdef FTT_2D
+#if (FTT_2D || FTT_2D3)
  {{-1., 1.,0.},
   { 1., 1.,0.},
   {-1.,-1.,0.},
@@ -738,6 +803,68 @@ void ftt_cell_pos (const FttCell * cell,
   }
 }
 
+/**
+ * ftt_corner_relative_pos:
+ * @cell: a #FttCell.
+ * @d: a set of perpendicular directions.
+ * @pos: a #FttVector.
+ *
+ * Fills @pos with the coordinates (normalised by the size of @cell)
+ * of the corner of @cell defined by @d relative to the position of
+ * the center of @cell.
+ */
+void ftt_corner_relative_pos (const FttCell * cell,
+			      FttDirection d[FTT_DIMENSION],
+			      FttVector * pos)
+{
+  static gdouble coords[FTT_NEIGHBORS][3] =
+#if FTT_2D
+    {{0.5,0.,0.},{-0.5,0.,0.},{0.,0.5,0.},{0.,-0.5,0.}};
+#else  /* FTT_3D || FTT_2D3 */
+    {{0.5,0.,0.},{-0.5,0.,0.},{0.,0.5,0.},{0.,-0.5,0.},{0.,0.,0.5},{0.,0.,-0.5}};
+#endif /* FTT_3D || FTT_2D3 */
+
+  g_return_if_fail (cell != NULL);
+  g_return_if_fail (pos != NULL);
+
+#if FTT_2D
+  pos->x = coords[d[0]][0] + coords[d[1]][0];
+  pos->y = coords[d[0]][1] + coords[d[1]][1];
+  pos->z = 0.;
+#else  /* FTT_3D || FTT_2D3 */
+  pos->x = coords[d[0]][0] + coords[d[1]][0] + coords[d[2]][0];
+  pos->y = coords[d[0]][1] + coords[d[1]][1] + coords[d[2]][1];
+  pos->z = coords[d[0]][2] + coords[d[1]][2] + coords[d[2]][2];
+#endif /* FTT_3D || FTT_2D3 */
+}
+
+/**
+ * ftt_corner_pos:
+ * @cell: a #FttCell.
+ * @d: a set of perpendicular directions.
+ * @pos: a #FttVector.
+ *
+ * Fills @pos with the coordinates of the corner of @cell defined by
+ * @d.
+ */
+void ftt_corner_pos (const FttCell * cell,
+		     FttDirection d[FTT_DIMENSION],
+		     FttVector * pos)
+{
+  gdouble size;
+  FttVector p;
+
+  g_return_if_fail (cell != NULL);
+  g_return_if_fail (pos != NULL);
+
+  ftt_corner_relative_pos (cell, d, pos);
+  ftt_cell_pos (cell, &p);
+  size = ftt_cell_size (cell);
+  pos->x = p.x + size*pos->x;
+  pos->y = p.y + size*pos->y;
+  pos->z = p.z + size*pos->z;
+}
+
 /**
  * ftt_face_pos:
  * @face: a #FttCellFace.
@@ -749,11 +876,11 @@ void ftt_face_pos (const FttCellFace * face, FttVector * pos)
 {
   gdouble size;
   static gdouble coords[FTT_NEIGHBORS][3] =
-#ifdef FTT_2D
+#if FTT_2D
   {{1.,0.,0.},{-1.,0.,0.},{0.,1.,0.},{0.,-1.,0.}};
-#else  /* FTT_3D */
+#else  /* FTT_3D || FTT_2D3 */
   {{1.,0.,0.},{-1.,0.,0.},{0.,1.,0.},{0.,-1.,0.},{0.,0.,1.},{0.,0.,-1.}};
-#endif /* FTT_3D */
+#endif /* FTT_3D || FTT_2D3 */
 
   g_return_if_fail (face != NULL);
   g_return_if_fail (pos != NULL);
@@ -838,7 +965,7 @@ static void update_neighbor (FttCell * cell,
     if (neighbor) {
       FttOct * oct = cell->children;
       FttCellChildren children;
-      guint i;
+      guint i, n;
 
       g_assert (oct->neighbors.c[d] == NULL ||
 		oct->neighbors.c[d] == neighbor);
@@ -850,8 +977,8 @@ static void update_neighbor (FttCell * cell,
       }
       
       g_assert (ftt_cell_level (oct->neighbors.c[d]) == oct->level);
-      ftt_cell_children_direction (cell, d, &children);
-      for (i = 0; i < FTT_CELLS/2; i++)
+      n = ftt_cell_children_direction (cell, d, &children);
+      for (i = 0; i < n; i++)
 	if (children.c[i])
 	  update_neighbor (children.c[i], d, init, init_data);
     }
@@ -914,7 +1041,7 @@ static void update_neighbor_match (FttCell * cell,
     if (neighbor) {
       FttOct * oct = cell->children;
       FttCellChildren children;
-      guint i;
+      guint i, n;
 
       g_assert (oct->neighbors.c[d] == NULL ||
 		oct->neighbors.c[d] == neighbor);
@@ -928,8 +1055,8 @@ static void update_neighbor_match (FttCell * cell,
 	oct_new (neighbor, TRUE, init, init_data);
       
       g_assert (ftt_cell_level (oct->neighbors.c[d]) == oct->level);
-      ftt_cell_children_direction (cell, d, &children);
-      for (i = 0; i < FTT_CELLS/2; i++)
+      n = ftt_cell_children_direction (cell, d, &children);
+      for (i = 0; i < n; i++)
 	if (children.c[i])
 	  update_neighbor_match (children.c[i], d, init, init_data);
     }
@@ -941,11 +1068,11 @@ static void update_neighbor_match (FttCell * cell,
       g_assert (ftt_cell_level (cell) == ftt_cell_level (neighbor));
       if (!FTT_CELL_IS_LEAF (neighbor)) {
 	FttCellChildren children;
-	guint i;
+	guint i, n;
 
 	oct_new (cell, TRUE, init, init_data);
-	ftt_cell_children_direction (cell, d, &children);
-	for (i = 0; i < FTT_CELLS/2; i++)
+	n = ftt_cell_children_direction (cell, d, &children);
+	for (i = 0; i < n; i++)
 	  if (children.c[i])
 	    update_neighbor_match (children.c[i], d, init, init_data);
       }
@@ -1286,12 +1413,12 @@ static void cell_traverse_boundary_pre_order_all (FttCell * cell,
 
   if (!FTT_CELL_IS_LEAF (cell)) {
     FttCellChildren children;
-    guint n;
+    guint i, n;
 
-    ftt_cell_children_direction (cell, d, &children);
-    for (n = 0; n < FTT_CELLS/2; n++)
-      if (children.c[n])
-	cell_traverse_boundary_pre_order_all (children.c[n], d, 
+    n = ftt_cell_children_direction (cell, d, &children);
+    for (i = 0; i < n; i++)
+      if (children.c[i])
+	cell_traverse_boundary_pre_order_all (children.c[i], d, 
 					      max_depth, func, data);
   }
 }
@@ -1307,12 +1434,12 @@ static void cell_traverse_boundary_post_order_all (FttCell * cell,
 
   if (!FTT_CELL_IS_LEAF (cell)) {
     FttCellChildren children;
-    guint n;
+    guint i, n;
 
-    ftt_cell_children_direction (cell, d, &children);
-    for (n = 0; n < FTT_CELLS/2; n++)
-      if (children.c[n])
-	cell_traverse_boundary_post_order_all (children.c[n], d, 
+    n = ftt_cell_children_direction (cell, d, &children);
+    for (i = 0; i < n; i++)
+      if (children.c[i])
+	cell_traverse_boundary_post_order_all (children.c[i], d, 
 					       max_depth, func, data);
   }
 
@@ -1332,12 +1459,12 @@ static void cell_traverse_boundary_leafs (FttCell * cell,
     (* func) (cell, data);
   else {
     FttCellChildren children;
-    guint n;
+    guint i, n;
 
-    ftt_cell_children_direction (cell, d, &children);
-    for (n = 0; n < FTT_CELLS/2; n++)
-      if (children.c[n])
-	cell_traverse_boundary_leafs (children.c[n], d, 
+    n = ftt_cell_children_direction (cell, d, &children);
+    for (i = 0; i < n; i++)
+      if (children.c[i])
+	cell_traverse_boundary_leafs (children.c[i], d, 
 				      max_depth, func, data);
   }
 }
@@ -1359,12 +1486,12 @@ static void cell_traverse_boundary_pre_order_nonleafs (FttCell * cell,
     g_assert (parent == NULL || parent->children != NULL);
     if (!FTT_CELL_IS_LEAF (cell)) {
       FttCellChildren children;
-      guint n;
+      guint i, n;
 
-      ftt_cell_children_direction (cell, d, &children);
-      for (n = 0; n < FTT_CELLS/2; n++)
-	if (children.c[n])
-	  cell_traverse_boundary_pre_order_nonleafs (children.c[n], d, 
+      n = ftt_cell_children_direction (cell, d, &children);
+      for (i = 0; i < n; i++)
+	if (children.c[i])
+	  cell_traverse_boundary_pre_order_nonleafs (children.c[i], d, 
 						     max_depth, func, data);
     }
   }
@@ -1381,12 +1508,12 @@ static void cell_traverse_boundary_post_order_nonleafs (FttCell * cell,
 
   if (!FTT_CELL_IS_LEAF (cell)) {
     FttCellChildren children;
-    guint n;
+    guint i, n;
 
-    ftt_cell_children_direction (cell, d, &children);
-    for (n = 0; n < FTT_CELLS/2; n++)
-      if (children.c[n])
-	cell_traverse_boundary_post_order_nonleafs (children.c[n], d, 
+    n = ftt_cell_children_direction (cell, d, &children);
+    for (i = 0; i < n; i++)
+      if (children.c[i])
+	cell_traverse_boundary_post_order_nonleafs (children.c[i], d, 
 						    max_depth, func, data);
     (* func) (cell, data);
   }
@@ -1402,12 +1529,12 @@ static void cell_traverse_boundary_level (FttCell * cell,
     (* func) (cell, data);
   else if (!FTT_CELL_IS_LEAF (cell)) {
     FttCellChildren children;
-    guint n;
+    guint i, n;
 
-    ftt_cell_children_direction (cell, d, &children);
-    for (n = 0; n < FTT_CELLS/2; n++)
-      if (children.c[n])
-	cell_traverse_boundary_level (children.c[n], d, 
+    n = ftt_cell_children_direction (cell, d, &children);
+    for (i = 0; i < n; i++)
+      if (children.c[i])
+	cell_traverse_boundary_level (children.c[i], d, 
 				      max_depth, func, data);
   }
 }
@@ -1422,12 +1549,12 @@ static void cell_traverse_boundary_level_leafs (FttCell * cell,
     (* func) (cell, data);
   else if (!FTT_CELL_IS_LEAF (cell)) {
     FttCellChildren children;
-    guint n;
+    guint i, n;
 
-    ftt_cell_children_direction (cell, d, &children);
-    for (n = 0; n < FTT_CELLS/2; n++)
-      if (children.c[n])
-	cell_traverse_boundary_level_leafs (children.c[n], d, 
+    n = ftt_cell_children_direction (cell, d, &children);
+    for (i = 0; i < n; i++)
+      if (children.c[i])
+	cell_traverse_boundary_level_leafs (children.c[i], d, 
 					    max_depth, func, data);
   }
 }
@@ -1442,12 +1569,12 @@ static void cell_traverse_boundary_level_non_leafs (FttCell * cell,
     (* func) (cell, data);
   else if (!FTT_CELL_IS_LEAF (cell)) {
     FttCellChildren children;
-    guint n;
+    guint i, n;
 
-    ftt_cell_children_direction (cell, d, &children);
-    for (n = 0; n < FTT_CELLS/2; n++)
-      if (children.c[n])
-	cell_traverse_boundary_level_non_leafs (children.c[n], d, 
+    n = ftt_cell_children_direction (cell, d, &children);
+    for (i = 0; i < n; i++)
+      if (children.c[i])
+	cell_traverse_boundary_level_non_leafs (children.c[i], d, 
 						max_depth, func, data);
   }
 }
@@ -1513,7 +1640,8 @@ void ftt_cell_traverse_boundary (FttCell * root,
 }
 
 static void oct_destroy (FttOct * oct,
-			 FttCellCleanupFunc cleanup)
+			 FttCellCleanupFunc cleanup,
+			 gpointer data)
 {
   guint n;
 
@@ -1522,7 +1650,7 @@ static void oct_destroy (FttOct * oct,
 
   oct->parent->children = NULL;
   for (n = 0; n < FTT_CELLS; n++)
-    ftt_cell_destroy (&(oct->cell[n]), cleanup);
+    ftt_cell_destroy (&(oct->cell[n]), cleanup, data);
   g_free (oct);
 }
 
@@ -1530,13 +1658,15 @@ static void oct_destroy (FttOct * oct,
  * ftt_cell_destroy:
  * @cell: a #FttCell.
  * @cleanup: a #FttCellCleanupFunc to call before destroying @cell or %NULL.
+ * @data: user data to pass to @cleanup.
  *
  * Frees all memory allocated for @cell and its descendants.
  *
  * The user-defined function @cleanup is called prior to freeing memory.
  */
 void ftt_cell_destroy (FttCell * cell,
-		       FttCellCleanupFunc cleanup)
+		       FttCellCleanupFunc cleanup,
+		       gpointer data)
 {
   FttCellNeighbors neighbor;
   guint i, level;
@@ -1550,12 +1680,12 @@ void ftt_cell_destroy (FttCell * cell,
   level = ftt_cell_level (cell);
 
   if (cleanup)
-    (* cleanup) (cell);
+    (* cleanup) (cell, data);
   cell->flags |= FTT_FLAG_DESTROYED;
 
   /* destroy children */
   if (!FTT_CELL_IS_LEAF (cell))
-    oct_destroy (cell->children, cleanup);
+    oct_destroy (cell->children, cleanup, data);
 
   /* update relationships for neighbors */
   for (i = 0; i < FTT_NEIGHBORS; i++)
@@ -1568,13 +1698,8 @@ void ftt_cell_destroy (FttCell * cell,
 	g_assert (opneighbor == cell);
 	FTT_ROOT_CELL (neighbor.c[i])->neighbors.c[od] = NULL;
       }
-      if (!FTT_CELL_IS_LEAF (neighbor.c[i])) {
-	FttOct * children = neighbor.c[i]->children;
-	FttCell * opneighbor = children->neighbors.c[od];
-	
-	/* g_assert (opneighbor == NULL || opneighbor == cell); */
-	children->neighbors.c[od] = NULL;
-      }
+      if (!FTT_CELL_IS_LEAF (neighbor.c[i]))
+	neighbor.c[i]->children->neighbors.c[od] = NULL;
     }
   
   if (FTT_CELL_IS_ROOT (cell))
@@ -1588,7 +1713,7 @@ void ftt_cell_destroy (FttCell * cell,
       if (!FTT_CELL_IS_DESTROYED (&(parent->cell[i])))
 	empty = FALSE;
     if (empty)
-      oct_destroy (parent, NULL);
+      oct_destroy (parent, NULL, NULL);
   }
 }
 
@@ -1597,6 +1722,7 @@ void ftt_cell_destroy (FttCell * cell,
  * @root: the root cell of a cell tree.
  * @children: a #FttCellChildren.
  * @cleanup: a #FttCellCleanupFunc to call before destroying a cell.
+ # @data: user data to pass to @cleanup.
  *
  * Destroys the root cell of a cell tree but not its children. Each
  * child becomes the root cell of a new cell tree. The new (orphaned)
@@ -1606,7 +1732,8 @@ void ftt_cell_destroy (FttCell * cell,
  */
 void ftt_cell_destroy_root (FttCell * root,
 			    FttCellChildren * children,
-			    FttCellCleanupFunc cleanup)
+			    FttCellCleanupFunc cleanup,
+			    gpointer data)
 {
   guint i;
   FttCellNeighbors neighbor;
@@ -1619,7 +1746,7 @@ void ftt_cell_destroy_root (FttCell * root,
   g_return_if_fail (children != NULL);
 
   if (cleanup)
-    (* cleanup) (root);
+    (* cleanup) (root, data);
   root->flags |= FTT_FLAG_DESTROYED;
 
   ftt_cell_neighbors (root, &neighbor);
@@ -1679,6 +1806,7 @@ void ftt_cell_destroy_root (FttCell * root,
  * @root: the root of the cell tree to flatten.
  * @d: the direction in which to flatten.
  * @cleanup: a #FttCellCleanupFunc to call before destroying a cell.
+ * @data: user data to pass to @cleanup.
  *
  * Recursively destroys all the cells of the tree defined by @root
  * which do not form the boundary in direction @d. The resulting cell
@@ -1688,21 +1816,28 @@ void ftt_cell_destroy_root (FttCell * root,
  */
 void ftt_cell_flatten (FttCell * root,
 		       FttDirection d,
-		       FttCellCleanupFunc cleanup)
+		       FttCellCleanupFunc cleanup,
+		       gpointer data)
 {
   g_return_if_fail (root != NULL);
   g_return_if_fail (d < FTT_NEIGHBORS);
 
+#if FTT_2D3
+  if (d >= FTT_NEIGHBORS_2D)
+    return;
+#endif /* 2D3 */
+
   if (!FTT_CELL_IS_LEAF (root)) {
     struct _FttOct * oct;
     guint i;
-    static gint index[FTT_NEIGHBORS][FTT_CELLS/2] =
-#ifdef    FTT_2D
+#if (FTT_2D || FTT_2D3)
+    static gint index[FTT_NEIGHBORS_2D][FTT_CELLS/2] =
     {{1, 3},
      {0, 2},
      {0, 1},
      {2, 3}};
 #else  /* FTT_3D */
+    static gint index[FTT_NEIGHBORS][FTT_CELLS/2] =
     {{1, 3, 5, 7},
      {0, 2, 4, 6},
      {0, 1, 4, 5},
@@ -1716,12 +1851,12 @@ void ftt_cell_flatten (FttCell * root,
     for (i = 0; i < FTT_CELLS/2; i++) {
       FttCell * c = &(oct->cell[index[od][i]]);
       if (!FTT_CELL_IS_DESTROYED (c))
-	ftt_cell_destroy (c, cleanup);
+	ftt_cell_destroy (c, cleanup, data);
     }
     if (!FTT_CELL_IS_LEAF (root))
       for (i = 0; i < FTT_CELLS/2; i++)
 	if (!FTT_CELL_IS_DESTROYED (&(oct->cell[index[d][i]])))
-	  ftt_cell_flatten (&(oct->cell[index[d][i]]), d, cleanup);
+	  ftt_cell_flatten (&(oct->cell[index[d][i]]), d, cleanup, data);
   }
 }
 
@@ -1844,6 +1979,48 @@ void ftt_cell_write (const FttCell * root,
   }
 }
 
+/**
+ * ftt_cell_write_binary:
+ * @root: a #FttCell.
+ * @max_depth: the maximum depth at which to stop writing (-1 means no limit).
+ * @fp: a file pointer.
+ * @write: a #FttCellWriteFunc function or %NULL.
+ * @data: user data to pass to @write.
+ *
+ * Writes in the file pointed to by @fp a binary representation of the
+ * cell tree starting at @root. If not %NULL, the user-defined
+ * function @write is used to write the extra user data associated
+ * with each cell.  
+ */
+void ftt_cell_write_binary (const FttCell * root,
+			    gint max_depth,
+			    FILE * fp,
+			    FttCellWriteFunc write,
+			    gpointer data)
+{
+  guint flags;
+
+  g_return_if_fail (root != NULL);
+  g_return_if_fail (fp != NULL);
+
+  flags = root->flags;
+  if (FTT_CELL_IS_LEAF (root) || ftt_cell_level (root) == max_depth)
+    flags |= FTT_FLAG_LEAF;
+
+  fwrite (&flags, sizeof (guint), 1, fp);
+  if (write && !FTT_CELL_IS_DESTROYED (root))
+    (* write) (root, fp, data);
+
+  if ((flags & FTT_FLAG_LEAF) == 0) {
+    FttOct * oct;
+    guint i;
+
+    oct = root->children;
+    for (i = 0; i < FTT_CELLS; i++)
+      ftt_cell_write_binary (&(oct->cell[i]), max_depth, fp, write, data);
+  }
+}
+
 #define FTT_CELL_IS_FLAGGED_LEAF(cell) (((cell)->flags & FTT_FLAG_LEAF) != 0)
 
 static gboolean oct_read (FttCell * parent, 
@@ -1895,6 +2072,9 @@ static gboolean oct_read (FttCell * parent,
   oct = g_malloc0 (sizeof (FttOct));
   oct->level = ftt_cell_level (parent);
   oct->parent = parent;
+#ifdef FTT_2D3
+  oct->dz = ftt_cell_dz (parent);
+#endif
   parent->children = oct;
   ftt_cell_pos (parent, &(oct->pos));
   
@@ -1950,6 +2130,105 @@ FttCell * ftt_cell_read (GtsFile * fp,
   return root;
 }
 
+static gboolean oct_read_binary (FttCell * parent, 
+				 GtsFile * fp,
+				 FttCellReadFunc read,
+				 gpointer data);
+
+static gboolean cell_read_binary (FttCell * cell, 
+				  GtsFile * fp,
+				  FttCellReadFunc read,
+				  gpointer data)
+{
+  guint flags;
+
+  if (gts_file_read (fp, &flags, sizeof (guint), 1) != 1) {
+    gts_file_error (fp, "expecting an integer (flags)");
+    return FALSE;
+  }
+  if (FTT_CELL_ID (cell) != (flags & FTT_FLAG_ID)) {
+    gts_file_error (fp, "flags `%d' != (flags & FTT_FLAG_ID) `%d'",
+		    flags, (flags & FTT_FLAG_ID));
+    return FALSE;
+  }
+  cell->flags = flags;
+
+  if (read && !FTT_CELL_IS_DESTROYED (cell))
+    (* read) (cell, fp, data);
+  if (fp->type == GTS_ERROR)
+    return FALSE;
+
+  if (!FTT_CELL_IS_DESTROYED (cell) && !FTT_CELL_IS_FLAGGED_LEAF (cell))
+    return oct_read_binary (cell, fp, read, data);
+
+  cell->flags &= ~FTT_FLAG_LEAF;
+  return TRUE;
+}
+
+static gboolean oct_read_binary (FttCell * parent,
+				 GtsFile * fp,
+				 FttCellReadFunc read,
+				 gpointer data)
+{
+  FttOct * oct;
+  guint n;
+
+  oct = g_malloc0 (sizeof (FttOct));
+  oct->level = ftt_cell_level (parent);
+  oct->parent = parent;
+#ifdef FTT_2D3
+  oct->dz = ftt_cell_dz (parent);
+#endif
+  parent->children = oct;
+  ftt_cell_pos (parent, &(oct->pos));
+  
+  for (n = 0; n < FTT_CELLS; n++) {
+    oct->cell[n].parent = oct;
+    oct->cell[n].flags = n;
+  }
+
+  for (n = 0; n < FTT_CELLS; n++)
+    if (!cell_read_binary (&(oct->cell[n]), fp, read, data))
+      return FALSE;
+  
+  return TRUE;
+}
+
+/**
+ * ftt_cell_read_binary:
+ * @fp: a #GtsFile.
+ * @read: a #FttCellReadFunc function or %NULL.
+ * @data: user data to pass to @read.
+ *
+ * If an error occurs (i.e. corrupted file or file format incorrect),
+ * the @error field of @fp is set. A possibly incomplete tree is then
+ * returned.
+ *
+ * Returns: the root cell of the tree contained in the file pointed to
+ * by @fp. If not %NULL, the user-defined function @read is used to
+ * read the extra user data associated with each cell.  
+ */
+FttCell * ftt_cell_read_binary (GtsFile * fp,
+				FttCellReadFunc read,
+				gpointer data)
+{
+  FttCell * root;
+  guint l, depth;
+
+  g_return_val_if_fail (fp != NULL, NULL);
+
+  root = ftt_cell_new (NULL, NULL);
+  cell_read_binary (root, fp, read, data);
+
+  depth = ftt_cell_depth (root);
+  for (l = 0; l < depth; l++)
+    ftt_cell_traverse (root, FTT_PRE_ORDER, 
+		       FTT_TRAVERSE_LEVEL|FTT_TRAVERSE_NON_LEAFS, l, 
+		       (FttCellTraverseFunc) set_neighbors, NULL);
+
+  return root;
+}
+
 /**
  * ftt_refine_corner:
  * @cell: a #FttCell.
@@ -1969,20 +2248,24 @@ gboolean ftt_refine_corner (const FttCell * cell)
     return FALSE;
 
   ftt_cell_neighbors (cell, &neighbor);
+#if FTT_2D3
+  for (i = 0; i < FTT_NEIGHBORS_2D; i++) {
+#else
   for (i = 0; i < FTT_NEIGHBORS; i++) {
+#endif
     FttCell * n = neighbor.c[i];
 
     if (n && !FTT_CELL_IS_LEAF (n)) {
       FttCellChildren child;
-      guint j;
+      guint j, k;
 
-      ftt_cell_children_direction (n, FTT_OPPOSITE_DIRECTION (i), &child);
-      for (j = 0; j < FTT_CELLS/2; j++) {
+      k = ftt_cell_children_direction (n, FTT_OPPOSITE_DIRECTION (i), &child);
+      for (j = 0; j < k; j++) {
 	FttCell * c = child.c[j];
 
 	if (c) {
-#ifdef FTT_2D
-	  static guint perpendicular[FTT_NEIGHBORS][FTT_CELLS/2] =
+#if (FTT_2D || FTT_2D3)
+	  static guint perpendicular[FTT_NEIGHBORS_2D][FTT_CELLS/2] =
 	  {{2,3},
 	   {2,3},
 	   {1,0},
@@ -2166,17 +2449,7 @@ static void traverse_face_boundary (FttCell * cell, gpointer * datum)
   face.d = *d;
   face.cell = cell;
   face.neighbor = ftt_cell_neighbor (cell, face.d);
-  if (face.neighbor) {
-#if 0
-    /* boundaries are always fine/fine */
-    g_assert (ftt_face_type (&face) == FTT_FINE_FINE);
-    g_assert (!FTT_CELL_IS_LEAF (face.cell) || 
-	      FTT_CELL_IS_LEAF (face.neighbor));
-#endif
-    (* func) (&face, data);
-  }
-  else
-    (* func) (&face, data);
+  (* func) (&face, data);
 }
 
 /**
@@ -2229,7 +2502,7 @@ static gboolean cell_is_in_box (FttCell * cell, GtsBBox * box)
   bb.y1 = p.y - h;
   bb.x2 = p.x + h;
   bb.y2 = p.y + h;
-#ifdef FTT_2D
+#if FTT_2D
   bb.z1 = bb.z2 = 0.;
 #else /* 3D */
   bb.z1 = p.z - h;
@@ -2505,8 +2778,9 @@ void ftt_cell_traverse_box (FttCell * root,
  * ftt_cell_coarsen:
  * @root: a #FttCell root of a cell tree to coarsen.
  * @coarsen: a #FttCellCoarsenFunc.
- * @data: user data to pass to @coarsen.
+ * @coarsen_data: user data to pass to @coarsen.
  * @cleanup: a #FttCellCleanupFunc to call before destroying a cell or %NULL.
+ * @cleanup_data: user data to pass to @cleanup.
  *
  * Coarsens the cell tree defined by @root according to @coarsen.
  *
@@ -2515,23 +2789,25 @@ void ftt_cell_traverse_box (FttCell * root,
  */
 gboolean ftt_cell_coarsen (FttCell * root,
 			   FttCellCoarsenFunc coarsen,
-			   gpointer data,
-			   FttCellCleanupFunc cleanup)
+			   gpointer coarsen_data,
+			   FttCellCleanupFunc cleanup,
+			   gpointer cleanup_data)
 {
-  guint i;
+  guint i, n;
   gboolean coarsenable = TRUE;
   
   g_return_val_if_fail (root != NULL, FALSE);
   g_return_val_if_fail (coarsen != NULL, FALSE);
 
   if (FTT_CELL_IS_LEAF (root))
-    return TRUE;
+    return (* coarsen) (root, coarsen_data);
 
   for (i = 0; i < FTT_CELLS; i++)
     if (!FTT_CELL_IS_DESTROYED (&(root->children->cell[i])))
       coarsenable &= ftt_cell_coarsen (&(root->children->cell[i]), 
-				       coarsen, data, cleanup);
-  if (!coarsenable || !(* coarsen) (root, data))
+				       coarsen, coarsen_data, 
+				       cleanup, cleanup_data);
+  if (!coarsenable || !(* coarsen) (root, coarsen_data))
     return FALSE;
 
   {
@@ -2540,22 +2816,22 @@ gboolean ftt_cell_coarsen (FttCell * root,
     for (d = 0; d < FTT_NEIGHBORS; d++) {
       FttCellChildren child;
 
-      ftt_cell_children_direction (root, d, &child);
-      for (i = 0; i < FTT_CELLS/2; i++) {
+      n = ftt_cell_children_direction (root, d, &child);
+      for (i = 0; i < n; i++) {
 	FttCell * neighbor;
 
 	if (child.c[i] && (neighbor = ftt_cell_neighbor (child.c[i], d)) &&
 	    !FTT_CELL_IS_LEAF (neighbor)) {
 	  FttCellChildren child1;
-	  guint j;
+	  guint j, k;
 	  gboolean empty = TRUE;
 
-	  ftt_cell_children_direction (neighbor, FTT_OPPOSITE_DIRECTION (d), 
-				       &child1);
-	  for (j = 0; j < FTT_CELLS/2 && empty; j++)
+	  k = ftt_cell_children_direction (neighbor, FTT_OPPOSITE_DIRECTION (d), &child1);
+	  for (j = 0; j < k && empty; j++)
 	    if (child1.c[j])
 	      empty = FALSE;
-	  if (!empty && !ftt_cell_coarsen (neighbor, coarsen, data, cleanup))
+	  if (!empty && !ftt_cell_coarsen (neighbor, coarsen, coarsen_data, 
+					   cleanup, cleanup_data))
 	    return FALSE;
 	}
       }
@@ -2565,7 +2841,7 @@ gboolean ftt_cell_coarsen (FttCell * root,
   if (cleanup)
     for (i = 0; i < FTT_CELLS; i++)
       if (!FTT_CELL_IS_DESTROYED (&(root->children->cell[i])))
-	(* cleanup) (&(root->children->cell[i]));
+	(* cleanup) (&(root->children->cell[i]), cleanup_data);
   g_free (root->children);
   root->children = NULL;
 
@@ -2669,3 +2945,4 @@ void ftt_cell_traverse_destroy (FttCellTraverse * t)
   g_free (t->cells);
   g_free (t);
 }
+
