@@ -33,7 +33,7 @@ extern "C" {
 
 #define FTT_MAINTAINER "s.popinet@niwa.cri.nz"
 
-#ifdef FTT_2D
+#if (FTT_2D || FTT_2D3)
 # define FTT_CELLS     4
 #else  /* FTT_3D */
 # define FTT_CELLS     8
@@ -71,24 +71,37 @@ typedef enum
   FTT_LEFT,
   FTT_TOP,
   FTT_BOTTOM,
-#ifndef FTT_2D
+#if (!FTT_2D)
   FTT_FRONT,
   FTT_BACK,
-#endif /* FTT_3D */  
+#endif /* FTT_3D || FTT_2D3 */
   FTT_NEIGHBORS
 } FttDirection;
 
-gchar * ftt_direction_name[FTT_NEIGHBORS]; /* defined in ftt.c */
+#define FTT_NEIGHBORS_2D (FTT_BOTTOM + 1)
+
+#if FTT_2D3
+# define FTT_CELLS_DIRECTION(d) ((d) < FTT_NEIGHBORS_2D ? FTT_CELLS/2 : FTT_CELLS)
+#else  /* 2D && 3D */
+# define FTT_CELLS_DIRECTION(d) (FTT_CELLS/2)
+#endif /* 2D && 3D */
+
+GTS_C_VAR gchar * ftt_direction_name[FTT_NEIGHBORS]; /* defined in ftt.c */
 
 typedef enum
 {
   FTT_X = 0,
   FTT_Y,
-#ifndef FTT_2D
+#if (!FTT_2D)
   FTT_Z,
-#endif /* FTT_3D */  
+#endif /* FTT_3D || FTT_2D3 */
   FTT_DIMENSION,
+  FTT_XY,
+#if FTT_2D
+  FTT_XYZ = FTT_XY
+#else  /* FTT_3D || FTT_2D3 */
   FTT_XYZ
+#endif /* FTT_3D || FTT_2D3 */
 } FttComponent;
 
 typedef enum {
@@ -128,6 +141,9 @@ struct _FttRootCell {
   FttCellNeighbors neighbors;
   FttVector pos;
   guint level;
+#if FTT_2D3
+  gdouble dz;
+#endif
 };
 
 struct _FttOct {
@@ -135,6 +151,9 @@ struct _FttOct {
   FttCell * parent;
   FttCellNeighbors neighbors;
   FttVector pos;
+#if FTT_2D3
+  gdouble dz;
+#endif
 
   FttCell cell[FTT_CELLS];
 };
@@ -161,12 +180,41 @@ typedef enum {
     (dst)->neighbor = (src)->cell,\
     (dst)->d = FTT_OPPOSITE_DIRECTION((src)->d))
 
-GUTILS_C_VAR
+GTS_C_VAR
 gint                 ftt_opposite_direction[FTT_NEIGHBORS];
 
 #define FTT_OPPOSITE_DIRECTION(d)     (ftt_opposite_direction[d])
 #define FTT_ORTHOGONAL_COMPONENT(c)   (((c) + 1) % FTT_DIMENSION)
 
+#ifdef G_DISABLE_ASSERT
+
+#define g_assert_not_implemented()
+
+#else /* !G_DISABLE_ASSERT */
+
+#ifdef __GNUC__
+
+#define g_assert_not_implemented()      G_STMT_START{		\
+     g_log (G_LOG_DOMAIN,					\
+	    G_LOG_LEVEL_ERROR,					\
+	    "file %s: line %d (%s): not implemented (yet)",	\
+	    __FILE__,						\
+	    __LINE__,						\
+	    __PRETTY_FUNCTION__);	}G_STMT_END
+
+#else /* !__GNUC__ */
+
+#define g_assert_not_implemented()	G_STMT_START{	\
+     g_log (G_LOG_DOMAIN,				\
+	    G_LOG_LEVEL_ERROR,				\
+	    "file %s: line %d: not implemented (yet)",	\
+	    __FILE__,					\
+	    __LINE__);		}G_STMT_END
+
+#endif /* __GNUC__ */
+
+#endif /* !G_DISABLE_ASSERT */
+
 FttCell *            ftt_cell_new                    (FttCellInitFunc init,
 						      gpointer data);
 #define              ftt_cell_level(c)  ((c)->parent ?\
@@ -174,6 +222,13 @@ FttCell *            ftt_cell_new                    (FttCellInitFunc init,
                                          ((struct _FttRootCell *) c)->level)
 #define              ftt_cell_parent(c) ((c)->parent ?\
                                          (c)->parent->parent : NULL)
+#ifdef FTT_2D3
+# define             ftt_cell_dz(c)     ((c)->parent ?\
+                                         (c)->parent->dz :\
+                                         ((struct _FttRootCell *) c)->dz)
+#else  /* 2D or 3D */
+# define             ftt_cell_dz(c)     (1.)
+#endif /* 2D or 3D */
 guint                ftt_cell_depth                  (const FttCell * root);
 G_INLINE_FUNC
 gdouble              ftt_level_size                  (guint level);
@@ -185,7 +240,7 @@ G_INLINE_FUNC
 void                 ftt_cell_children           (const FttCell * cell,
 						  FttCellChildren * children);
 G_INLINE_FUNC
-void                 ftt_cell_children_direction (const FttCell * cell,
+guint                ftt_cell_children_direction (const FttCell * cell,
 						  FttDirection d,
 						  FttCellChildren * children);
 G_INLINE_FUNC
@@ -210,6 +265,10 @@ FttCellFace          ftt_cell_face              (FttCell * cell,
 G_INLINE_FUNC
 FttFaceType          ftt_face_type              (const FttCellFace * face);
 
+G_INLINE_FUNC
+gboolean             ftt_cell_neighbor_is_brother (FttCell * cell, 
+						   FttDirection d);
+
 #ifdef G_CAN_INLINE
 G_INLINE_FUNC
 gdouble ftt_level_size (guint level)
@@ -240,7 +299,7 @@ gdouble ftt_cell_volume (const FttCell * cell)
   g_return_val_if_fail (cell != NULL, 0.);
 
   size = ftt_level_size (ftt_cell_level (cell));
-#ifdef FTT_2D
+#if (FTT_2D || FTT_2D3)
   return size*size;
 #else  /* FTT_3D */
   return size*size*size;
@@ -260,24 +319,24 @@ void ftt_cell_children (const FttCell * cell,
 
   oct = cell->children;
   for (i = 0; i < FTT_CELLS; i++)
-    children->c[i] = FTT_CELL_IS_DESTROYED (&(oct->cell[i])) ? 
-      NULL : &(oct->cell[i]);
+    children->c[i] = FTT_CELL_IS_DESTROYED (&(oct->cell[i])) ? NULL : &(oct->cell[i]);
 }
 
 G_INLINE_FUNC
-void ftt_cell_children_direction (const FttCell * cell,
+guint ftt_cell_children_direction (const FttCell * cell,
 				  FttDirection d,
 				  FttCellChildren * children)
 {
   struct _FttOct * oct;
   guint i;
-  static gint index[FTT_NEIGHBORS][FTT_CELLS/2] =
-#ifdef    FTT_2D
+#if (FTT_2D || FTT_2D3)
+  static gint index[FTT_NEIGHBORS_2D][FTT_CELLS/2] =
   {{1, 3},
    {0, 2},
    {0, 1},
    {2, 3}};
 #else  /* FTT_3D */
+  static gint index[FTT_NEIGHBORS][FTT_CELLS/2] =
   {{1, 3, 5, 7},
    {0, 2, 4, 6},
    {0, 1, 4, 5},
@@ -286,23 +345,33 @@ void ftt_cell_children_direction (const FttCell * cell,
    {4, 5, 6, 7}};
 #endif /* FTT_3D */
 
-  g_return_if_fail (cell != NULL);
-  g_return_if_fail (!FTT_CELL_IS_LEAF (cell));
-  g_return_if_fail (d < FTT_NEIGHBORS);
-  g_return_if_fail (children != NULL);
+  g_return_val_if_fail (cell != NULL, 0);
+  g_return_val_if_fail (!FTT_CELL_IS_LEAF (cell), 0);
+  g_return_val_if_fail (d < FTT_NEIGHBORS, 0);
+  g_return_val_if_fail (children != NULL, 0);
 
   oct = cell->children;
+
+#if FTT_2D3
+  if (d >= FTT_NEIGHBORS_2D) {
+    for (i = 0; i < FTT_CELLS; i++)
+      children->c[i] = FTT_CELL_IS_DESTROYED (&(oct->cell[i])) ? NULL : &(oct->cell[i]);
+    return FTT_CELLS;
+  }
+#endif /* 2D3 */
+
   for (i = 0; i < FTT_CELLS/2; i++)
     children->c[i] = FTT_CELL_IS_DESTROYED (&(oct->cell[index[d][i]])) ? 
       NULL : &(oct->cell[index[d][i]]);
+  return FTT_CELLS/2;
 }
 
 G_INLINE_FUNC
 FttCell * ftt_cell_child_corner (const FttCell * cell,
 				 FttDirection d[FTT_DIMENSION])
 {
-#ifdef FTT_2D  
-  static gint index[FTT_NEIGHBORS][FTT_NEIGHBORS] = {
+#if (FTT_2D || FTT_2D3)
+  static gint index[FTT_NEIGHBORS_2D][FTT_NEIGHBORS_2D] = {
     {-1,-1,1,3},
     {-1,-1,0,2},
     {1,0,-1,-1},
@@ -312,8 +381,16 @@ FttCell * ftt_cell_child_corner (const FttCell * cell,
 
   g_return_val_if_fail (cell != NULL, NULL);
   g_return_val_if_fail (!FTT_CELL_IS_LEAF (cell), NULL);
+
+  g_assert (d[0] < FTT_NEIGHBORS);
+  g_assert (d[1] < FTT_NEIGHBORS);
+
   g_return_val_if_fail (d[0] < FTT_NEIGHBORS, NULL);
   g_return_val_if_fail (d[1] < FTT_NEIGHBORS, NULL);
+#  if FTT_2D3
+  if (d[0] >= FTT_NEIGHBORS_2D || d[1] >= FTT_NEIGHBORS_2D)
+    g_assert_not_implemented ();
+#  endif
 
   i = index[d[0]][d[1]];
 #else  /* FTT_3D */
@@ -350,7 +427,8 @@ FttCell * ftt_cell_child_corner (const FttCell * cell,
 
   g_return_val_if_fail (i >= 0, NULL);
 
-  return &(cell->children->cell[i]);
+  return FTT_CELL_IS_DESTROYED (&(cell->children->cell[i])) ? NULL:
+    &(cell->children->cell[i]);
 }
 
 G_INLINE_FUNC
@@ -358,11 +436,18 @@ void ftt_cell_neighbors_not_cached (const FttCell * cell,
 				    FttCellNeighbors * neighbors)
 {
   static gint neighbor_index[FTT_NEIGHBORS][FTT_CELLS]
-#ifdef FTT_2D
+#if FTT_2D
     = {{1,-1,3,-3},
        {-2,0,-4,2},
        {-3,-4,0,1},
        {2,3,-1,-2}};
+#elif FTT_2D3
+    = {{1,-1,3,-3},
+       {-2,0,-4,2},
+       {-3,-4,0,1},
+       {2,3,-1,-2},
+       {-1,-2,-3,-4},
+       {-1,-2,-3,-4}};
 #else  /* FTT_3D */
     = {{1,-1,3,-3,5,-5,7,-7},
        {-2,0,-4,2,-6,4,-8,6},
@@ -408,11 +493,18 @@ FttCell * ftt_cell_neighbor_not_cached (const FttCell * cell,
 					FttDirection d)
 {
   static gint neighbor_index[FTT_NEIGHBORS][FTT_CELLS]
-#ifdef FTT_2D
+#if FTT_2D
     = {{1,-1,3,-3},
        {-2,0,-4,2},
        {-3,-4,0,1},
        {2,3,-1,-2}};
+#elif FTT_2D3
+    = {{1,-1,3,-3},
+       {-2,0,-4,2},
+       {-3,-4,0,1},
+       {2,3,-1,-2},
+       {-1,-2,-3,-4},
+       {-1,-2,-3,-4}};
 #else  /* FTT_3D */
     = {{1,-1,3,-3,5,-5,7,-7},
        {-2,0,-4,2,-6,4,-8,6},
@@ -500,6 +592,28 @@ FttFaceType ftt_face_type (const FttCellFace * face)
   g_assert (ftt_cell_level (face->cell) == ftt_cell_level (face->neighbor));
   return FTT_FINE_FINE;
 }
+
+G_INLINE_FUNC
+gboolean ftt_cell_neighbor_is_brother (FttCell * cell, 
+				       FttDirection d)
+{
+  static gboolean b[FTT_CELLS][FTT_NEIGHBORS] = {
+#if FTT_2D
+    {1,0,0,1}, {0,1,0,1}, {1,0,1,0}, {0,1,1,0}
+#elif FTT_2D3
+    {1,0,0,1,0,0}, {0,1,0,1,0,0}, {1,0,1,0,0,0}, {0,1,1,0,0,0}
+#else  /* 3D */
+    {1,0,0,1,0,1}, {0,1,0,1,0,1}, {1,0,1,0,0,1}, {0,1,1,0,0,1},
+    {1,0,0,1,1,0}, {0,1,0,1,1,0}, {1,0,1,0,1,0}, {0,1,1,0,1,0}
+#endif /* 3D */
+  };
+
+  g_return_val_if_fail (cell != NULL, FALSE);
+  
+  if (FTT_CELL_IS_ROOT (cell))
+    return FALSE;
+  return b[FTT_CELL_ID (cell)][d];
+}
 #endif	/* G_CAN_INLINE */
 
 void                 ftt_cell_set_neighbor           (FttCell * root,
@@ -516,6 +630,12 @@ void                 ftt_cell_relative_pos           (const FttCell * cell,
 						      FttVector * pos);
 void                 ftt_cell_pos                    (const FttCell * cell,
 						      FttVector * pos);
+void                 ftt_corner_relative_pos         (const FttCell * cell,
+						      FttDirection d[FTT_DIMENSION],
+						      FttVector * pos);
+void                 ftt_corner_pos                  (const FttCell * cell,
+						      FttDirection d[FTT_DIMENSION],
+						      FttVector * pos);
 void                 ftt_face_pos                    (const FttCellFace * face,
 						      FttVector * pos);
 void                 ftt_cell_set_pos                (FttCell * root,
@@ -591,27 +711,40 @@ void                 ftt_cell_write                  (const FttCell * root,
 						      FILE * fp,
 						      FttCellWriteFunc write,
 						      gpointer data);
+void                 ftt_cell_write_binary           (const FttCell * root,
+						      gint max_depth,
+						      FILE * fp,
+						      FttCellWriteFunc write,
+						      gpointer data);
 typedef void      (* FttCellReadFunc)                (FttCell * cell,
 						      GtsFile * fp,
 						      gpointer data);
 FttCell *            ftt_cell_read                   (GtsFile * fp,
 						      FttCellReadFunc read,
 						      gpointer data);
-typedef void      (* FttCellCleanupFunc)             (FttCell * cell);
+FttCell *            ftt_cell_read_binary            (GtsFile * fp,
+						      FttCellReadFunc read,
+						      gpointer data);
+typedef void      (* FttCellCleanupFunc)             (FttCell * cell,
+						      gpointer data);
 void                 ftt_cell_destroy           (FttCell * cell,
-						 FttCellCleanupFunc cleanup);
+						 FttCellCleanupFunc cleanup,
+						 gpointer data);
 void                 ftt_cell_destroy_root      (FttCell * root,
 						 FttCellChildren * children,
-						 FttCellCleanupFunc cleanup);
+						 FttCellCleanupFunc cleanup,
+						 gpointer data);
 void                 ftt_cell_flatten           (FttCell * root, 
 						 FttDirection d, 
-						 FttCellCleanupFunc cleanup);
+						 FttCellCleanupFunc cleanup,
+						 gpointer data);
 typedef gboolean  (* FttCellCoarsenFunc)        (FttCell * cell,
 						 gpointer data);
 gboolean             ftt_cell_coarsen           (FttCell * root,
 						 FttCellCoarsenFunc coarsen,
-						 gpointer data,
-						 FttCellCleanupFunc cleanup);
+						 gpointer coarsen_data,
+						 FttCellCleanupFunc cleanup,
+						 gpointer cleanup_data);
 FttDirection         ftt_direction_from_name    (const gchar * name);
 
 struct _FttCellTraverse {
