@@ -23,6 +23,7 @@
 
 #include "config.h"
 #include "graphic.h"
+#include "solid.h"
 
 typedef struct {
   GPtrArray * colors;
@@ -115,40 +116,135 @@ static GtsColor colormap_color (Colormap * cmap, gdouble val)
   return c;
 }
 
+/* VertexCellFace: Header */
+
+typedef struct _VertexCellFace         VertexCellFace;
+
+struct _VertexCellFace {
+  /*< private >*/
+  GtsVertex parent;
+
+  /*< public >*/
+  guint index;
+  FttCell * cell;
+  FttCellFace face;
+};
+
+#define VERTEX_CELL_FACE(obj)            GTS_OBJECT_CAST (obj,\
+					         VertexCellFace,\
+					         vertex_cell_face_class ())
+
+/* VertexCellFace: Object */
+
+static GtsVertexClass * vertex_cell_face_class (void)
+{
+  static GtsVertexClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo vertex_cell_face_info = {
+      "VertexCellFace",
+      sizeof (VertexCellFace),
+      sizeof (GtsVertexClass),
+      (GtsObjectClassInitFunc) NULL,
+      (GtsObjectInitFunc) NULL,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gts_vertex_class ()),
+				  &vertex_cell_face_info);
+  }
+
+  return klass;
+}
+
 static void triangulate (FttCell * cell, gpointer * data)
 {
   GtsSurface * s = data[0];
   GfsVariable * var = data[1];
-  GtsMatrix * transform = data[2];
-  gdouble * z = data[3];
-  GfsNorm * norm = data[4];
   GtsVertex * v;
   FttVector pos;
 
-  ftt_cell_pos (cell, &pos);
+  if (var && var->centered)
+    ftt_cell_pos (cell, &pos);
+  else 
+    gfs_cell_cm (cell, &pos);
+
   v = gts_vertex_new (s->vertex_class, pos.x, pos.y, pos.z);
-  gts_point_transform (GTS_POINT (v), transform);
-  GTS_POINT (v)->z = *z + GFS_VARIABLE (cell, var->i)/(norm->infty*1000.);
+  if (var) {
+    GtsMatrix * transform = data[2];
+    gdouble * z = data[3];
+    GfsNorm * norm = data[4];
+
+    gts_point_transform (GTS_POINT (v), transform);
+    GTS_POINT (v)->z = *z + GFS_VARIABLE (cell, var->i)/(norm->infty*1000.);
+  }
   g_assert (gts_delaunay_add_vertex (s, v, NULL) == NULL);
+  VERTEX_CELL_FACE (v)->cell = cell;
+}
+
+static void triangulate_face (FttCell * cell, gpointer * data)
+{
+  GtsSurface * s = data[0];
+  GfsVariable * var = data[1];
+  FttDirection * d = data[5];
+  GtsVertex * v;
+  FttVector pos;
+  FttCellFace face;
+
+  face.cell = cell;
+  face.d = *d;
+  face.neighbor = ftt_cell_neighbor (cell, face.d);
+  ftt_face_pos (&face, &pos);
+  v = gts_vertex_new (s->vertex_class, pos.x, pos.y, pos.z);
+  if (var) {
+    GtsMatrix * transform = data[2];
+    gdouble * z = data[3];
+    GfsNorm * norm = data[4];
+
+    gts_point_transform (GTS_POINT (v), transform);
+    if (face.neighbor)
+      GTS_POINT (v)->z = *z + 
+	gfs_face_interpolated_value  (&face, var->i)/(norm->infty*1000.);
+    else
+      GTS_POINT (v)->z = *z + GFS_VARIABLE (cell, var->i)/(norm->infty*1000.);
+  }
+  g_assert (gts_delaunay_add_vertex (s, v, NULL) == NULL);
+  VERTEX_CELL_FACE (v)->face = face;
+}
+
+static void add_long_segment (GtsSegment * s, GSList ** list)
+{
+  FttCell * c1 = VERTEX_CELL_FACE (s->v1)->cell ? VERTEX_CELL_FACE (s->v1)->cell :
+    VERTEX_CELL_FACE (s->v1)->face.cell;
+  FttCell * c2 = VERTEX_CELL_FACE (s->v2)->cell ? VERTEX_CELL_FACE (s->v2)->cell :
+    VERTEX_CELL_FACE (s->v2)->face.cell;
+  gdouble s1 = ftt_cell_size (c1);
+  gdouble s2 = ftt_cell_size (c2);
+  gdouble size = MIN (s1, s2);
+
+  if (gts_point_distance2 (GTS_POINT (s->v1), GTS_POINT (s->v2)) > 
+      16.*size*size)
+    *list = g_slist_prepend (*list, s);
 }
 
 void gfs_write_gts (GfsDomain * domain, 
-		   GfsVariable * v, 
-		   FttTraverseFlags flags,
-		   gint level,
-		   GtsBBox * box,
-		   FILE * fp)
+		    GfsVariable * v, 
+		    FttTraverseFlags flags,
+		    gint level,
+		    GtsBBox * box,
+		    FILE * fp)
 {
   GtsSurface * s;
   GtsVertex * v1, * v2, * v3;
   GtsEdge * e1, * e2, * e3;
-  gpointer data[5];
+  gpointer data[6];
   GtsMatrix * transform, * inv;
   gdouble z = 0.;
   GfsNorm norm;
+  GSList * long_segments = NULL;
 
   g_return_if_fail (domain != NULL);
-#ifndef FTT_2D
+#if (!FTT_2D)
   g_return_if_fail (box != NULL);
 #endif /* 3D */
   g_return_if_fail (fp != NULL);
@@ -162,36 +258,37 @@ void gfs_write_gts (GfsDomain * domain,
   s = gts_surface_new (gts_surface_class (), 
 		       gts_face_class (), 
 		       gts_edge_class (), 
-		       gts_vertex_class ());
+		       vertex_cell_face_class ());
   gts_surface_add_face (s, gts_face_new (gts_face_class (), e1, e2, e3));
 
   norm = gfs_domain_norm_variable (domain, v, flags, level);
   if (norm.infty == 0.)
     norm.infty = 1.;
-#ifdef FTT_2D
-  transform = gts_matrix_new (1., 0., 0.,
-			      0., 1., 0.,
-			      0., 0., 1.);
+#if FTT_2D
+  transform = gts_matrix_identity (NULL);
 #else /* 3D */
   if (box->x2 - box->x1 < box->z2 - box->z1 &&
       box->x2 - box->x1 < box->y2 - box->y1) {
     z = box->x2 = box->x1 = (box->x2 + box->x1)/2. + 1e-30;
-    transform = gts_matrix_new (0., 1., 0.,
-				0., 0., 1.,
-				1., 0., 0.);
+    transform = gts_matrix_new (0., 1., 0., 0.,
+				0., 0., 1., 0.,
+				1., 0., 0., 0.,
+				0., 0., 0., 0.);
   }
   else if (box->y2 - box->y1 < box->z2 - box->z1 &&
 	   box->y2 - box->y1 < box->x2 - box->x1) {
     z = box->y2 = box->y1 = (box->y2 + box->y1)/2. + 1e-30;
-    transform = gts_matrix_new (0., 0., 1.,
-				1., 0., 0.,
-				0., 1., 0.);
+    transform = gts_matrix_new (0., 0., 1., 0.,
+				1., 0., 0., 0.,
+				0., 1., 0., 0.,
+				0., 0., 0., 0.);
   }
   else {
     z = box->z2 = box->z1 = (box->z2 + box->z1)/2. + 1e-30;
-    transform = gts_matrix_new (1., 0., 0.,
-				0., 1., 0.,
-				0., 0., 1.);
+    transform = gts_matrix_new (1., 0., 0., 0.,
+				0., 1., 0., 0.,
+				0., 0., 1., 0.,
+				0., 0., 0., 0.);
   }
 #endif /* 3D */
   
@@ -200,9 +297,17 @@ void gfs_write_gts (GfsDomain * domain,
   data[2] = transform;
   data[3] = &z;
   data[4] = &norm;
-  if (box == NULL)
+  if (box == NULL) {
+    FttDirection d;
+
     gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, flags, level, 
 			     (FttCellTraverseFunc) triangulate, data);
+    data[5] = &d;
+    for (d = 0; d < FTT_NEIGHBORS; d++)
+      gfs_domain_cell_traverse_boundary (domain, d, 
+					 FTT_PRE_ORDER, flags, level, 
+		   (FttCellTraverseFunc) triangulate_face, data);
+  }
   else
     gfs_domain_cell_traverse_box (domain, box, FTT_PRE_ORDER, flags, level, 
 				 (FttCellTraverseFunc) triangulate, data);
@@ -213,7 +318,13 @@ void gfs_write_gts (GfsDomain * domain,
   gts_object_destroy (GTS_OBJECT (v3));
   gts_allow_floating_vertices = FALSE;
 
-  inv = gts_matrix_inverse (transform);
+  gts_surface_foreach_edge (s, (GtsFunc) add_long_segment, &long_segments);
+  gts_allow_floating_edges = TRUE;
+  g_slist_foreach (long_segments, (GFunc) gts_object_destroy, NULL);
+  gts_allow_floating_edges = FALSE;
+  g_slist_free (long_segments);
+
+  inv = gts_matrix3_inverse (transform);
   gts_matrix_destroy (transform);
   transform = inv;
   gts_surface_foreach_vertex (s, (GtsFunc) gts_point_transform, transform);
@@ -223,7 +334,251 @@ void gfs_write_gts (GfsDomain * domain,
   gts_matrix_destroy (transform);
 }
 
-#ifdef FTT_2D
+static void write_vertex (GtsPoint * p, gpointer * data)
+{
+  guint * index = data[0];
+  FILE * fp = data[1];
+
+  fprintf (fp, "%g %g %g\n", p->x, p->y, p->z);
+  VERTEX_CELL_FACE (p)->index = (*index)++;
+}
+
+static void write_face (GtsTriangle * t, FILE * fp)
+{
+  GtsVertex * v1, * v2, * v3;
+
+  gts_triangle_vertices (t, &v1, &v2, &v3);
+  fprintf (fp, "3 %u %u %u\n",
+	   VERTEX_CELL_FACE (v1)->index, 
+	   VERTEX_CELL_FACE (v2)->index,
+	   VERTEX_CELL_FACE (v3)->index);
+}
+
+static gdouble variable_value (VertexCellFace * p, GfsVariable * v)
+{
+  if (p->cell)
+    return GFS_VARIABLE (p->cell, v->i);
+  else if (p->face.neighbor)
+    return gfs_face_interpolated_value (&p->face, v->i);
+  else
+    return GFS_VARIABLE (p->face.cell, v->i);
+}
+
+static void write_variable (VertexCellFace * p, gpointer * data)
+{
+  fprintf (data[1], "%g\n", variable_value (p, data[0]));
+}
+
+static void write_vector (VertexCellFace * p, gpointer * data)
+{
+  GfsVariable * v = data[0];
+
+#if FTT_2D
+  fprintf (data[1], "%g %g 0\n", 
+	   variable_value (p, v), 
+	   variable_value (p, v->next));
+#else  /* 3D */
+  fprintf (data[1], "%g %g %g\n",
+	   variable_value (p, v), 
+	   variable_value (p, v->next),
+	   variable_value (p, v->next->next));
+#endif /* 3D */
+}
+
+void gfs_write_vtk (GfsDomain * domain,
+		    FttTraverseFlags flags,
+		    gint level,
+		    FILE * fp)
+{
+  GtsSurface * s;
+  GtsVertex * v1, * v2, * v3;
+  GtsEdge * e1, * e2, * e3;
+  gpointer data[6] = { NULL, NULL, NULL, NULL, NULL };
+  guint nvertices, nfaces, index = 0, i;
+  GfsVariable * v;
+  GSList * long_segments = NULL;
+  FttDirection d;
+
+  g_return_if_fail (domain != NULL);
+  g_return_if_fail (fp != NULL);
+
+  v1 = gts_vertex_new (gts_vertex_class (), -100., -100., 0.);
+  v2 = gts_vertex_new (gts_vertex_class (), 100., -100., 0.);
+  v3 = gts_vertex_new (gts_vertex_class (), 0., 100., 0.);
+  e1 = gts_edge_new (gts_edge_class (), v1, v2);
+  e2 = gts_edge_new (gts_edge_class (), v2, v3);
+  e3 = gts_edge_new (gts_edge_class (), v3, v1);
+  s = gts_surface_new (gts_surface_class (), 
+		       gts_face_class (), 
+		       gts_edge_class (), 
+		       vertex_cell_face_class ());
+  gts_surface_add_face (s, gts_face_new (gts_face_class (), e1, e2, e3));
+
+  data[0] = s;
+  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, flags, level, 
+			    (FttCellTraverseFunc) triangulate, data);
+  data[5] = &d;
+  for (d = 0; d < FTT_NEIGHBORS; d++)
+    gfs_domain_cell_traverse_boundary (domain, d, FTT_PRE_ORDER, flags, level, 
+				       (FttCellTraverseFunc) triangulate_face,
+				       data);
+
+  gts_allow_floating_vertices = TRUE;
+  gts_object_destroy (GTS_OBJECT (v1));
+  gts_object_destroy (GTS_OBJECT (v2));
+  gts_object_destroy (GTS_OBJECT (v3));
+  gts_allow_floating_vertices = FALSE;
+
+  gts_surface_foreach_edge (s, (GtsFunc) add_long_segment, &long_segments);
+  gts_allow_floating_edges = TRUE;
+  g_slist_foreach (long_segments, (GFunc) gts_object_destroy, NULL);
+  gts_allow_floating_edges = FALSE;
+  g_slist_free (long_segments);
+
+  nvertices = gts_surface_vertex_number (s);
+  fprintf (fp, 
+	   "# vtk DataFile Version 2.0\n"
+	   "Generated by Gerris\n"
+           "ASCII\n"
+	   "DATASET UNSTRUCTURED_GRID\n"
+	   "POINTS %u float\n",
+	   nvertices);
+  data[0] = &index;
+  data[1] = fp;
+  gts_surface_foreach_vertex (s, (GtsFunc) write_vertex, data);
+  nfaces = gts_surface_face_number (s);
+  fprintf (fp, "CELLS %u %u\n", nfaces, 4*nfaces);
+  gts_surface_foreach_face (s, (GtsFunc) write_face, fp);
+  fprintf (fp, "CELL_TYPES %u\n", nfaces);
+  for (i = 0; i < nfaces; i++)
+    fputs ("5\n", fp);
+  fprintf (fp, "POINT_DATA %u\n", nvertices);
+  v = domain->variables;
+  while (v) {
+    guint d;
+
+    if (!strcmp (v->name, "U") && v->next &&
+	!strcmp (v->next->name, "V")
+#if (!FTT_2D)
+	&& v->next->next && !strcmp (v->next->next->name, "W")
+#endif /* not FTT_2D */
+	) {
+      fputs ("VECTORS U float\n", fp);
+      data[0] = v;
+      data[1] = fp;
+      gts_surface_foreach_vertex (s, (GtsFunc) write_vector, data);
+      for (d = 0; d < FTT_DIMENSION; d++) v = v->next;
+    }
+    else if (v->name[strlen (v->name) - 1] == 'x' && v->next &&
+	     v->next->name[strlen (v->name) - 1] == 'y'
+#if (!FTT_2D)
+	     && v->next->next &&
+	     v->next->next->name[strlen (v->name) - 1] == 'z'
+#endif /* not FTT_2D */
+	     ) {
+      gchar * name = g_strndup (v->name, strlen (v->name) - 1);
+
+      fprintf (fp, "VECTORS %s float\n", name);
+      g_free (name);
+      data[0] = v;
+      data[1] = fp;
+      gts_surface_foreach_vertex (s, (GtsFunc) write_vector, data);
+      for (d = 0; d < FTT_DIMENSION; d++) v = v->next;
+    }
+    else {
+      fprintf (fp, 
+	       "SCALARS %s float 1\n"
+	       "LOOKUP_TABLE default\n",
+	       v->name);
+      data[0] = v;
+      data[1] = fp;
+      gts_surface_foreach_vertex (s, (GtsFunc) write_variable, data);
+      v = v->next;
+    }
+  }
+	 
+  gts_object_destroy (GTS_OBJECT (s));
+}
+
+static void extent (FttCell * cell, gpointer * data)
+{
+  FttVector * min = data[0];
+  FttVector * max = data[1];
+  FttVector pos;
+  
+  ftt_cell_pos (cell, &pos);
+  if (pos.x > max->x) max->x = pos.x;
+  if (pos.y > max->y) max->y = pos.y;
+  if (pos.z > max->z) max->z = pos.z;
+  if (pos.x < min->x) min->x = pos.x;
+  if (pos.y < min->y) min->y = pos.y;
+  if (pos.z < min->z) min->z = pos.z;
+}
+
+static void iso_func (gdouble ** a, GtsCartesianGrid g, guint k,
+		      gpointer * data)
+{
+  GfsDomain * domain = data[0];
+  guint * level = data[1], i, j;
+  GfsVariable * v = data[2];
+  FttVector p;
+  fprintf (stderr, "\rslice %4d/%d", k + 1, g.nz);
+  p.z = g.z;
+  for (i = 0, p.x = g.x; i < g.nx; i++, p.x += g.dx)
+    for (j = 0, p.y = g.y; j < g.ny; j++, p.y += g.dy) {
+      FttCell * cell = gfs_domain_locate (domain, p, *level);
+
+      if (cell == NULL)
+	a[i][j] = 0.;
+      else
+	a[i][j] = gfs_interpolate (cell, p, v);
+    }
+}
+
+GtsSurface * gfs_isosurface (GfsDomain * domain, 
+			     GfsVariable * v, gdouble val,
+			     gint level)
+{
+  FttVector cmax = { - G_MAXDOUBLE, - G_MAXDOUBLE, - G_MAXDOUBLE };
+  FttVector cmin = { G_MAXDOUBLE, G_MAXDOUBLE, G_MAXDOUBLE };
+  guint depth;
+  GtsCartesianGrid g;
+  gpointer data[3];
+  GtsSurface * s;
+
+  g_return_val_if_fail (domain != NULL, NULL);
+  g_return_val_if_fail (v != NULL, NULL);
+
+  if (level < 0)
+    depth = gfs_domain_depth (domain);
+  else
+    depth = level;
+
+  data[0] = &cmin;
+  data[1] = &cmax;
+  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, 
+			    FTT_TRAVERSE_LEAFS|FTT_TRAVERSE_LEVEL, depth,
+			    (FttCellTraverseFunc) extent, data);
+  if (cmin.x == G_MAXDOUBLE)
+    return NULL;
+
+  g.dx = g.dy = g.dz = ftt_level_size (depth);
+  g.x = cmin.x; g.y = cmin.y; g.z = cmin.z;
+  g.nx = (cmax.x - cmin.x)/g.dx + 1;
+  g.ny = (cmax.y - cmin.y)/g.dy + 1;
+  g.nz = (cmax.z - cmin.z)/g.dz + 1;
+  
+  s = gts_surface_new (gts_surface_class (), 
+		       gts_face_class (), 
+		       gts_edge_class (), 
+		       gts_vertex_class ());
+  data[0] = domain;
+  data[1] = &depth;
+  data[2] = v;
+  gts_isosurface_cartesian (s, g, (GtsIsoCartesianFunc) iso_func, data, val);
+
+  return s;
+}
 
 static void write_gnuplot (FttCell * cell, gpointer * data)
 {
@@ -232,7 +587,11 @@ static void write_gnuplot (FttCell * cell, gpointer * data)
   GtsBBox * bbox = data[2];
   FttVector pos;
   
-  ftt_cell_pos (cell, &pos);
+  if (v->centered)
+    ftt_cell_pos (cell, &pos);
+  else
+    gfs_cell_cm (cell, &pos);
+
   if (bbox == NULL || (pos.x >= bbox->x1 && pos.x <= bbox->x2 &&
 		       pos.y >= bbox->y1 && pos.y <= bbox->y2 &&
 		       pos.z >= bbox->z1 && pos.z <= bbox->z2))
@@ -259,24 +618,6 @@ void gfs_write_gnuplot (GfsDomain * domain,
 			   (FttCellTraverseFunc) write_gnuplot, data);
 }
 
-static void max_extent (FttCell * cell, FttVector * max)
-{
-  FttVector pos;
-  
-  ftt_cell_pos (cell, &pos);
-  if (pos.x > max->x) max->x = pos.x;
-  if (pos.y > max->y) max->y = pos.y;
-}
-
-static void min_extent (FttCell * cell, FttVector * min)
-{
-  FttVector pos;
-  
-  ftt_cell_pos (cell, &pos);
-  if (pos.x < min->x) min->x = pos.x;
-  if (pos.y < min->y) min->y = pos.y;
-}
-
 typedef struct {
   guchar r, g, b;
 } Color;
@@ -337,26 +678,22 @@ static void image_draw_square (Image * im,
 			       FttVector * p1, FttVector * p2,
 			       Color c)
 {
-  guint i1, j1, i2, j2, i, j;
+  gint i1, j1, i2, j2, i, j;
 
   i1 = (p1->x - im->min.x)*im->size;
   i2 = (p2->x - im->min.x)*im->size;
   j1 = (p1->y - im->min.y)*im->size;
   j2 = (p2->y - im->min.y)*im->size;
 
-  g_assert (i1 < im->width);
-  g_assert (i2 < im->width);
-  g_assert (j1 < im->height);
-  g_assert (j2 < im->height);
-
   j1 = im->height - 1 - j1;
   j2 = im->height - 1 - j2;
   for (i = i1; i <= i2; i++)
-    for (j = j2; j <= j1; j++) {
-      im->data[j][i][0] = c.r;
-      im->data[j][i][1] = c.g;
-      im->data[j][i][2] = c.b;
-    }
+    for (j = j2; j <= j1; j++) 
+      if (i >= 0 && i < im->width && j >= 0 && j < im->height) {
+	im->data[j][i][0] = c.r;
+	im->data[j][i][1] = c.g;
+	im->data[j][i][2] = c.b;
+      }
 }
 
 static void write_image_square (FttCell * cell, gpointer * data)
@@ -368,8 +705,7 @@ static void write_image_square (FttCell * cell, gpointer * data)
   Image * image = data[4];
   FttVector * lambda = data[5];
   FttVector p;
-  GtsColor fc = colormap_color (colormap, 
-			(GFS_VARIABLE (cell, v->i) - *min)/(*max - *min));
+  GtsColor fc = colormap_color (colormap, (GFS_VARIABLE (cell, v->i) - *min)/(*max - *min));
   Color c;
   gdouble size = ftt_cell_size (cell)/2.;
   FttVector p1, p2;
@@ -385,14 +721,34 @@ static void write_image_square (FttCell * cell, gpointer * data)
   image_draw_square (image, &p1, &p2, c);
 }
 
+static void max_extent (FttCell * cell, FttVector * max)
+{
+  FttVector pos;
+  
+  ftt_cell_pos (cell, &pos);
+  if (pos.x > max->x) max->x = pos.x;
+  if (pos.y > max->y) max->y = pos.y;
+  if (pos.z > max->z) max->z = pos.z;
+}
+
+static void min_extent (FttCell * cell, FttVector * min)
+{
+  FttVector pos;
+  
+  ftt_cell_pos (cell, &pos);
+  if (pos.x < min->x) min->x = pos.x;
+  if (pos.y < min->y) min->y = pos.y;
+  if (pos.z < min->z) min->z = pos.z;
+}
+
 void gfs_write_ppm (GfsDomain * domain, 
+		    GtsBBox * box,
 		    GfsVariable * v, gdouble min, gdouble max,
 		    FttTraverseFlags flags,
 		    gint level,
 		    FILE * fp)
 {
   Colormap * colormap;
-  gdouble h;
   guint depth, size = 1;
   Image * image;
   FttVector cmax = { - G_MAXDOUBLE, - G_MAXDOUBLE, - G_MAXDOUBLE };
@@ -411,21 +767,33 @@ void gfs_write_ppm (GfsDomain * domain,
   while (depth-- > 0)
     size *= 2;
 
-  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL, 
-			    domain->rootlevel,
-			    (FttCellTraverseFunc) min_extent, &cmin);
-  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL,
-			    domain->rootlevel,
-			    (FttCellTraverseFunc) max_extent, &cmax);
-  if (cmin.x == G_MAXDOUBLE)
-    return;
+  if (box != NULL) {
+    cmin.x = box->x1/domain->lambda.x; 
+    cmin.y = box->y1/domain->lambda.y; 
+    cmin.z = box->z1;
+    cmax.x = box->x2/domain->lambda.x; 
+    cmax.y = box->y2/domain->lambda.y; 
+    cmax.z = box->z2;
+  }
+  else {
+    gdouble h;
+
+    gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL, 
+			      domain->rootlevel,
+			      (FttCellTraverseFunc) min_extent, &cmin);
+    gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL,
+			      domain->rootlevel,
+			      (FttCellTraverseFunc) max_extent, &cmax);
+    if (cmin.x == G_MAXDOUBLE)
+      return;
+    h = ftt_level_size (domain->rootlevel)/2.;
+    cmin.x = (cmin.x - h)/domain->lambda.x; 
+    cmin.y = (cmin.y - h)/domain->lambda.y;
+    cmax.x = (cmax.x + h)/domain->lambda.x; 
+    cmax.y = (cmax.y + h)/domain->lambda.y;
+  }
 
   colormap = colormap_jet ();
-  h = ftt_level_size (domain->rootlevel)/2.;
-  cmin.x = (cmin.x - h)/domain->lambda.x; 
-  cmin.y = (cmin.y - h)/domain->lambda.y;
-  cmax.x = (cmax.x + h)/domain->lambda.x; 
-  cmax.y = (cmax.y + h)/domain->lambda.y;
   image = image_new (cmin, cmax, size);
 
   data[0] = colormap;
@@ -434,8 +802,12 @@ void gfs_write_ppm (GfsDomain * domain,
   data[3] = v;
   data[4] = image;
   data[5] = &domain->lambda;
-  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, flags, level,
-			   (FttCellTraverseFunc) write_image_square, data);
+  if (box != NULL)
+    gfs_domain_cell_traverse_box (domain, box, FTT_PRE_ORDER, flags, level,
+				  (FttCellTraverseFunc) write_image_square, data);
+  else
+    gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, flags, level,
+			      (FttCellTraverseFunc) write_image_square, data);
   image_write (image, fp);
   image_destroy (image);
   colormap_destroy (colormap);
@@ -551,8 +923,6 @@ gint gfs_combine_ppm (gchar ** fname, guint nname, FILE * fp)
   }
 }
 
-#endif /* 2D */
-
 static void write_square (FttCell * cell, gpointer * data)
 {
   Colormap * colormap = data[0];
@@ -567,7 +937,7 @@ static void write_square (FttCell * cell, gpointer * data)
   ftt_cell_pos (cell, &p);
   c = colormap_color (colormap, 
 		      (GFS_VARIABLE (cell, v->i) - *min)/(*max - *min));
-#ifdef FTT_2D    
+#if FTT_2D    
   fprintf (fp, 
 	   "OFF 4 1 4\n"
 	   "%g %g 0\n%g %g 0\n%g %g 0\n%g %g 0\n"
@@ -661,7 +1031,7 @@ static void write_mac (FttCellFace * face, gpointer * data)
     switch (c) {
     case FTT_X: f.x = un; break;
     case FTT_Y: f.y = un; break;
-#ifndef FTT_2D
+#if (!FTT_2D)
     case FTT_Z: f.z = un; break;
 #endif /* not FTT_2D */
     default: g_assert_not_reached ();
@@ -701,7 +1071,7 @@ void gfs_write_mac_velocity (GfsDomain * domain,
   norm = gfs_domain_norm_velocity (domain, flags, level);
   scale = norm.infty > 0. ? 
     ftt_level_size (level < 0 ? gfs_domain_depth (domain) : level)*
-    scale/norm.infty : 1.;
+    scale/norm.infty : scale;
   data[0] = &scale;
   data[1] = fp;
   data[2] = bbox;
@@ -711,9 +1081,9 @@ void gfs_write_mac_velocity (GfsDomain * domain,
 }
 
 void gfs_draw_cells (FttCell * cell, 
-		    FttTraverseFlags flags,
-		    gint level,
-		    FILE * fp)
+		     FttTraverseFlags flags,
+		     gint level,
+		     FILE * fp)
 {
   g_return_if_fail (cell != NULL);
   g_return_if_fail (fp != NULL);
@@ -750,7 +1120,7 @@ static void draw_box_boundaries (GfsBox * box, FILE * fp)
       if (o->klass->color) {
 	GtsColor c = (* o->klass->color) (o);
 
-#ifdef FTT_2D
+#if FTT_2D
 	fprintf (fp, "appearance { material { edgecolor %g %g %g } }\n", 
 		 c.r, c.g, c.b);
 #else /* 3D */
@@ -766,7 +1136,7 @@ static void draw_box_boundaries (GfsBox * box, FILE * fp)
       fputs ("}\n", fp);
     }
     else if (GFS_IS_BOX (o) && box->pid != GFS_BOX (o)->pid) {
-#ifdef FTT_2D
+#if FTT_2D
       fputs ("appearance { material { edgecolor 1 0 0 } }\n", fp);
 #else /* 3D */
       fputs ("appearance { material { ambient 1 0 0 diffuse 1 0 0 } }\n", fp);
@@ -795,7 +1165,7 @@ void gfs_draw_boundary_conditions (GfsDomain * domain, FILE * fp)
 
   fputs ("(geometry \"conditions\" = \n"
 	 "LIST {\n", fp);
-#ifdef FTT_2D
+#if FTT_2D
   fputs ("appearance { linewidth 2 }\n", fp);
 #endif /* 2D */
   gts_container_foreach (GTS_CONTAINER (domain), 
@@ -829,7 +1199,7 @@ void gfs_draw_solid_boundaries (GfsDomain * domain, FILE * fp)
   g_return_if_fail (fp != NULL);
 
   fputs ("(geometry \"solid\" = {\n", fp);
-#ifdef FTT_2D
+#if FTT_2D
   fputs ("appearance { linewidth 2 }\n", fp);
 #endif /* 2D */
   fputs ("LIST{\n", fp);
@@ -920,7 +1290,7 @@ static GtsColor variable_color (GtsObject * o)
 
   cell = gfs_domain_locate (domain, pos, -1);
   if (cell) {
-    val = gfs_interpolate (cell, pos, v->i);
+    val = gfs_interpolate (cell, pos, v);
     c = colormap_color (colormap, (val - *min)/(*max - *min));
   }
   else
@@ -943,8 +1313,7 @@ void gfs_draw_surface (GfsDomain * domain,
 
   gfs_domain_cell_traverse (domain, 
 			    FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-			    (FttCellTraverseFunc) gfs_get_from_below_intensive,
-			    v);
+			    (FttCellTraverseFunc) v->fine_coarse, v);
   if (min == max)
     max = min + 1.;
   colormap = colormap_jet ();
@@ -1194,7 +1563,7 @@ GtsPointClass * gfs_twisted_vertex_class (void)
     GtsObjectClassInfo gfs_twisted_vertex_info = {
       "GfsTwistedVertex",
       sizeof (GfsTwistedVertex),
-      sizeof (GtsPointClass),
+      sizeof (GtsVertexClass),
       (GtsObjectClassInitFunc) gfs_twisted_vertex_class_init,
       (GtsObjectInitFunc) NULL,
       (GtsArgSetFunc) NULL,
@@ -1259,17 +1628,19 @@ static void edge_list (GtsMatrix * b, GtsPoint * o,
   if (GFS_IS_TWISTED_VERTEX (o)) {
     gdouble t = GFS_TWISTED_VERTEX (o)->theta;
     gdouble sint = sin (t), cost = cos (t);
-    GtsMatrix * r = gts_matrix_new (cost, -sint, 0.,
-				    sint,  cost, 0.,
-				    0.,      0., 1.);
+    GtsMatrix * r = gts_matrix_new (cost, -sint, 0., 0.,
+				    sint,  cost, 0., 0.,
+				    0.,      0., 1., 0.,
+				    0.,      0., 0., 0.);
     
     c = gts_matrix_product (b, r);
     gts_matrix_destroy (r);
   }
   else
-    c = gts_matrix_new (b[0][0], b[0][1], b[0][2],
-			b[1][0], b[1][1], b[1][2],
-			b[2][0], b[2][1], b[2][2]);
+    c = gts_matrix_new (b[0][0], b[0][1], b[0][2], 0.,
+			b[1][0], b[1][1], b[1][2], 0.,
+			b[2][0], b[2][1], b[2][2], 0.,
+			0., 0., 0., 0.);
 
   for (i = 0; i <= ne; i++, profile = profile->next) {
     GtsPoint * p = profile->data;
@@ -1360,9 +1731,7 @@ size /= 4.;
   if (path == NULL)
     return;
   p1 = path->data;
-  r = gts_matrix_new (1., 0., 0.,
-		      0., 1., 0.,
-		      0., 0., 1.);
+  r = gts_matrix_identity (NULL);
   ne = g_slist_length (profile) - 1;
   e1 = g_malloc (sizeof (GtsEdge *)*ne);
   e2 = g_malloc (sizeof (GtsEdge *)*ne);
@@ -1426,7 +1795,7 @@ static GSList * ribbon_profile (GtsPointClass * klass,
   return lp;
 }
 
-#ifndef FTT_2D
+#if (!FTT_2D)
 static void vorticity_vector (FttCell * cell)
 {
   gdouble size = ftt_cell_size (cell);
@@ -1440,15 +1809,16 @@ static void vorticity_vector (FttCell * cell)
 }
 #endif /* 3D */
 
-GSList * gfs_streamline_new (GfsDomain * domain,
-			     FttVector p,
-			     GfsVariable * var,
-			     gdouble min, 
-			     gdouble max,
-			     gboolean twist)
+static GSList * grow_curve (GfsDomain * domain,
+			    FttVector p,
+			    GfsVariable * var,
+			    gdouble min, 
+			    gdouble max,
+			    gboolean twist,
+			    GSList * path,
+			    gdouble direction)
 {
   FttCell * cell;
-  GSList * path = NULL;
   gdouble delta = 0.1;
   GtsPoint * oldp = NULL;
   FttVector p1, p2;
@@ -1457,13 +1827,12 @@ GSList * gfs_streamline_new (GfsDomain * domain,
   guint nstep = 0, nmax = 10000;
   GtsPointClass * path_class = gfs_vertex_class ();
   Colormap * colormap = NULL;
-
-  g_return_val_if_fail (domain != NULL, NULL);
+  GfsVariable * U;
 
   if (min < max)
     colormap = colormap_jet ();
 
-#ifndef FTT_2D
+#if (!FTT_2D)
   if (twist) {
     path_class = GTS_POINT_CLASS (gfs_twisted_vertex_class ());
     {
@@ -1474,18 +1843,19 @@ GSList * gfs_streamline_new (GfsDomain * domain,
 			     (FttCellTraverseFunc) vorticity_vector, NULL);
       for (c = 0, g = gfs_gx; c < FTT_DIMENSION; c++, g = g->next)
 	gfs_domain_cell_traverse (domain,
-				 FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-				 (FttCellTraverseFunc) 
-				 gfs_get_from_below_intensive, g);
+				  FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
+				  (FttCellTraverseFunc) g->fine_coarse, g);
     }
   }
 #endif /* 3D */  
 
+  U = gfs_variable_from_name (domain->variables, "U");
   p1 = p2 = p;
   while ((cell = gfs_domain_locate (domain, p, -1)) != NULL && nmax--) {
     gdouble h = delta*ftt_cell_size (cell);
     FttVector u;
     FttComponent c;
+    GfsVariable * v;
     gdouble nu = 0.;
 
     cost += curve_cost (p1, p2, p);
@@ -1494,7 +1864,7 @@ GSList * gfs_streamline_new (GfsDomain * domain,
     if (oldp == NULL || cost > maxcost) {
       oldp = gts_point_new (path_class, p.x, p.y, p.z);
       if (var)
-	GFS_VERTEX (oldp)->v = gfs_interpolate (cell, p, var->i);
+	GFS_VERTEX (oldp)->v = gfs_interpolate (cell, p, var);
       if (colormap)
 	GTS_COLORED_VERTEX (oldp)->c = 
 	  colormap_color (colormap, (GFS_VERTEX (oldp)->v - min)/(max - min));
@@ -1505,8 +1875,8 @@ GSList * gfs_streamline_new (GfsDomain * domain,
       nstep = 0;
     }
 
-    for (c = 0; c < FTT_DIMENSION; c++) {
-      ((gdouble *) &u)[c] = gfs_interpolate (cell, p, GFS_VELOCITY_INDEX (c));
+    for (c = 0, v = U; c < FTT_DIMENSION; c++, v = v->next) {
+      ((gdouble *) &u)[c] = direction*gfs_interpolate (cell, p, v);
       nu += ((gdouble *) &u)[c]*((gdouble *) &u)[c];
     }
     if (nu > 0. && nstep++ < nmax) {
@@ -1516,14 +1886,14 @@ GSList * gfs_streamline_new (GfsDomain * domain,
       nu = sqrt (nu);
       for (c = 0; c < FTT_DIMENSION; c++)
 	((gdouble *) &p)[c] += h*((gdouble *) &u)[c]/nu;
-#ifndef FTT_2D
+#if (!FTT_2D)
       if (twist) {
 	GtsVector rot;
 	GtsVector dx;
 
 	dx[0] = p1.x - p.x; dx[1] = p1.y - p.y; dx[2] = p1.z - p.z;
-	for (c = 0; c < FTT_DIMENSION; c++)
-	  rot[c] = gfs_interpolate (cell, p1, GFS_GRADIENT_INDEX (c));
+	for (c = 0, v = gfs_gx; c < FTT_DIMENSION; c++, v = v->next)
+	  rot[c] = gfs_interpolate (cell, p1, v);
 	theta += gts_vector_scalar (rot, dx)/nu;
       }
 #endif /* 3D */
@@ -1536,7 +1906,7 @@ GSList * gfs_streamline_new (GfsDomain * domain,
     if (cell) {
       oldp = gts_point_new (path_class, p2.x, p2.y, p2.z);
       if (var)
-	GFS_VERTEX (oldp)->v = gfs_interpolate (cell, p2, var->i);
+	GFS_VERTEX (oldp)->v = gfs_interpolate (cell, p2, var);
       if (twist)
 	GFS_TWISTED_VERTEX (oldp)->theta = theta;
       path = g_slist_prepend (path, oldp);
@@ -1546,7 +1916,21 @@ GSList * gfs_streamline_new (GfsDomain * domain,
   if (colormap)
     colormap_destroy (colormap);
 
-  return g_slist_reverse (path);
+  return direction > 0. ? g_slist_reverse (path) : path;
+}
+
+GSList * gfs_streamline_new (GfsDomain * domain,
+			     FttVector p,
+			     GfsVariable * var,
+			     gdouble min, 
+			     gdouble max,
+			     gboolean twist)
+{
+  GSList * path;
+
+  path = grow_curve (domain, p, var, min, max, twist, NULL, 1.);
+  path = grow_curve (domain, p, var, min, max, twist, path, -1.);
+  return path;
 }
 
 void gfs_streamline_write (GSList * stream, FILE * fp)
@@ -1676,8 +2060,8 @@ void gfs_draw_stream_ribbon (GfsDomain * domain,
 }
 
 void gfs_draw_streamline (GfsDomain * domain,
-			 FttVector p,
-			 FILE * fp)
+			  FttVector p,
+			  FILE * fp)
 {
   GSList * path;
 
