@@ -17,6 +17,8 @@
  * 02111-1307, USA.  
  */
 
+#include <stdlib.h>
+#include <math.h>
 #include <gmodule.h>
 #include "config.h"
 #include "gfsconfig.h"
@@ -27,6 +29,44 @@
 #include "adaptive.h"
 #include "source.h"
 #include "vof.h"
+#include "tension.h"
+
+/* GfsVariableTracer: object */
+
+static void variable_tracer_init (GfsVariableTracer * v)
+{
+  gfs_advection_params_init (&v->advection);
+  v->advection.gradient = gfs_center_van_leer_gradient;
+  v->advection.flux = gfs_face_advection_flux;
+  v->advection.v = GFS_VARIABLE1 (v);
+  v->advection.fv = gfs_res;
+
+  gfs_multilevel_params_init (&v->diffusion);
+  v->diffusion.tolerance = 1e-6;
+}
+
+GfsVariableClass * gfs_variable_tracer_class (void)
+{
+  static GfsVariableClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_variable_tracer_info = {
+      "GfsVariableTracer",
+      sizeof (GfsVariableTracer),
+      sizeof (GfsVariableClass),
+      (GtsObjectClassInitFunc) NULL,
+      (GtsObjectInitFunc) variable_tracer_init,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_variable_class ()), 
+				  &gfs_variable_tracer_info);
+  }
+
+  return klass;
+}
+
+/* GfsSimulation: object */
 
 static void simulation_destroy (GtsObject * object)
 {
@@ -56,6 +96,7 @@ static void simulation_destroy (GtsObject * object)
     i = i->next;
   }
   g_slist_free (sim->modules);
+  g_slist_free (sim->variables);
 
   (* GTS_OBJECT_CLASS (gfs_simulation_class ())->parent_class->destroy) 
     (object);
@@ -65,6 +106,7 @@ static void simulation_write (GtsObject * object, FILE * fp)
 {
   GfsSimulation * sim = GFS_SIMULATION (object);
   GSList * i;
+  GfsVariable * v;
 
   (* GTS_OBJECT_CLASS (gfs_simulation_class ())->parent_class->write)
     (object, fp);
@@ -76,20 +118,36 @@ static void simulation_write (GtsObject * object, FILE * fp)
   gfs_physical_params_write (&sim->physical_params, fp);
   fputs ("\n  GfsAdvectionParams ", fp);
   gfs_advection_params_write (&sim->advection_params, fp);
-  fputs ("\n  GfsTracerAdvectionParams ", fp);
-  gfs_advection_params_write (&sim->tracer_advection_params, fp);
   fputs ("\n  GfsApproxProjectionParams ", fp);
-  gfs_projection_params_write (&sim->approx_projection_params, fp);
+  gfs_multilevel_params_write (&sim->approx_projection_params, fp);
   fputs ("\n  GfsProjectionParams ", fp);
-  gfs_projection_params_write (&sim->projection_params, fp);
+  gfs_multilevel_params_write (&sim->projection_params, fp);
   fputc ('\n', fp);
 
+  i = sim->variables;
+  while (i) {
+    fputs ("  ", fp);
+    (* GTS_OBJECT (i->data)->klass->write) (i->data, fp);
+    fputc ('\n', fp);
+    i = i->next;
+  }
+
   i = sim->modules;
   while (i) {
     fprintf (fp, "  GModule %s\n", g_module_name (i->data));
     i = i->next;
   }
 
+  v = GFS_DOMAIN (sim)->variables;
+  while (v) {
+    if (v->surface_bc) {
+      fputs ("  ", fp);
+      (* GTS_OBJECT (v->surface_bc)->klass->write) (GTS_OBJECT (v->surface_bc), fp);
+      fputc ('\n', fp);
+    }
+    v = v->next;
+  }
+
   if (GFS_DOMAIN (sim)->max_depth_write < -1) {
     i = sim->refines->items;
     while (i) {
@@ -130,9 +188,23 @@ static void simulation_write (GtsObject * object, FILE * fp)
 
   if (sim->surface) {
     fputs ("  GtsSurface { ", fp);
-    gts_surface_write (sim->surface, fp);
+    if (GFS_DOMAIN (sim)->binary) {
+      gboolean binary = GTS_POINT_CLASS (sim->surface->vertex_class)->binary;
+      GTS_POINT_CLASS (sim->surface->vertex_class)->binary = TRUE;
+      gts_surface_write (sim->surface, fp);
+      GTS_POINT_CLASS (sim->surface->vertex_class)->binary = binary;
+    }
+    else
+      gts_surface_write (sim->surface, fp);
     fputs ("}\n", fp);
   }
+#if 1
+  if (sim->interface) {
+    fputs ("  GtsInterface { ", fp);
+    gts_surface_write (sim->interface, fp);
+    fputs ("}\n", fp);
+  }
+#endif
   fputc ('}', fp);
 }
 
@@ -160,6 +232,123 @@ static void check_solid_surface (GtsSurface * s,
   g_string_free (name, TRUE);
 }
 
+static GtsSurface * read_surface_file (GtsFile * fp, GtsSurface * surface)
+{
+  GtsSurface * s;
+  FILE * fptr;
+  GtsFile * fp1;
+  
+  gts_file_next_token (fp);
+  if (fp->type != GTS_STRING) {
+    gts_file_error (fp, "expecting a string (filename)");
+    return NULL;
+  }
+  s = gts_surface_new (gts_surface_class (),
+		       gts_face_class (),
+		       gts_edge_class (),
+		       surface ? surface->vertex_class : gts_vertex_class ());
+  fptr = fopen (fp->token->str, "rt");
+  if (fptr == NULL) {
+    gts_file_error (fp, "cannot open file `%s'", fp->token->str);
+    return NULL;
+  }
+  fp1 = gts_file_new (fptr);
+  if (gts_surface_read (s, fp1)) {
+    gts_file_error (fp, 
+		    "file `%s' is not a valid GTS file\n"
+		    "%s:%d:%d: %s",
+		    fp->token->str, fp->token->str,
+		    fp1->line, fp1->pos, fp1->error);
+    gts_file_destroy (fp1);
+    fclose (fptr);
+    gts_object_destroy (GTS_OBJECT (s));
+    return NULL;
+  }
+  gts_file_destroy (fp1);
+  fclose (fptr);
+  
+  check_solid_surface (s, fp->token->str, fp);
+  if (fp->type == GTS_ERROR) {
+    gts_object_destroy (GTS_OBJECT (s));
+    return NULL;
+  }
+  
+  if (surface) {
+    GtsSurface * self;
+    
+    gts_surface_merge (surface, s);
+    gts_object_destroy (GTS_OBJECT (s));
+    if ((self = gts_surface_is_self_intersecting (surface))) {
+      gts_object_destroy (GTS_OBJECT (self));
+      gts_file_error (fp, "merged surface is self-intersecting");
+      return NULL;
+    }
+    return surface;
+  }
+  return s;
+}
+
+static GtsSurface * read_surface (GtsFile * fp, GtsSurface * surface)
+{
+  GtsSurface * s;
+
+  gts_file_next_token (fp);
+  if (fp->type != '{') {
+    gts_file_error (fp, "expecting an opening brace");
+    return NULL;
+  }
+  fp->scope_max++;
+  gts_file_next_token (fp);
+  
+  s = gts_surface_new (gts_surface_class (),
+		       gts_face_class (),
+		       gts_edge_class (),
+		       surface ? surface->vertex_class : gts_vertex_class ());
+  
+  if (gts_surface_read (s, fp)) {
+    gts_object_destroy (GTS_OBJECT (s));
+    return NULL;
+  }
+  if (fp->type != '}') {
+    gts_object_destroy (GTS_OBJECT (s));
+    gts_file_error (fp, "expecting a closing brace");
+    return NULL;
+  }
+  fp->scope_max--;
+  
+  check_solid_surface (s, NULL, fp);
+  if (fp->type == GTS_ERROR) {
+    gts_object_destroy (GTS_OBJECT (s));
+    return NULL;
+  }
+  
+  if (surface) {
+    GtsSurface * self;
+    
+    gts_surface_merge (surface, s);
+    gts_object_destroy (GTS_OBJECT (s));
+    if ((self = gts_surface_is_self_intersecting (surface))) {
+      gts_object_destroy (GTS_OBJECT (self));
+      gts_file_error (fp, "merged surface is self-intersecting");
+      return NULL;
+    }
+    return surface;
+  }
+  return s;
+}
+
+static gboolean strmatch (const gchar * s, const gchar * s1)
+{
+  gboolean m = !strcmp (s, s1);
+
+  if (!m) {
+    gchar * s2 = g_strconcat ("Gfs", s, NULL);
+    m = !strcmp (s2, s1);
+    g_free (s2);
+  }
+  return m;
+}
+
 static void simulation_read (GtsObject ** object, GtsFile * fp)
 {
   GfsSimulation * sim = GFS_SIMULATION (*object);
@@ -188,102 +377,57 @@ static void simulation_read (GtsObject ** object, GtsFile * fp)
     }
 
     /* ------------ GtsSurface ------------ */
-    if (!strcmp (fp->token->str, "GtsSurface")) {
+    if (strmatch (fp->token->str, "GtsSurface")) {
       GtsSurface * s;
-
-      gts_file_next_token (fp);
-      if (fp->type != '{') {
-	gts_file_error (fp, "expecting an opening brace");
-	return;
-      }
-      fp->scope_max++;
-      gts_file_next_token (fp);
-
-      s = gts_surface_new (gts_surface_class (),
-			   gts_face_class (),
-			   gts_edge_class (),
-			   gts_vertex_class ());
       
-      if (gts_surface_read (s, fp)) {
-	gts_object_destroy (GTS_OBJECT (s));
+      if ((s = read_surface (fp, sim->surface)) == NULL)
 	return;
-      }
-      if (fp->type != '}') {
-	gts_object_destroy (GTS_OBJECT (s));
-	gts_file_error (fp, "expecting a closing brace");
-	return;
-      }
-      fp->scope_max--;
-
-      check_solid_surface (s, NULL, fp);
-      if (fp->type == GTS_ERROR) {
-	gts_object_destroy (GTS_OBJECT (s));
-	return;
-      }
-      
-      if (sim->surface) {
-	gts_surface_merge (sim->surface, s);
-	gts_object_destroy (GTS_OBJECT (s));
-      }
-      else
-	sim->surface = s;
-
+      sim->surface = s;
       gts_file_next_token (fp);
     }
 
     /* ------------ GtsSurfaceFile ------------ */
-    else if (!strcmp (fp->token->str, "GtsSurfaceFile")) {
+    else if (strmatch (fp->token->str, "GtsSurfaceFile")) {
       GtsSurface * s;
-      FILE * fptr;
-      GtsFile * fp1;
 
+      if ((s = read_surface_file (fp, sim->surface)) == NULL)
+	return;
+      sim->surface = s;
       gts_file_next_token (fp);
-      if (fp->type != GTS_STRING) {
-	gts_file_error (fp, "expecting a string (filename)");
-	return;
-      }
-      s = gts_surface_new (gts_surface_class (),
-			   gts_face_class (),
-			   gts_edge_class (),
-			   gts_vertex_class ());
-      fptr = fopen (fp->token->str, "rt");
-      if (fptr == NULL) {
-	gts_file_error (fp, "cannot open file `%s'", fp->token->str);
-	return;
-      }
-      fp1 = gts_file_new (fptr);
-      if (gts_surface_read (s, fp1)) {
-	gts_file_error (fp, 
-			"file `%s' is not a valid GTS file\n"
-			"%s:%d:%d: %s",
-			fp->token->str, fp->token->str,
-			fp1->line, fp1->pos, fp1->error);
-	gts_file_destroy (fp1);
-	fclose (fptr);
-	gts_object_destroy (GTS_OBJECT (s));
-	return;
-      }
-      gts_file_destroy (fp1);
-      fclose (fptr);
+    }
 
-      check_solid_surface (s, fp->token->str, fp);
-      if (fp->type == GTS_ERROR) {
-	gts_object_destroy (GTS_OBJECT (s));
+    /* ------------ GtsInterface ------------ */
+    else if (strmatch (fp->token->str, "GtsInterface")) {
+      GtsSurface * s;
+
+      if (!sim->interface)
+	sim->interface = gts_surface_new (gts_surface_class (),
+					  gts_face_class (),
+					  gts_edge_class (),
+					  gts_vertex_normal_class ());
+      if ((s = read_surface (fp, sim->interface)) == NULL)
 	return;
-      }
+      sim->interface = s;
+      gts_file_next_token (fp);
+    }
 
-      if (sim->surface) {
-	gts_surface_merge (sim->surface, s);
-	gts_object_destroy (GTS_OBJECT (s));
-      }
-      else
-	sim->surface = s;
+    /* ------------ GtsInterfaceFile ------------ */
+    else if (strmatch (fp->token->str, "GtsInterfaceFile")) {
+      GtsSurface * s;
 
+      if (!sim->interface)
+	sim->interface = gts_surface_new (gts_surface_class (),
+					  gts_face_class (),
+					  gts_edge_class (),
+					  gts_vertex_normal_class ());
+      if ((s = read_surface_file (fp, sim->interface)) == NULL)
+	return;
+      sim->interface = s;
       gts_file_next_token (fp);
     }
 
     /* ------------ GModule ------------ */
-    else if (!strcmp (fp->token->str, "GModule")) {
+    else if (strmatch (fp->token->str, "GModule")) {
       gts_file_next_token (fp);
       if (fp->type != GTS_STRING) {
 	gts_file_error (fp, "expecting a string (filename)");
@@ -292,15 +436,14 @@ static void simulation_read (GtsObject ** object, GtsFile * fp)
       if (!g_module_supported ())
 	g_warning ("modules are not supported on this system");
       else {
-	GString * name = g_string_new (GFS_MODULES_DIR);
+	gchar * name, * path;
 	GModule * module;
 
-	g_string_append (name, "/lib");
-	g_string_append (name, fp->token->str);
- 	g_string_append (name, FTT_DIMENSION == 2 ? "2D" : "3D");
-	g_string_append (name, ".so");
-	module = g_module_open (name->str, 0);
-	g_string_free (name, TRUE);	
+	name = g_strconcat (fp->token->str, FTT_DIMENSION == 2 ? "2D" : "3D");
+	path = g_module_build_path (GFS_MODULES_DIR, name);
+	g_free (name);
+	module = g_module_open (path, 0);
+	g_free (path);
 	if (module == NULL)
 	  module = g_module_open (fp->token->str, 0);
 	if (module == NULL) {
@@ -314,7 +457,7 @@ static void simulation_read (GtsObject ** object, GtsFile * fp)
     }
 
     /* ------------ GfsTime ------------ */
-    else if (!strcmp (fp->token->str, "GfsTime")) {
+    else if (strmatch (fp->token->str, "GfsTime")) {
       gts_file_next_token (fp);
       gfs_time_read (&sim->time, fp);
       if (fp->type == GTS_ERROR)
@@ -322,7 +465,7 @@ static void simulation_read (GtsObject ** object, GtsFile * fp)
     }
 
     /* ------------ GfsPhysicalParams ------------ */
-    else if (!strcmp (fp->token->str, "GfsPhysicalParams")) {
+    else if (strmatch (fp->token->str, "GfsPhysicalParams")) {
       gts_file_next_token (fp);
       gfs_physical_params_read (&sim->physical_params, fp);
       if (fp->type == GTS_ERROR)
@@ -330,51 +473,45 @@ static void simulation_read (GtsObject ** object, GtsFile * fp)
     }
 
     /* ------------ GfsProjectionParams ------------ */
-    else if (!strcmp (fp->token->str, "GfsProjectionParams")) {
+    else if (strmatch (fp->token->str, "GfsProjectionParams")) {
       gts_file_next_token (fp);
-      gfs_projection_params_read (&sim->projection_params, fp);
+      gfs_multilevel_params_read (&sim->projection_params, fp);
       if (fp->type == GTS_ERROR)
 	return;
     }
 
     /* ------------ GfsApproxProjectionParams ------------ */
-    else if (!strcmp (fp->token->str, "GfsApproxProjectionParams")) {
+    else if (strmatch (fp->token->str, "GfsApproxProjectionParams")) {
       gts_file_next_token (fp);
-      gfs_projection_params_read (&sim->approx_projection_params, fp);
+      gfs_multilevel_params_read (&sim->approx_projection_params, fp);
       if (fp->type == GTS_ERROR)
 	return;
     }
 
     /* ------------ GfsAdvectionParams ------------ */
-    else if (!strcmp (fp->token->str, "GfsAdvectionParams")) {
+    else if (strmatch (fp->token->str, "GfsAdvectionParams")) {
       gts_file_next_token (fp);
       gfs_advection_params_read (&sim->advection_params, fp);
       if (fp->type == GTS_ERROR)
 	return;
     }
 
-    /* ------------ GfsTracerAdvectionParams ------------ */
-    else if (!strcmp (fp->token->str, "GfsTracerAdvectionParams")) {
-      gts_file_next_token (fp);
-      gfs_advection_params_read (&sim->tracer_advection_params, fp);
-      if (fp->type == GTS_ERROR)
-	return;
-    }
-
     /* ------------ GtsObject ------------ */
     else {
-      GtsObjectClass * klass = gts_object_class_from_name (fp->token->str);
+      GtsObjectClass * klass = gfs_object_class_from_name (fp->token->str);
       GtsObject * object;
 
       if (klass == NULL ||
 	  (!gts_object_class_is_from_class (klass, gfs_refine_class ()) &&
-	   !gts_object_class_is_from_class (klass, gfs_event_class ()))) {
+	   !gts_object_class_is_from_class (klass, gfs_event_class ()) &&
+	   !gts_object_class_is_from_class (klass, gfs_variable_class ()) &&
+	   !gts_object_class_is_from_class (klass, gfs_surface_generic_bc_class ()))) {
 	gts_file_error (fp, "unknown keyword `%s'", fp->token->str);
 	return;
       }
 
       object = gts_object_new (klass);
-      object->reserved = sim;
+      gfs_object_simulation (object) = sim;
 
       g_assert (klass->read);
       (* klass->read) (&object, fp);
@@ -392,6 +529,21 @@ static void simulation_read (GtsObject ** object, GtsFile * fp)
       else if (GFS_IS_EVENT (object))
 	gts_container_add (GTS_CONTAINER (sim->events), 
 			   GTS_CONTAINEE (object));
+      else if (GFS_IS_VARIABLE (object)) {
+	GfsVariable * v = GFS_VARIABLE1 (object);
+	GfsVariable * old = gfs_variable_from_name (GFS_DOMAIN (sim)->variables, v->name);
+
+	if (old == NULL)
+	  gfs_domain_add_new_variable (GFS_DOMAIN (sim), v);
+	else {
+	  gfs_domain_replace_variable (GFS_DOMAIN (sim), old, v);
+	  gts_object_destroy (GTS_OBJECT (old));
+	  sim->variables = g_slist_remove (sim->variables, old);
+	}
+	sim->variables = g_slist_append (sim->variables, v);
+      }
+      else if (GFS_IS_SURFACE_GENERIC_BC (object))
+	;
       else
 	g_assert_not_reached ();
     }
@@ -411,19 +563,123 @@ static void simulation_read (GtsObject ** object, GtsFile * fp)
     if (gts_surface_volume (sim->surface) < 0.)
       sim->is_open = TRUE;
   }
+  if (sim->interface) {
+    if (sim->itree)
+      gts_bb_tree_destroy (sim->itree, TRUE);
+    sim->itree = gts_bb_tree_surface (sim->interface);
+    if (gts_surface_volume (sim->interface) < 0.)
+      sim->i_is_open = TRUE;
+  }
   sim->refines->items = g_slist_reverse (sim->refines->items);
   sim->adapts->items = g_slist_reverse (sim->adapts->items);
   sim->events->items = g_slist_reverse (sim->events->items);
   sim->modules = g_slist_reverse (sim->modules);
-  sim->tracer_advection_params.cfl = sim->advection_params.cfl;
   sim->advection_params.rho = sim->physical_params.rho;
 }
 
+static void simulation_run (GfsSimulation * sim)
+{
+  GfsVariable * v, * c, * ch;
+  GfsDomain * domain;
+
+  domain = GFS_DOMAIN (sim);
+
+  gfs_simulation_refine (sim);
+
+  gfs_simulation_event_init (sim, sim->events->items);
+  gfs_simulation_event_init (sim, sim->adapts->items);
+
+  gfs_set_merged (domain);
+  v = domain->variables;
+  while (v) {
+    gfs_domain_bc (domain, FTT_TRAVERSE_LEAFS, -1, v);
+    v = v->next;
+  }
+  c = gfs_variable_from_name (domain->variables, "C");
+  g_assert (c == NULL);
+  ch = gfs_variable_from_name (domain->variables, "CH");
+  g_assert (ch == NULL);
+
+  sim->advection_params.c = c;
+  gfs_approximate_projection (domain,
+      			      &sim->approx_projection_params,
+      			      &sim->advection_params);
+
+  gts_range_init (&domain->mpi_wait);
+  while (sim->time.t < sim->time.end &&
+	 sim->time.i < sim->time.iend) {
+    gfs_domain_cell_traverse (domain,
+			      FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
+			      (FttCellTraverseFunc) gfs_cell_coarse_init, domain);
+    gfs_simulation_event (sim, sim->events->items);
+
+    g_timer_start (domain->timer);
+
+    gfs_simulation_set_timestep (sim);
+
+    sim->advection_params.c = c;
+
+    gfs_predicted_face_velocities (domain, FTT_DIMENSION, &sim->advection_params);
+    gfs_mac_projection (domain, 
+    			&sim->projection_params, 
+    			&sim->advection_params);
+
+    v = domain->variables;
+    while (v) {
+      if (GFS_IS_VARIABLE_TRACER (v)) {
+	GfsVariableTracer * t = GFS_VARIABLE_TRACER (v);
+
+	t->advection.dt = sim->advection_params.dt;
+	switch (t->advection.scheme) {
+	case GFS_GODUNOV:
+	  gfs_tracer_advection_diffusion (domain, &t->advection, &t->diffusion, NULL);
+	  break;
+	case GFS_VOF:
+	  gfs_tracer_vof_advection (domain, &t->advection, NULL);
+	  gfs_domain_variable_centered_sources (domain, v, v, t->advection.dt);
+	  break;
+	case GFS_NONE:
+	  break;
+	}
+      }
+      v = v->next;
+    }
+
+    gfs_simulation_event_half (sim, sim->events->items);
+
+    sim->advection_params.c = ch;
+    gfs_centered_velocity_advection_diffusion (domain,
+					       FTT_DIMENSION,
+					       &sim->advection_params,
+					       &sim->diffusion_params);
+    gfs_simulation_adapt (sim, NULL);
+    gfs_approximate_projection (domain,
+   				&sim->approx_projection_params, 
+    				&sim->advection_params);
+
+    sim->time.t = sim->tnext;
+    sim->time.i++;
+
+    g_timer_stop (domain->timer);
+    gts_range_add_value (&domain->timestep,
+			 g_timer_elapsed (domain->timer, NULL));
+    gts_range_update (&domain->timestep);
+    gts_range_add_value (&domain->size, 
+			 gfs_domain_size (domain, FTT_TRAVERSE_LEAFS, -1));
+    gts_range_update (&domain->size);
+  }
+  gfs_simulation_event (sim, sim->events->items);
+  gts_container_foreach (GTS_CONTAINER (sim->events),
+			 (GtsFunc) gts_object_destroy, NULL);
+}
+
 static void gfs_simulation_class_init (GfsSimulationClass * klass)
 {
   GTS_OBJECT_CLASS (klass)->write =   simulation_write;
   GTS_OBJECT_CLASS (klass)->read =    simulation_read;
   GTS_OBJECT_CLASS (klass)->destroy = simulation_destroy;
+
+  klass->run = simulation_run;
 }
 
 static void gfs_simulation_init (GfsSimulation * object)
@@ -433,25 +689,28 @@ static void gfs_simulation_init (GfsSimulation * object)
 
   gfs_advection_params_init (&object->advection_params);
   object->advection_params.flux = gfs_face_velocity_advection_flux;
+  object->advection_params.fv = gfs_res;
+  gfs_multilevel_params_init (&object->diffusion_params);
+  object->diffusion_params.tolerance = 1e-6;
 
-  gfs_advection_params_init (&object->tracer_advection_params);
-  object->tracer_advection_params.gradient = gfs_center_van_leer_gradient;
-  object->tracer_advection_params.flux = gfs_face_advection_flux;
-  object->tracer_advection_params.fv = gfs_div;
-
-  gfs_projection_params_init (&object->projection_params);
-  gfs_projection_params_init (&object->approx_projection_params);
+  gfs_multilevel_params_init (&object->projection_params);
+  gfs_multilevel_params_init (&object->approx_projection_params);
 
   object->surface = NULL;
   object->stree = NULL;
   object->is_open = FALSE;
 
+  object->interface = NULL;
+  object->itree = NULL;
+  object->i_is_open = FALSE;
+
   object->refines = GTS_SLIST_CONTAINER (gts_container_new 
 					 (GTS_CONTAINER_CLASS 
 					  (gts_slist_container_class ())));
   object->adapts = GTS_SLIST_CONTAINER (gts_container_new 
 					(GTS_CONTAINER_CLASS 
 					 (gts_slist_container_class ())));
+  gfs_adapt_stats_init (&object->adapts_stats);
   object->events = GTS_SLIST_CONTAINER (gts_container_new 
 					(GTS_CONTAINER_CLASS 
 					 (gts_slist_container_class ())));
@@ -496,8 +755,18 @@ static void box_init_solid_fractions (GfsBox * box, GfsSimulation * sim)
 {
   gfs_cell_init_solid_fractions (box->root, 
 				 sim->surface, sim->stree, sim->is_open,
-				 TRUE, gfs_cell_cleanup);
-  g_assert (!FTT_CELL_IS_DESTROYED (box->root));
+				 TRUE, (FttCellCleanupFunc) gfs_cell_cleanup,
+				 NULL);
+  if (FTT_CELL_IS_DESTROYED (box->root)) {
+    FttVector p;
+
+    ftt_cell_pos (box->root, &p);
+    g_warning ("%s centered at (%g,%g,%g) is entirely filled by a solid.\n"
+	       "Aborting...\n", 
+	       GTS_OBJECT (box)->klass->info.name,
+	       p.x, p.y, p.z);
+    exit (1);
+  }
 }
 
 static void refine_cell_corner (FttCell * cell, GfsDomain * domain)
@@ -506,10 +775,24 @@ static void refine_cell_corner (FttCell * cell, GfsDomain * domain)
     ftt_cell_refine_single (cell, (FttCellInitFunc) gfs_cell_init, domain);
 }
 
-static void check_solid_fractions (GfsBox * box, GfsSimulation * sim)
+static void check_face (FttCellFace * f, guint * nf)
 {
-  gfs_cell_check_solid_fractions (box->root, 
-				 sim->surface, sim->stree, sim->is_open);
+  GfsSolidVector * s = GFS_STATE (f->cell)->solid;
+
+  if (s && !f->neighbor && s->s[f->d] > 0. && s->s[f->d] < 1.)
+    (*nf)++;
+}
+
+static void check_solid_fractions (GfsBox * box, gpointer * data)
+{
+  GfsSimulation * sim = data[0];
+  guint * nf = data[1];
+  FttDirection d;
+
+  gfs_cell_check_solid_fractions (box->root, sim->surface, sim->stree, sim->is_open);
+  for (d = 0; d < FTT_NEIGHBORS; d++)
+    ftt_face_traverse_boundary (box->root, d, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+				(FttFaceTraverseFunc) check_face, nf);
 }
 
 /**
@@ -522,8 +805,9 @@ static void check_solid_fractions (GfsBox * box, GfsSimulation * sim)
 void gfs_simulation_refine (GfsSimulation * sim)
 {
   GSList * i;
-  guint depth;
+  guint depth, nf = 0;
   gint l;
+  gpointer data[2];
 
   g_return_if_fail (sim != NULL);
 
@@ -546,33 +830,39 @@ void gfs_simulation_refine (GfsSimulation * sim)
   gfs_domain_match (GFS_DOMAIN (sim));
 
   if (sim->surface) {
-    gts_container_foreach (GTS_CONTAINER (sim),
-			   (GtsFunc) box_init_solid_fractions, sim);
+    gts_container_foreach (GTS_CONTAINER (sim), (GtsFunc) box_init_solid_fractions, sim);
     gfs_domain_match (GFS_DOMAIN (sim));
   }
-  gts_container_foreach (GTS_CONTAINER (sim),
-			 (GtsFunc) check_solid_fractions, sim);
+  data[0] = sim;
+  data[1] = &nf;
+  gts_container_foreach (GTS_CONTAINER (sim), (GtsFunc) check_solid_fractions, data);
+  if (nf > 0)
+    g_warning ("the solid surface cuts %d boundary cells,\n"
+	       "this may cause errors for diffusion terms\n", nf);
 }
 
 /**
  * gfs_simulation_read:
- * @sim: a #GfsSimulation.
  * @fp: a #GtsFile.
  *
- * Read a simulation file from @fp in @sim.
+ * Reads a simulation file from @fp.
  *
- * Returns: 0 in case of success or the line number at wich an error
- * occured, in which case the @pos and @error fields of @fp are set.
+ * Returns: the #GfsSimulation or %NULL if an error occured, in which
+ * case the @pos and @error fields of @fp are set.
  */
-guint gfs_simulation_read (GfsSimulation * sim, GtsFile * fp)
+GfsSimulation * gfs_simulation_read (GtsFile * fp)
 {
-  g_return_val_if_fail (sim != NULL, 1);
-  g_return_val_if_fail (fp != NULL, 1);
+  GfsDomain * d;
 
-  if (gfs_domain_read (GFS_DOMAIN (sim), fp))
-    return fp->line;
+  g_return_val_if_fail (fp != NULL, NULL);
 
-  return 0;
+  d = gfs_domain_read (fp);
+  if (d != NULL && !GFS_IS_SIMULATION (d)) {
+    gts_file_error (fp, "parent graph is not a GfsSimulation");
+    gts_object_destroy (GTS_OBJECT (d));
+    return NULL;
+  }
+  return GFS_SIMULATION (d);
 }
 
 /**
@@ -586,8 +876,8 @@ guint gfs_simulation_read (GfsSimulation * sim, GtsFile * fp)
  * smaller or equal to -2, no cell tree data is written.  
  */
 void gfs_simulation_write (GfsSimulation * sim,
-			  gint max_depth,		  
-			  FILE * fp)
+			   gint max_depth,		  
+			   FILE * fp)
 {
   gint depth;
   GfsDomain * domain;
@@ -624,9 +914,10 @@ void gfs_simulation_set_timestep (GfsSimulation * sim)
   g_return_if_fail (sim != NULL);
 
   t = sim->time.t;
-  sim->advection_params.dt = 
-    sim->advection_params.cfl*gfs_domain_cfl (GFS_DOMAIN (sim), 
-					      FTT_TRAVERSE_LEAFS, -1);
+  sim->advection_params.dt =
+    sim->advection_params.cfl*gfs_domain_cfl (GFS_DOMAIN (sim), FTT_TRAVERSE_LEAFS, -1);
+  if (sim->advection_params.dt > sim->time.dtmax)
+    sim->advection_params.dt = sim->time.dtmax;
   sim->tnext = t + sim->advection_params.dt;
 
   i = sim->events->items;
@@ -644,8 +935,6 @@ void gfs_simulation_set_timestep (GfsSimulation * sim)
     sim->advection_params.dt = sim->time.end - t;
     sim->tnext = sim->time.end;
   }
-
-  sim->tracer_advection_params.dt = sim->advection_params.dt;
 }
 
 /**
@@ -657,7 +946,7 @@ void gfs_simulation_set_timestep (GfsSimulation * sim)
  * activates it if necessary.
  */
 void gfs_simulation_event (GfsSimulation * sim,
-			  GSList * events)
+			   GSList * events)
 {
   g_return_if_fail (sim != NULL);
 
@@ -672,6 +961,30 @@ void gfs_simulation_event (GfsSimulation * sim,
   }
 }
 
+/**
+ * gfs_simulation_event_half:
+ * @sim: a #GfsSimulation.
+ * @events: a list of #GfsEvent.
+ * 
+ * Checks if any half-timestep event associated with @sim must be
+ * activated and activates it if necessary.
+ */
+void gfs_simulation_event_half (GfsSimulation * sim,
+				GSList * events)
+{
+  g_return_if_fail (sim != NULL);
+
+  while (events) {
+    GfsEvent * event = events->data;
+    GSList * next = events->next;
+
+    if (event->realised && GFS_EVENT_CLASS (GTS_OBJECT (event)->klass)->event_half)
+      (* GFS_EVENT_CLASS (GTS_OBJECT (event)->klass)->event_half) (event, sim);
+    
+    events = next;
+  }
+}
+
 /**
  * gfs_simulation_event_init:
  * @sim: a #GfsSimulation.
@@ -681,7 +994,7 @@ void gfs_simulation_event (GfsSimulation * sim,
  * "init" events are activated by this function.
  */
 void gfs_simulation_event_init (GfsSimulation * sim,
-			       GSList * events)
+				GSList * events)
 {
   g_return_if_fail (sim != NULL);
 
@@ -699,16 +1012,8 @@ void gfs_simulation_event_init (GfsSimulation * sim,
       g_assert (GFS_EVENT_CLASS (GTS_OBJECT (event)->klass)->event);
       (* GFS_EVENT_CLASS (GTS_OBJECT (event)->klass)->event) (event, sim);
     }
-    else if (event->end_event) {
-      if (sim->time.iend < G_MAXINT)
-	event->i = event->istart = sim->time.iend;
-      else
-	event->i = event->istart = G_MAXINT/2.;
-      if (sim->time.end < G_MAXDOUBLE)
-	event->t = event->start = sim->time.end;
-      else
-	event->t = event->start = G_MAXDOUBLE/2.;
-    }
+    else if (event->end_event)
+      event->t = event->start = G_MAXDOUBLE/2.;
     else {
       if (event->istep < G_MAXINT)
 	while (event->i < sim->time.i) {
@@ -746,6 +1051,8 @@ void gfs_time_write (GfsTime * t, FILE * fp)
     fprintf (fp, "end = %g ", t->end);
   if (t->iend < G_MAXINT)
     fprintf (fp, "iend = %u ", t->iend);
+  if (t->dtmax < G_MAXDOUBLE)
+    fprintf (fp, "dtmax = %g ", t->dtmax);
   fputc ('}', fp);
 }
 
@@ -764,6 +1071,8 @@ void gfs_time_init (GfsTime * t)
 
   t->i = t->istart = 0;
   t->iend = G_MAXINT;
+
+  t->dtmax = G_MAXDOUBLE;
 }
 
 /**
@@ -782,6 +1091,7 @@ void gfs_time_read (GfsTime * t, GtsFile * fp)
     {GTS_UINT,   "i",      TRUE},
     {GTS_UINT,   "istart", TRUE},
     {GTS_UINT,   "iend",   TRUE},
+    {GTS_DOUBLE, "dtmax",  TRUE},
     {GTS_NONE}
   };
 
@@ -794,6 +1104,7 @@ void gfs_time_read (GfsTime * t, GtsFile * fp)
   var[3].data = &t->i;
   var[4].data = &t->istart;
   var[5].data = &t->iend;
+  var[6].data = &t->dtmax;
 
   gfs_time_init (t);
   gts_file_assign_variables (fp, var);
@@ -817,7 +1128,7 @@ void gfs_physical_params_write (GfsPhysicalParams * p, FILE * fp)
   g_return_if_fail (p != NULL);
   g_return_if_fail (fp != NULL);
 
-  fprintf (fp, "{ rho = %g }", p->rho);
+  fprintf (fp, "{ rho = %g sigma = %g g = %g }", p->rho, p->sigma, p->g);
 }
 
 /**
@@ -831,6 +1142,8 @@ void gfs_physical_params_init (GfsPhysicalParams * p)
   g_return_if_fail (p != NULL);
   
   p->rho = 1.;
+  p->sigma = 0.;
+  p->g = 1.;
 }
 
 /**
@@ -843,7 +1156,9 @@ void gfs_physical_params_init (GfsPhysicalParams * p)
 void gfs_physical_params_read (GfsPhysicalParams * p, GtsFile * fp)
 {
   GtsFileVariable var[] = {
-    {GTS_DOUBLE, "rho", TRUE},
+    {GTS_DOUBLE, "rho",   TRUE},
+    {GTS_DOUBLE, "sigma", TRUE},
+    {GTS_DOUBLE, "g",     TRUE},
     {GTS_NONE}
   };
 
@@ -851,12 +1166,15 @@ void gfs_physical_params_read (GfsPhysicalParams * p, GtsFile * fp)
   g_return_if_fail (fp != NULL);
 
   var[0].data = &p->rho;
+  var[1].data = &p->sigma;
+  var[2].data = &p->g;
 
   gfs_physical_params_init (p);
   gts_file_assign_variables (fp, var);
   if (p->rho <= 0.)
-    gts_file_variable_error (fp, var, "rho", 
-			     "rho must be strictly positive");
+    gts_file_variable_error (fp, var, "rho", "rho must be strictly positive");
+  if (p->sigma < 0.)
+    gts_file_variable_error (fp, var, "sigma", "sigma must be positive");
 }
 
 /**
@@ -867,82 +1185,7 @@ void gfs_physical_params_read (GfsPhysicalParams * p, GtsFile * fp)
  */
 void gfs_simulation_run (GfsSimulation * sim)
 {
-  GfsVariable * v, * c, * ch;
-  GfsDomain * domain;
-
   g_return_if_fail (sim != NULL);
 
-  domain = GFS_DOMAIN (sim);
-
-  gfs_simulation_refine (sim);
-
-  gfs_simulation_event_init (sim, sim->events->items);
-  gfs_simulation_event_init (sim, sim->adapts->items);
-
-  gfs_set_merged (domain);
-  v = domain->variables;
-  while (v) {
-    gfs_domain_center_bc (domain, v);
-    v = v->next;
-  }
-  c = sim->tracer_advection_params.v = 
-    gfs_variable_from_name (domain->variables, "C");
-  ch = gfs_variable_from_name (domain->variables, "CH");
-
-  sim->advection_params.c = c;
-  gfs_approximate_projection (domain, 
-			      &sim->approx_projection_params,
-			      &sim->advection_params);
-
-  gts_range_init (&domain->mpi_wait);
-  while (sim->time.t < sim->time.end &&
-	 sim->time.i < sim->time.iend) {
-    v = domain->variables;
-    while (v) {
-      gfs_domain_cell_traverse (domain,
-				FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-				(FttCellTraverseFunc) 
-				gfs_get_from_below_intensive, v);
-      v = v->next;
-    }
-    gfs_simulation_event (sim, sim->events->items);
-
-    g_timer_start (domain->timer);
-
-    gfs_simulation_set_timestep (sim);
-
-    sim->advection_params.c = c;
-    gfs_predicted_face_velocities (domain, &sim->advection_params);
-    gfs_mac_projection (domain, 
-			&sim->projection_params, 
-			&sim->advection_params);
-#if 1
-    gfs_tracer_advection (domain, &sim->tracer_advection_params, ch);
-#else
-    gfs_tracer_vof_advection (domain, &sim->tracer_advection_params, ch);
-#endif
-    gfs_variable_sources (domain, c, &sim->tracer_advection_params);
-
-    sim->advection_params.c = ch;
-    gfs_centered_velocity_advection (domain, &sim->advection_params);
-    gfs_velocity_sources (domain, &sim->advection_params);
-    gfs_simulation_adapt (sim);
-    gfs_approximate_projection (domain, 
-				&sim->approx_projection_params, 
-				&sim->advection_params);
-
-    sim->time.t = sim->tnext;
-    sim->time.i++;
-
-    g_timer_stop (domain->timer);
-    gts_range_add_value (&domain->timestep,
-			 g_timer_elapsed (domain->timer, NULL));
-    gts_range_update (&domain->timestep);
-    gts_range_add_value (&domain->size, 
-			 gfs_domain_size (domain, FTT_TRAVERSE_LEAFS, -1));
-    gts_range_update (&domain->size);
-  }
-  gfs_simulation_event (sim, sim->events->items);
-  gts_container_foreach (GTS_CONTAINER (sim->events),
-			 (GtsFunc) gts_object_destroy, NULL);
+  (* GFS_SIMULATION_CLASS (GTS_OBJECT (sim)->klass)->run) (sim);
 }
