@@ -20,6 +20,7 @@
 #include <math.h>
 #include <string.h>
 #include "solid.h"
+#include "vof.h"
 
 /* GfsFace: Header */
 
@@ -153,16 +154,26 @@ static void surface_add_box (GtsSurface * s,
   gts_surface_add_face (s, GTS_FACE (gfs_face_new (klass, e18, e14, e6, 3)));
 }
 
+static void cell_size (FttCell * cell, GtsVector s)
+{
+#if FTT_2D3
+  s[0] = s[1] = ftt_cell_size (cell);
+  s[2] = 1.;
+#else  /* 2D or 3D */
+  s[0] = s[1] = s[2] = ftt_cell_size (cell);
+#endif /* 2D or 3D */
+}
+
 static GtsBBox * bbox_cell (GtsBBoxClass * klass, FttCell * cell)
 {
   FttVector p;
-  gdouble size;
+  GtsVector size;
 
   ftt_cell_pos (cell, &p);
-  size = ftt_cell_size (cell)/2.;
+  cell_size (cell, size);
   return gts_bbox_new (klass, cell, 
-		       p.x - size, p.y - size, p.z - size,
-		       p.x + size, p.y + size, p.z + size);
+		       p.x - size[0]/2., p.y - size[1]/2., p.z - size[2]/2.,
+		       p.x + size[0]/2., p.y + size[1]/2., p.z + size[2]/2.);
 }
 
 /**
@@ -215,7 +226,8 @@ static void set_solid_or_fluid (FttCell * cell,
 				GNode * tree,
 				gboolean is_open,
 				gboolean destroy_solid,
-				FttCellCleanupFunc cleanup)
+				FttCellCleanupFunc cleanup,
+				gpointer data)
 {
   FttVector pos;
   GtsPoint * p;
@@ -225,8 +237,14 @@ static void set_solid_or_fluid (FttCell * cell,
   if (gts_point_is_inside_surface (p, tree, is_open))
     gfs_cell_fluid (cell);
   else {
-    if (destroy_solid)
-      ftt_cell_destroy (cell, cleanup);
+    if (destroy_solid) {
+      if (FTT_CELL_IS_ROOT (cell))
+	g_log (G_LOG_DOMAIN,
+	       G_LOG_LEVEL_ERROR,
+	       "root cell is entirely outside of the fluid domain\n"
+	       "the solid surface orientation may be incorrect\n");
+      ftt_cell_destroy (cell, cleanup, data);
+    }
     else
       gfs_cell_solid (cell);
   }
@@ -241,20 +259,63 @@ static void add_surface_fraction (GfsFace * f, GfsSolidVector * solid)
 
 static guint warning_number = 0;
 
+static void write_surface_warning (GtsSurfaceInter * si,
+				   GtsSurface * cs,
+				   GtsSurface * inter)
+{
+  gchar s[80];
+  FILE * fp;
+  
+  sprintf (s, "/tmp/gerris_warning.%d", warning_number++);
+  fp = fopen (s, "wt");
+  if (fp != NULL) {
+#if 1
+    gts_surface_write (cs, fp);
+#else
+    GSList * i = si->edges;
+    
+    fputs ("(geometry \"cell\" = {\n", fp);
+    gts_surface_write_oogl (cs, fp);
+    fputs ("})\n(geometry \"inter\" = {\n", fp);
+    gts_surface_write_oogl (inter, fp);
+    fputs ("})\n(geometry \"curve\" = LIST {\n", fp);
+    while (i) {
+      GtsPoint * p1 = GTS_POINT (GTS_SEGMENT (i->data)->v1);
+      GtsPoint * p2 = GTS_POINT (GTS_SEGMENT (i->data)->v2);
+      
+      fprintf (fp, "VECT 1 2 0 2 0 %g %g %g %g %g %g\n",
+	       p1->x, p1->y, p1->z, p2->x, p2->y, p2->z);
+      i = i->next;
+    }
+    fputs ("})\n", fp);
+#endif
+    fclose (fp);
+  }
+}
+
+static void bbox_size (GtsBBox * bbox, GtsVector s)
+{
+  s[0] = bbox->x2 - bbox->x1;
+  s[1] = bbox->y2 - bbox->y1;
+  s[2] = bbox->z2 - bbox->z1;
+}
+
 static void set_solid_fractions_from_surface (FttCell * cell,
 					      GtsBBox * bbox,
 					      GtsSurface * s,
 					      GNode * stree,
 					      gboolean is_open,
 					      gboolean destroy_solid,
-					      FttCellCleanupFunc cleanup)
+					      FttCellCleanupFunc cleanup,
+					      gpointer data)
 {
   GtsSurface * cs; 
   GNode * ctree;
   GtsSurfaceInter * si;
-  gdouble size = bbox->x2 - bbox->x1;
+  GtsVector size;
   gboolean closed = TRUE;
 
+  bbox_size (bbox, size);
   cs = gts_surface_new (gts_surface_class (),
 			GTS_FACE_CLASS (gfs_face_class ()),
 			gts_edge_class (),
@@ -287,13 +348,17 @@ static void set_solid_fractions_from_surface (FttCell * cell,
   }
 #endif
   if (si->edges == NULL)
-    set_solid_or_fluid (cell, stree, is_open, destroy_solid, cleanup);
+    set_solid_or_fluid (cell, stree, is_open, destroy_solid, cleanup, data);
   else {
     GtsSurface * inter = gts_surface_new (gts_surface_class (),
 					  gts_face_class (),
 					  gts_edge_class (),
 					  gts_vertex_class ());
-    gdouble size2 = size*size;
+    GtsSurface * inter1 = gts_surface_new (gts_surface_class (),
+					   gts_face_class (),
+					   gts_edge_class (),
+					   gts_vertex_class ());
+    gdouble size2[3] = { size[1]*size[2], size[0]*size[2], size[0]*size[1] };
     guint i;
     GfsSolidVector * solid = GFS_STATE (cell)->solid;
 
@@ -316,44 +381,22 @@ static void set_solid_fractions_from_surface (FttCell * cell,
       solid->s[i] = 0.;
     gts_surface_foreach_face (inter, (GtsFunc) add_surface_fraction, solid);
     for (i = 0; i < FTT_NEIGHBORS; i++) {
-      solid->s[i] /= size2;
+      solid->s[i] /= size2[i/2];
 #if 0
       g_assert (solid->s[i] >= 0. && solid->s[i] <= 1.);
 #else
       if (solid->s[i] < 0. || solid->s[i] > 1.) {
 	FttVector p;
-	gchar s[80];
-	FILE * fp;
-	
-	sprintf (s, "/tmp/gerris_warning.%d", warning_number++);
-	fp = fopen (s, "wt");
-	if (fp != NULL) {
-	  GSList * i = si->edges;
-
-	  fputs ("(geometry \"cell\" = {\n", fp);
-	  gts_surface_write_oogl (cs, fp);
-	  fputs ("})\n(geometry \"inter\" = {\n", fp);
-	  gts_surface_write_oogl (inter, fp);
-	  fputs ("})\n(geometry \"curve\" = LIST {\n", fp);
-	  while (i) {
-	    GtsPoint * p1 = GTS_POINT (GTS_SEGMENT (i->data)->v1);
-	    GtsPoint * p2 = GTS_POINT (GTS_SEGMENT (i->data)->v2);
-
-	    fprintf (fp, "VECT 1 2 0 2 0 %g %g %g %g %g %g\n",
-		     p1->x, p1->y, p1->z, p2->x, p2->y, p2->z);
-	    i = i->next;
-	  }
-	  fputs ("})\n", fp);
-	  fclose (fp);
-	}
 	
+	write_surface_warning (si, cs, inter);
 	ftt_cell_pos (cell, &p);
-	g_warning ("file %s: line %d (%s): (%g,%g,%g): level %d: solid->s[%d] = %g. Surfaces have been written in %s.",
+	g_warning ("file %s: line %d (%s): (%g,%g,%g): level %d: solid->s[%d] = %g. "
+		   "Surfaces have been written in /tmp/gerris_warning.%d.",
 		   __FILE__, __LINE__, G_GNUC_PRETTY_FUNCTION,
 		   p.x, p.y, p.z,
 		   ftt_cell_level (cell),
 		   i, solid->s[i],
-		   fp != NULL ? s : "(null)");
+		   warning_number - 1);
 	if (solid->s[i] > 1.)
 	  solid->s[i] = 1.;
 	else if (solid->s[i] < 0.)
@@ -362,46 +405,26 @@ static void set_solid_fractions_from_surface (FttCell * cell,
 #endif
     }
 
-    gts_surface_inter_boolean (si, inter, GTS_2_IN_1);
-    solid->a = gts_surface_volume (inter)/(size2*size);
+    gts_surface_inter_boolean (si, inter1, GTS_2_IN_1);
+    gts_surface_merge (inter, inter1);
+    solid->a = gts_surface_center_of_mass (inter, &solid->cm.x)/(size[0]*size[1]*size[2]);
+    gts_surface_center_of_area (inter1, &solid->ca.x);
 
 #if 0
     g_assert (solid->a > 0. && solid->a < 1.);
 #else
     if (solid->a <= 0. || solid->a >= 1.) {
       FttVector p;
-      gchar s[80];
-      FILE * fp;
-
-      sprintf (s, "/tmp/gerris_warning.%d", warning_number++);
-      fp = fopen (s, "wt");
-      if (fp != NULL) {
-	GSList * i = si->edges;
-
-	fputs ("(geometry \"cell\" = {\n", fp);
-	gts_surface_write_oogl (cs, fp);
-	fputs ("})\n(geometry \"inter\" = {\n", fp);
-	gts_surface_write_oogl (inter, fp);
-	fputs ("})\n(geometry \"curve\" = LIST {\n", fp);
-	while (i) {
-	  GtsPoint * p1 = GTS_POINT (GTS_SEGMENT (i->data)->v1);
-	  GtsPoint * p2 = GTS_POINT (GTS_SEGMENT (i->data)->v2);
-	  
-	  fprintf (fp, "VECT 1 2 0 2 0 %g %g %g %g %g %g\n",
-		   p1->x, p1->y, p1->z, p2->x, p2->y, p2->z);
-	  i = i->next;
-	}
-	fputs ("})\n", fp);
-	fclose (fp);
-      }
 
+      write_surface_warning (si, cs, inter);
       ftt_cell_pos (cell, &p);
-      g_warning ("file %s: line %d (%s): (%g,%g,%g): level %d: solid->a = %g. Surfaces have been written in %s.",
+      g_warning ("file %s: line %d (%s): (%g,%g,%g): level %d: solid->a = %g. "
+		 "Surfaces have been written in /tmp/gerris_warning.%d.",
 		 __FILE__, __LINE__, G_GNUC_PRETTY_FUNCTION,
 		 p.x, p.y, p.z,
 		 ftt_cell_level (cell),
 		 solid->a,
-		 fp != NULL ? s : "(null)");
+		 warning_number - 1);
       if (solid->a > 1.)
 	solid->a = 1.;
       else if (solid->a < 0.)
@@ -410,6 +433,7 @@ static void set_solid_fractions_from_surface (FttCell * cell,
 #endif
 
     gts_object_destroy (GTS_OBJECT (inter));
+    gts_object_destroy (GTS_OBJECT (inter1));
   }
 
   gts_object_destroy (GTS_OBJECT (si));
@@ -417,6 +441,19 @@ static void set_solid_fractions_from_surface (FttCell * cell,
   gts_object_destroy (GTS_OBJECT (cs));
 }
 
+static gdouble solid_sa (GfsSolidVector * s)
+{
+  gdouble sa2 = 0.;
+  FttComponent c;
+
+  for (c = 0; c < FTT_DIMENSION; c++) {
+    gdouble n = s->s[2*c] - s->s[2*c + 1];
+
+    sa2 += n*n;
+  }
+  return sqrt (sa2);
+}
+
 /**
  * gfs_cell_init_solid_fractions_from_children:
  * @cell: a #FttCell.
@@ -430,23 +467,11 @@ void gfs_cell_init_solid_fractions_from_children (FttCell * cell)
 {
   FttCellChildren child;
   guint i, j;
-  gdouble w = 0.;
+  gdouble w = 0., wa = 0.;
   gboolean cell_is_solid = TRUE;
   gboolean cell_is_mixed = FALSE;
-  static guint index[FTT_NEIGHBORS][FTT_CELLS/2] =
-#ifdef FTT_2D
-    {{1, 3},
-     {0, 2},
-     {0, 1},
-     {2, 3}};
-#else  /* FTT_3D */
-    {{1, 3, 5, 7},
-     {0, 2, 4, 6},
-     {0, 1, 4, 5},
-     {2, 3, 6, 7},
-     {0, 1, 2, 3},
-     {4, 5, 6, 7}};
-#endif /* FTT_3D */
+  FttVector cm = { 0., 0., 0.};
+  FttVector ca = { 0., 0., 0.};
 
   g_return_if_fail (cell != NULL);
   g_return_if_fail (!FTT_CELL_IS_LEAF (cell));
@@ -455,11 +480,24 @@ void gfs_cell_init_solid_fractions_from_children (FttCell * cell)
   for (i = 0; i < FTT_CELLS; i++)
     if (child.c[i]) {
       if (GFS_IS_FLUID (child.c[i])) {
+	FttVector p;
+
 	w += 1.;
+	ftt_cell_pos (child.c[i], &p);
+	cm.x += p.x; cm.y += p.y; cm.z += p.z;
 	cell_is_solid = FALSE;
       }
       else {
-	w += GFS_STATE (child.c[i])->solid->a;
+	GfsSolidVector * solid = GFS_STATE (child.c[i])->solid;
+	gdouble sa = solid_sa (solid);
+
+	w += solid->a; wa += sa;
+	cm.x += solid->cm.x*solid->a;
+	cm.y += solid->cm.y*solid->a;
+	cm.z += solid->cm.z*solid->a;
+	ca.x += solid->ca.x*sa;
+	ca.y += solid->ca.y*sa;
+	ca.z += solid->ca.z*sa;
 	cell_is_mixed = TRUE;
       }
     }
@@ -468,14 +506,21 @@ void gfs_cell_init_solid_fractions_from_children (FttCell * cell)
     GfsSolidVector * solid = GFS_STATE (cell)->solid;
 
     if (solid == NULL)
-      GFS_STATE (cell)->solid = solid = 
-	g_malloc0 (sizeof (GfsSolidVector));
+      GFS_STATE (cell)->solid = solid = g_malloc0 (sizeof (GfsSolidVector));
 
-    solid->a = w/FTT_CELLS;
+    solid->a = w/FTT_CELLS; 
+    solid->cm.x = cm.x/w;
+    solid->cm.y = cm.y/w;
+    solid->cm.z = cm.z/w;
+    solid->ca.x = ca.x/wa;
+    solid->ca.y = ca.y/wa;
+    solid->ca.z = ca.z/wa;
     for (i = 0; i < FTT_NEIGHBORS; i++) {
+      guint n = ftt_cell_children_direction (cell, i, &child);
+
       w = 0.;
-      for (j = 0; j < FTT_CELLS/2; j++) {
-	FttCell * c = child.c[index[i][j]];
+      for (j = 0; j < n; j++) {
+	FttCell * c = child.c[j];
 
 	if (c) {
 	  if (GFS_IS_FLUID (c))
@@ -484,7 +529,7 @@ void gfs_cell_init_solid_fractions_from_children (FttCell * cell)
 	    w += GFS_STATE (c)->solid->s[i];
 	}
       }
-      solid->s[i] = w/(FTT_CELLS/2);
+      solid->s[i] = w/n;
     }
   }
   else { /* !cell_is_mixed */
@@ -505,6 +550,7 @@ void gfs_cell_init_solid_fractions_from_children (FttCell * cell)
  * enclosed by @s is negative, %FALSE otherwise.
  * @destroy_solid: controls what to do with solid cells.
  * @cleanup: a #FttCellCleanupFunc or %NULL.
+ * @data: user data to pass to @cleanup.
  *
  * Initializes the solid fractions of all the cells of the cell tree
  * starting at @root.
@@ -513,11 +559,12 @@ void gfs_cell_init_solid_fractions_from_children (FttCell * cell)
  * the solid are destroyed using @cleanup as cleanup function.  
  */
 void gfs_cell_init_solid_fractions (FttCell * root, 
-				   GtsSurface * s,
-				   GNode * stree,
-				   gboolean is_open,
-				   gboolean destroy_solid,
-				   FttCellCleanupFunc cleanup)
+				    GtsSurface * s,
+				    GNode * stree,
+				    gboolean is_open,
+				    gboolean destroy_solid,
+				    FttCellCleanupFunc cleanup,
+				    gpointer data)
 {
   GtsBBox * bbox;
 
@@ -526,11 +573,13 @@ void gfs_cell_init_solid_fractions (FttCell * root,
   g_return_if_fail (stree != NULL);
 
   bbox = bbox_cell (gts_bbox_class (), root);
-  if (gts_bb_tree_is_overlapping (stree, bbox)) {
+  if (!gts_bb_tree_is_overlapping (stree, bbox))
+    set_solid_or_fluid (root, stree, is_open, destroy_solid, cleanup, data);
+  else {
     if (FTT_CELL_IS_LEAF (root))
       set_solid_fractions_from_surface (root, bbox, 
 					s, stree, is_open,
-					destroy_solid, cleanup);
+					destroy_solid, cleanup, data);
     else { /* !FTT_CELL_IS_LEAF (root) */
       FttCellChildren child;
       guint i;
@@ -540,16 +589,20 @@ void gfs_cell_init_solid_fractions (FttCell * root,
 	if (child.c[i])
 	  gfs_cell_init_solid_fractions (child.c[i],
 					s, stree, is_open,
-					destroy_solid, cleanup);
-      if (FTT_CELL_IS_LEAF (root))
+					destroy_solid, cleanup, data);
+      if (FTT_CELL_IS_LEAF (root)) {
 	/* all the children have been destroyed i.e. the cell is solid */
-	ftt_cell_destroy (root, cleanup);
+	if (FTT_CELL_IS_ROOT (root))
+	  g_log (G_LOG_DOMAIN,
+		 G_LOG_LEVEL_ERROR,
+		 "root cell is entirely outside of the fluid domain\n"
+		 "the solid surface orientation may be incorrect\n");
+	ftt_cell_destroy (root, cleanup, data);
+      }
       else
 	gfs_cell_init_solid_fractions_from_children (root);
     }
   }
-  else /* !gts_bb_tree_is_overlapping (stree, bbox) */
-    set_solid_or_fluid (root, stree, is_open, destroy_solid, cleanup);
 
   gts_object_destroy (GTS_OBJECT (bbox));
 }
@@ -687,7 +740,7 @@ gboolean gfs_cell_check_solid_fractions (FttCell * root,
   g_return_val_if_fail (root != NULL, FALSE);
   g_return_val_if_fail (solid == NULL || stree != NULL, FALSE);
 
-#ifndef FTT_2D 
+#if (!FTT_2D)
   if (solid) {
     /* check that the total volume of the union of the solid with the
        domain is equal to the volume fraction of the root cell */
@@ -695,7 +748,7 @@ gboolean gfs_cell_check_solid_fractions (FttCell * root,
     GNode * dtree; 
     GtsSurfaceInter * si;
     FttVector p;
-    gdouble size = ftt_cell_size (root)/2.;
+    GtsVector size;
     gboolean closed = TRUE;
 
     domain = gts_surface_new (gts_surface_class (),
@@ -703,9 +756,10 @@ gboolean gfs_cell_check_solid_fractions (FttCell * root,
 			      gts_edge_class (),
 			      gts_vertex_class ());
     ftt_cell_pos (root, &p);
+    cell_size (root, size);
     surface_add_box (domain, 
-		     p.x - size, p.y - size, p.z - size, 
-		     p.x + size, p.y + size, p.z + size);
+		     p.x - size[0]/2., p.y - size[1]/2., p.z - size[2]/2., 
+		     p.x + size[0]/2., p.y + size[1]/2., p.z + size[2]/2.);
     dtree = gts_bb_tree_surface (domain);
     si = gts_surface_inter_new (gts_surface_inter_class (),
 				domain, solid, dtree, stree, FALSE, is_open);
@@ -757,7 +811,8 @@ gboolean gfs_cell_check_solid_fractions (FttCell * root,
 	    fclose (fp);
 	  }
 	  
-	  g_warning ("file %s: line %d (%s): solid->a: %g volume: %g. Surface has been written in %s.",
+	  g_warning ("file %s: line %d (%s): solid->a: %g volume: %g. "
+		     "Surface has been written in %s.",
 		     __FILE__, __LINE__, G_GNUC_PRETTY_FUNCTION,
 		     GFS_STATE (root)->solid->a, volume,
 		     fp != NULL ? s : "(null)");
@@ -848,7 +903,6 @@ static void set_fraction_from_surface (FttCell * cell,
   GtsSurface * cs; 
   GNode * ctree;
   GtsSurfaceInter * si;
-  gdouble size = bbox->x2 - bbox->x1;
   gboolean closed = TRUE;
 
   cs = gts_surface_new (gts_surface_class (),
@@ -869,12 +923,13 @@ static void set_fraction_from_surface (FttCell * cell,
 					  gts_face_class (),
 					  gts_edge_class (),
 					  gts_vertex_class ());
-    gdouble size2 = size*size;
+    GtsVector size;
 
     g_assert (closed);
     gts_surface_inter_boolean (si, inter, GTS_1_IN_2);
     gts_surface_inter_boolean (si, inter, GTS_2_IN_1);
-    GFS_VARIABLE (cell, c->i) = gts_surface_volume (inter)/(size2*size);
+    bbox_size (bbox, size);
+    GFS_VARIABLE (cell, c->i) = gts_surface_volume (inter)/(size[0]*size[1]*size[2]);
     g_assert (GFS_VARIABLE (cell, c->i) > -1e-9 && 
 	      GFS_VARIABLE (cell, c->i) < 1. + 1e-9);
     gts_object_destroy (GTS_OBJECT (inter));
@@ -931,3 +986,73 @@ void gfs_cell_init_fraction (FttCell * root,
 
   gts_object_destroy (GTS_OBJECT (bbox));
 }
+
+/**
+ * gfs_cell_cm:
+ * @cell: a #FttCell.
+ * @cm: a #FttVector.
+ *
+ * Fills @cm with the coordinates of the center of mass of @cell.
+ */
+void gfs_cell_cm (const FttCell * cell, FttVector * cm)
+{
+  g_return_if_fail (cell != NULL);
+  g_return_if_fail (cm != NULL);
+
+  if (GFS_IS_MIXED (cell))
+    *cm = GFS_STATE (cell)->solid->cm;
+  else
+    ftt_cell_pos (cell, cm);
+}
+
+/**
+ * gfs_face_ca:
+ * @face: a #FttCellFace.
+ * @ca: a #FttVector.
+ *
+ * Fills @ca with the coordinates of the center of area of @face.
+ */
+void gfs_face_ca (const FttCellFace * face, FttVector * ca)
+{
+  gdouble f;
+
+  g_return_if_fail (face != NULL);
+  g_return_if_fail (ca != NULL);
+
+  ftt_face_pos (face, ca);
+  if ((f = GFS_FACE_FRACTION (face)) < 1.) {
+    GfsSolidVector * s = GFS_STATE (face->cell)->solid;
+    gdouble h = ftt_cell_size (face->cell);
+#if FTT_2D
+    FttComponent cp = FTT_ORTHOGONAL_COMPONENT (face->d/2);
+
+    (&ca->x)[cp] += (s->s[2*cp] > s->s[2*cp + 1]) ? (1. - f)/2.*h : (f - 1.)/2.*h;
+#else /* 3D */
+    static guint perpendicular[FTT_DIMENSION][2] = {
+      {FTT_Y, FTT_Z}, {FTT_Z, FTT_X}, {FTT_X, FTT_Y}
+    };
+    FttComponent c0 = face->d/2;
+    FttComponent c1 = perpendicular[c0][0];
+    FttComponent c2 = perpendicular[c0][1];
+    gboolean s1, s2;
+    FttVector m, p;
+    gdouble n, alpha;
+
+    m.x = s->s[2*c1 + 1] - s->s[2*c1];
+    m.y = s->s[2*c2 + 1] - s->s[2*c2];
+    s1 = (m.x < 0.);
+    s2 = (m.y < 0.);
+    m.x = fabs (m.x) + 1e-6;
+    m.y = fabs (m.y) + 1e-6;
+    n = m.x + m.y;
+    m.x /= n;
+    m.y /= n;
+    alpha = gfs_line_alpha (&m, f);
+    gfs_line_center (&m, alpha, f, &p);
+    if (s1) p.x = 1. - p.x;
+    if (s2) p.y = 1. - p.y;
+    (&ca->x)[c1] += (p.x - 0.5)*h;
+    (&ca->x)[c2] += (p.y - 0.5)*h;
+#endif /* 3D */
+  }
+}
