@@ -26,11 +26,25 @@ extern "C" {
 
 #include "advection.h"
 
+gdouble gfs_line_area              (FttVector * m, 
+				    gdouble alpha, 
+				    gdouble c1);
+void    gfs_line_center            (FttVector * m, 
+				    gdouble alpha, 
+				    gdouble c, 
+				    FttVector * p);
+gdouble gfs_line_alpha             (FttVector * m, 
+				    gdouble c);
+#if FTT_2D
+#  define gfs_plane_volume         gfs_line_area
+#  define gfs_plane_alpha          gfs_line_alpha
+#else /* 3D */
 gdouble gfs_plane_volume           (FttVector * m, 
 				    gdouble alpha, 
 				    gdouble c1);
 gdouble gfs_plane_alpha            (FttVector * m, 
 				    gdouble c);
+#endif /* 3D */
 void    gfs_cell_vof_advection     (FttCell * cell,
 				    FttComponent c,
 				    GfsAdvectionParams * par);
