@@ -29,7 +29,7 @@ for file in $1; do
     else
         all=`expr $all + 1`;
 	failed=`expr $failed + 1`;
-	echo "FAIL: $sname [$avg]";
+	echo "FAIL: $sname [$avg] [$avgref]";
     fi
     expr `date +%s` - $startdate > timestamp
 done
