@@ -33,28 +33,124 @@
 #include "poisson.h"
 #include "init.h"
 
-#ifndef PI
-# define PI 3.14159265359
-#endif
-
 static gdouble K = 3.;
 static gdouble L = 3.;
 #ifndef   FTT_2D
 static gdouble M = 3.;
 #endif /* not FTT_2D */
 
+static void fix_volume (FttCell * cell)
+{
+  if (GFS_IS_MIXED (cell)) {
+    GfsSolidVector * s = GFS_STATE (cell)->solid;
+    FttVector p;
+    gdouble h = ftt_cell_size (cell);
+    
+    ftt_cell_pos (cell, &p);
+    if (s->s[0] > 0. && s->s[0] < 1. &&
+	s->s[1] > 0. && s->s[1] < 1.) {
+      s->ca.x = p.x;
+      s->a = (s->s[0] + s->s[1])/2.;
+      if (s->s[3] == 1.) {
+	g_assert (s->s[2] == 0.);
+	s->ca.y = p.y + h*(s->a - 0.5);
+      }
+      else {
+	g_assert (s->s[2] == 1. && s->s[3] == 0.);
+	s->ca.y = p.y - h*(s->a - 0.5);	
+      }
+    }
+    else if (s->s[2] > 0. && s->s[2] < 1. &&
+	     s->s[3] > 0. && s->s[3] < 1.) {
+      s->ca.y = p.y;
+      s->a = (s->s[2] + s->s[3])/2.;
+      if (s->s[1] == 1.) {
+	g_assert (s->s[0] == 0.);
+	s->ca.x = p.x + h*(s->a - 0.5);
+      }
+      else {
+	g_assert (s->s[0] == 1. && s->s[1] == 0.);
+	s->ca.x = p.x - h*(s->a - 0.5);	
+      }
+    }
+    else if (s->s[0] > 0. && s->s[0] < 1. &&
+	     s->s[2] > 0. && s->s[2] < 1.) {
+      if (s->s[1] == 1.) {
+	g_assert (s->s[3] == 1.);
+	s->a = 1. - (1. - s->s[0])*(1. - s->s[2])/2.;
+	s->ca.x = p.x + h*s->s[2]/2.;
+	s->ca.y = p.y + h*s->s[0]/2.;
+      }
+      else {
+	g_assert (s->s[3] == 0. && s->s[1] == 0.);
+	s->a = s->s[0]*s->s[2]/2.;
+      	s->ca.x = p.x + h*(1. - s->s[2])/2.;
+	s->ca.y = p.y + h*(1. - s->s[0])/2.;
+      }
+    }
+    else if (s->s[0] > 0. && s->s[0] < 1. &&
+	     s->s[3] > 0. && s->s[3] < 1.) {
+      if (s->s[1] == 1.) {
+	g_assert (s->s[2] == 1.);
+	s->a = 1. - (1. - s->s[0])*(1. - s->s[3])/2.;
+	s->ca.x = p.x + h*s->s[3]/2.;
+	s->ca.y = p.y - h*s->s[0]/2.;
+      }
+      else {
+	g_assert (s->s[2] == 0. && s->s[1] == 0.);
+	s->a = s->s[0]*s->s[3]/2.;
+      	s->ca.x = p.x + h*(1. - s->s[3])/2.;
+	s->ca.y = p.y - h*(1. - s->s[0])/2.;
+      }
+    }
+
+    else if (s->s[1] > 0. && s->s[1] < 1. &&
+	     s->s[2] > 0. && s->s[2] < 1.) {
+      if (s->s[0] == 1.) {
+	g_assert (s->s[3] == 1.);
+	s->a = 1. - (1. - s->s[1])*(1. - s->s[2])/2.;
+	s->ca.x = p.x - h*s->s[2]/2.;
+	s->ca.y = p.y + h*s->s[1]/2.;
+      }
+      else {
+	g_assert (s->s[3] == 0. && s->s[0] == 0.);
+	s->a = s->s[1]*s->s[2]/2.;
+      	s->ca.x = p.x - h*(1. - s->s[2])/2.;
+	s->ca.y = p.y + h*(1. - s->s[1])/2.;
+      }
+    }
+    else if (s->s[1] > 0. && s->s[1] < 1. &&
+	     s->s[3] > 0. && s->s[3] < 1.) {
+      if (s->s[0] == 1.) {
+	g_assert (s->s[2] == 1.);
+	s->a = 1. - (1. - s->s[1])*(1. - s->s[3])/2.;
+	s->ca.x = p.x - h*s->s[3]/2.;
+	s->ca.y = p.y - h*s->s[1]/2.;
+      }
+      else {
+	g_assert (s->s[2] == 0. && s->s[0] == 0.);
+	s->a = s->s[1]*s->s[3]/2.;
+      	s->ca.x = p.x - h*(1. - s->s[3])/2.;
+	s->ca.y = p.y - h*(1. - s->s[1])/2.;
+      }
+    }
+    else
+      g_assert_not_reached ();
+  }
+}
+
 static void init_sine_div (FttCell * cell)
 {
-  FttVector pos;
+  FttVector cm;
   gdouble size;
 
-  ftt_cell_pos (cell, &pos);
+  gfs_cell_cm (cell, &cm);
   size = ftt_cell_size (cell);
-#ifdef FTT_2D
-  GFS_STATE (cell)->div = - PI*PI*(K*K + L*L)*sin (PI*K*pos.x)*sin(PI*L*pos.y);
+#if FTT_2D
+  GFS_STATE (cell)->div = - M_PI*M_PI*(K*K + L*L)*sin (M_PI*K*cm.x)*sin(M_PI*L*cm.y);
 #else  /* FTT_3D */
-  GFS_STATE (cell)->div = - PI*PI*(K*K + L*L + M*M)*
-    sin (PI*K*pos.x)*sin(PI*L*pos.y)*sin(PI*M*pos.z);
+  GFS_STATE (cell)->div = - M_PI*M_PI*(K*K + L*L + M*M)*
+    sin (M_PI*K*cm.x)*sin(M_PI*L*cm.y)*sin(M_PI*M*cm[2]);
 #endif /* FTT_3D */
   if (GFS_IS_MIXED (cell))
     GFS_STATE (cell)->div *= size*size*GFS_STATE (cell)->solid->a; 
@@ -62,6 +158,191 @@ static void init_sine_div (FttCell * cell)
     GFS_STATE (cell)->div *= size*size;
 }
 
+#define EPS 1e-3
+
+static double dx (double x, double y, double (*f) (double, double))
+{
+  double f0 = (*f) (x, y);
+  
+  return (2.*((*f) (x + EPS, y) - f0)/3. -
+             ((*f) (x + 2.*EPS, y) - f0)/12. -
+          2.*((*f) (x - EPS, y) - f0)/3. +
+             ((*f) (x - 2.*EPS, y) - f0)/12.)/EPS;
+}
+
+static double dx2 (double x, double y, double (*f) (double, double))
+{
+  double f0 = (*f) (x, y);
+  
+  return (4.*((*f) (x + EPS, y) - f0)/3. -
+             ((*f) (x + 2.*EPS, y) - f0)/12. +
+          4.*((*f) (x - EPS, y) - f0)/3. -
+             ((*f) (x - 2.*EPS, y) - f0)/12.)/(EPS*EPS);
+}
+
+static double dy (double x, double y, double (*f) (double, double))
+{
+  double f0 = (*f) (x, y);
+  
+  return (2.*((*f) (x, y + EPS) - f0)/3. -
+             ((*f) (x, y + 2.*EPS) - f0)/12. -
+          2.*((*f) (x, y - EPS) - f0)/3. +
+             ((*f) (x, y - 2.*EPS) - f0)/12.)/EPS;
+}
+
+static double dy2 (double x, double y, double (*f) (double, double))
+{
+  double f0 = (*f) (x, y);
+  
+  return (4.*((*f) (x, y + EPS) - f0)/3. -
+             ((*f) (x, y + 2.*EPS) - f0)/12. +
+          4.*((*f) (x, y - EPS) - f0)/3. -
+             ((*f) (x, y - 2.*EPS) - f0)/12.)/(EPS*EPS);
+}
+
+static double annulus_p (double x, double y)
+{
+  return x*cos (M_PI*(0.25 - sqrt (x*x + y*y))/(0.25 - 0.5))/sqrt (x*x + y*y);
+  //  return cos (M_PI*(0.25 - sqrt (x*x + y*y))/(0.25 - 0.5));
+}
+
+static void init_annulus_div (FttCell * cell)
+{
+  FttVector cm;
+  gdouble size;
+
+  gfs_cell_cm (cell, &cm);
+  size = ftt_cell_size (cell);
+  GFS_STATE (cell)->div = (dx2 (cm.x, cm.y, annulus_p) +
+			   dy2 (cm.x, cm.y, annulus_p));
+  if (GFS_IS_MIXED (cell))
+    GFS_STATE (cell)->div *= size*size*GFS_STATE (cell)->solid->a; 
+  else
+    GFS_STATE (cell)->div *= size*size;
+}
+
+extern void gfs_face_gradient_flux_centered (const FttCellFace * face,
+					     GfsGradient * g,
+					     guint v,
+					     gint max_level);
+
+static void cell_annulus_error (FttCell * cell, gdouble * bias)
+{
+#if 1
+  FttVector pos;
+
+  ftt_cell_pos (cell, &pos);
+  GFS_STATE (cell)->v = GFS_STATE (cell)->dp = 
+    annulus_p (pos.x, pos.y) - GFS_STATE (cell)->p -
+    *bias;
+#else
+  FttCellFace f;
+  FttVector p;
+  GtsVector ca;
+
+  f.cell = cell;
+  f.d = FTT_RIGHT;
+  f.neighbor = ftt_cell_neighbor (f.cell, f.d);
+  ftt_face_pos (&f, &p);
+  if (f.neighbor) {
+    gfs_face_ca (&f, ca);
+    GFS_STATE (cell)->dp = dx (ca.x, ca.y, annulus_p) - 
+      GFS_STATE (cell)->f[f.d].un;
+  }
+  else
+    GFS_STATE (cell)->dp = 0.;
+#endif
+}
+
+static void init_day_pressure (FttCell * cell)
+{
+  FttVector p;
+  gdouble r2, theta;
+
+  ftt_cell_pos (cell, &p);
+  r2 = p.x*p.x + p.y*p.y;
+  theta = atan2 (p.y, p.x);
+  GFS_STATE (cell)->p = r2*r2*cos (3.*theta);
+}
+
+static void init_day_div (FttCell * cell)
+{
+  FttVector cm;
+  gdouble r2, theta;
+  gdouble h = ftt_cell_size (cell);
+  
+  gfs_cell_cm (cell, &cm);
+  r2 = cm.x*cm.x + cm.y*cm.y;
+  theta = atan2 (cm.y, cm.x);
+  GFS_STATE (cell)->div = 7.*r2*cos (3.*theta);
+  if (GFS_IS_MIXED (cell)) {
+    GfsSolidVector * s = GFS_STATE (cell)->solid;
+    GtsVector n;
+    gdouble a;
+    gdouble x = s->ca.x;
+    gdouble y = s->ca.y;
+
+    n[0] = s->s[1] - s->s[0];
+    n[1] = s->s[3] - s->s[2];
+    r2 = x*x + y*y;
+#if 0
+    a = sqrt (n[0]*n[0] + n[1]*n[1]);
+    //    fprintf (stderr, "%g %g\n", a, s->sa);
+    n[0] /= a;
+    n[1] /= a;
+    GFS_STATE (cell)->div = GFS_STATE (cell)->div*h*h*s->a -
+      ((4.*x*x*x*x - 3.*x*x*y*y - 3.*y*y*y*y)*n[0] + 
+       (-x*y*(5.*x*x + 9.*y*y))*n[1])/sqrt (r2)*h*s->sa;
+#else
+    GFS_STATE (cell)->div = GFS_STATE (cell)->div*h*h*s->a -
+      ((4.*x*x*x*x - 3.*x*x*y*y - 3.*y*y*y*y)*n[0] + 
+       (-x*y*(5.*x*x + 9.*y*y))*n[1])/sqrt (r2)*h;
+#endif
+  }
+  else {
+    FttCellFace f;
+    FttCellNeighbors n;
+
+    GFS_STATE (cell)->div *= h*h;
+    ftt_cell_neighbors (cell, &n);
+    f.cell = cell;
+    for (f.d = 0; f.d < FTT_NEIGHBORS; f.d++)
+      if (n.c[f.d] == NULL) {
+	FttVector p;
+	gdouble x, y, r2;
+	gdouble nx[4] = { 1., -1., 0., 0.};
+	gdouble ny[4] = { 0., 0., 1., -1.};
+
+	ftt_face_pos (&f, &p);
+	x = p.x; y = p.y;
+	r2 = x*x + y*y;
+	GFS_STATE (cell)->div -=
+	  ((4.*x*x*x*x - 3.*x*x*y*y - 3.*y*y*y*y)*nx[f.d] + 
+	   (-x*y*(5.*x*x + 9.*y*y))*ny[f.d])/sqrt (r2)*h;
+      }
+  }
+}
+
+static void cell_day_error (FttCell * cell, gdouble * bias)
+{
+  FttVector p;
+  gdouble r2, theta;
+
+  ftt_cell_pos (cell, &p);
+  r2 = p.x*p.x + p.y*p.y;
+  theta = atan2 (p.y, p.x);
+  GFS_STATE (cell)->v = GFS_STATE (cell)->dp = 
+    r2*r2*cos (3.*theta) - GFS_STATE (cell)->p - *bias;
+}
+
+static void cell_day_truncation (FttCell * cell, GfsNorm * norm)
+{
+  gdouble h = ftt_cell_size (cell);
+
+  GFS_STATE (cell)->v = GFS_STATE (cell)->dp = GFS_STATE (cell)->res/(h*h);
+  gfs_norm_add (norm, GFS_STATE (cell)->dp, h*h);
+}
+
 static void sum_volume (FttCell * cell, GtsRange * vol)
 {
   gdouble size = ftt_cell_size (cell);
@@ -82,7 +363,7 @@ static void add_ddiv (FttCell * cell, gdouble * ddiv)
     GFS_STATE (cell)->div += size*size*(*ddiv);
 }
 
-static void correct_div (FttCell * root)
+static void correct_div (FttCell * root, gboolean * verbose)
 {
   GtsRange div, vol;
   gdouble ddiv;
@@ -94,6 +375,8 @@ static void correct_div (FttCell * root)
   gts_range_update (&vol);
   ddiv = - div.mean/vol.mean;
 
+  if (*verbose)
+    fprintf (stderr, "Correcting divergence: %g\n", ddiv);
   ftt_cell_traverse (root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
 		     (FttCellTraverseFunc) add_ddiv, &ddiv);
 }
@@ -103,12 +386,12 @@ static void cell_error (FttCell * cell, gdouble * avg)
   FttVector pos;
 
   ftt_cell_pos (cell, &pos);
-#ifdef FTT_2D
+#if FTT_2D
   GFS_STATE (cell)->dp = (*avg) + 
-    sin (PI*K*pos.x)*sin (PI*L*pos.y) - GFS_STATE (cell)->p;
+    sin (M_PI*K*pos.x)*sin (M_PI*L*pos.y) - GFS_STATE (cell)->p;
 #else  /* FTT_3D */
   GFS_STATE (cell)->dp = (*avg) + 
-    sin (PI*K*pos.x)*sin (PI*L*pos.y)*sin (PI*M*pos.z) - GFS_STATE (cell)->p;
+    sin (M_PI*K*pos.x)*sin (M_PI*L*pos.y)*sin (M_PI*M*pos.z) - GFS_STATE (cell)->p;
 #endif /* FTT_3D */
 }
 
@@ -126,7 +409,7 @@ static void cell_error_ref (FttCell * cell, gpointer * data)
   locate = ftt_cell_locate (root, pos, level);
   g_assert (locate && ftt_cell_level (locate) == level);
   GFS_STATE (cell)->dp = GFS_STATE (locate)->p - *bias - GFS_STATE (cell)->p;
-  if (a >= 1e-2 && fabs (GFS_STATE (cell)->dp) > *dpmax) {
+  if (fabs (GFS_STATE (cell)->dp) > *dpmax) {
     *dpmax = fabs (GFS_STATE (cell)->dp);
     *amax = a;
   }
@@ -147,11 +430,6 @@ static void cell_error_hist (FttCell * cell, gpointer * data)
 	  GFS_STATE (cell)->solid->a : 1., GFS_STATE (cell)->dp);
 }
 
-static void correct_pressure (FttCell * cell)
-{
-  GFS_STATE (cell)->p += GFS_STATE (cell)->dp;
-}
-
 static void weighted_residual (FttCell * cell)
 {
   gdouble size = ftt_cell_size (cell);
@@ -184,7 +462,7 @@ static gboolean refine_inside_solid (FttCell * cell, gpointer * data)
     GtsPoint * p = gts_point_new (gts_point_class (), 0., 0., 0.);
     FttVector pos;
     static gint shift[FTT_CELLS][3] = 
-#ifdef FTT_2D
+#if FTT_2D
     {{-1,-1,0},{-1,1,0},{1,1,0},{1,-1,0}};
 #else  /* FTT_3D */
     {{-1,-1,-1},{-1,1,-1},{1,1,-1},{1,-1,-1},
@@ -225,8 +503,9 @@ static void init_solid_fractions (FttCell * cell, gpointer * data)
 
   gfs_cell_init (cell, domain);
   gfs_cell_init_solid_fractions (cell, 
-				data[0], data[1], *is_open, 
-				TRUE, gfs_cell_cleanup);
+				 data[0], data[1], *is_open, 
+				 TRUE, (FttCellCleanupFunc) gfs_cell_cleanup,
+				 NULL);
 }
 
 #define GFS_FLAG_MARKED  (1 << 5)
@@ -240,17 +519,21 @@ typedef enum
 } Method;
 
 static GfsBox * solution (guint levels,
-			    guint levelmax,
-			    gboolean relative,
-			    GtsSurface * solid,
-			    gboolean nosolid,
-			    gdouble probability,
-			    gboolean verbose,
-			    Method method,
-			    guint vcycle,
-			    guint nrelax,
-			    gdouble * ttime,
-			    GfsDomain ** retdomain)
+			  guint levelmax,
+			  gboolean relative,
+			  GtsSurface * solid,
+			  gboolean nosolid,
+			  gdouble probability,
+			  gboolean verbose,
+			  gboolean annulus,
+                          gboolean day,
+			  gboolean truncation,
+			  gboolean fix_vol,
+			  Method method,
+			  guint vcycle,
+			  guint nrelax,
+			  gdouble * ttime,
+			  GfsDomain ** retdomain)
 {
   GfsBox * box;
   GfsDomain * domain;
@@ -295,7 +578,9 @@ static GfsBox * solution (guint levels,
       gpointer data[4];
 	
       gfs_cell_init_solid_fractions (box->root, solid, tree, is_open, 
-				     TRUE, gfs_cell_cleanup);
+				     TRUE, 
+				     (FttCellCleanupFunc) gfs_cell_cleanup,
+				     NULL);
 
       data[0] = solid;
       data[1] = tree;
@@ -308,9 +593,13 @@ static GfsBox * solution (guint levels,
       ftt_cell_refine (box->root, 
 		       (FttCellRefineFunc) ftt_refine_corner, NULL,
 		       (FttCellInitFunc) init_solid_fractions, data);
-      
+
       g_assert (gfs_cell_check_solid_fractions (box->root, 
 						solid, tree, is_open));
+      if (fix_vol)
+	gfs_domain_cell_traverse (domain, 
+				  FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+				  (FttCellTraverseFunc) fix_volume, NULL);
     }
     gts_bb_tree_destroy (tree, TRUE);
   }
@@ -329,10 +618,24 @@ static GfsBox * solution (guint levels,
     }
     gfs_cell_fluid (box->root);
   }
-
-  ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-		     (FttCellTraverseFunc) init_sine_div, NULL);
-  correct_div (box->root); /* enforce solvability condition */
+  
+  if (annulus)
+    ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+		       (FttCellTraverseFunc) init_annulus_div, NULL);
+  else if (day) {
+    ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+		       (FttCellTraverseFunc) init_day_div, NULL);
+    if (truncation) {
+      ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			 (FttCellTraverseFunc) init_day_pressure, NULL);
+      gfs_poisson_coefficients (domain, NULL, 1.);
+      return box;
+    }
+  }
+  else
+    ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+		       (FttCellTraverseFunc) init_sine_div, NULL);
+  correct_div (box->root, &verbose); /* enforce solvability condition */
   if (verbose) {
     GtsRange stats = gfs_stats_variable (box->root, gfs_div, 
 					 FTT_TRAVERSE_LEAFS, -1);
@@ -342,7 +645,8 @@ static GfsBox * solution (guint levels,
   }
   
   levels = gfs_domain_depth (domain);
-  gfs_residual (domain, FTT_TRAVERSE_LEAFS, -1, gfs_p, gfs_div, gfs_res);
+  gfs_poisson_coefficients (domain, NULL, 1.);
+  gfs_residual (domain, FTT_DIMENSION, FTT_TRAVERSE_LEAFS, -1, gfs_p, gfs_div, gfs_res);
 
   timer = g_timer_new ();
   g_timer_start (timer);
@@ -360,8 +664,8 @@ static GfsBox * solution (guint levels,
 	
 	g_timer_stop (timer);
 	total_time += g_timer_elapsed (timer, NULL);
-	gfs_residual (domain, FTT_TRAVERSE_LEAFS, -1, gfs_p, gfs_div, gfs_res);
-	norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1);
+	gfs_residual (domain, FTT_DIMENSION, FTT_TRAVERSE_LEAFS, -1, gfs_p, gfs_div, gfs_res);
+	norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1, 1.);
 	fprintf (stderr, 
 		 "div n: %d time: %g first: %g second: %g infty: %g w: %g\n",
 		 cycle, total_time,
@@ -371,11 +675,10 @@ static GfsBox * solution (guint levels,
       }
       switch (method) {
       case POISSON:
-	gfs_poisson_coefficients (domain, NULL, 1.);
-	gfs_poisson (domain, 0, levels, nrelax, gfs_p, gfs_div);
+	gfs_poisson_cycle (domain, FTT_DIMENSION, 0, levels, nrelax, gfs_p, gfs_div);
 	break;
       case RELAX: {
-	gfs_relax (domain, -1, gfs_p, gfs_div);
+	gfs_relax (domain, FTT_DIMENSION, -1, gfs_p, gfs_div);
 	break;
       }
       default :
@@ -407,6 +710,10 @@ int main (int argc, char * argv[])
   gdouble probability = -1.;
   gboolean relative = FALSE;
   gboolean histogram = FALSE;
+  gboolean annulus = FALSE;
+  gboolean day = FALSE;
+  gboolean truncation = FALSE;
+  gboolean fix_vol = FALSE;
   Method method = POISSON;
   GfsVariable * output = NULL;
   gint olevel = -1;
@@ -428,6 +735,10 @@ int main (int argc, char * argv[])
   while (c != EOF) {
 #ifdef HAVE_GETOPT_LONG
     static struct option long_options[] = {
+      {"fixvolume", no_argument, NULL, 'f'},
+      {"truncation", no_argument, NULL, 'T'},
+      {"day", no_argument, NULL, 'd'},
+      {"annulus", no_argument, NULL, 'A'},
       {"histogram", no_argument, NULL, 'i'},
       {"solid", required_argument, NULL, 's'},
       {"refine", required_argument, NULL, 'r'},
@@ -449,11 +760,23 @@ int main (int argc, char * argv[])
       {"version", no_argument, NULL, 'V'}
     };
     int option_index = 0;
-    switch ((c = getopt_long (argc, argv, "hvo:m:Ml:s:Vr:nRa:te:x:k:u:w:i",
+    switch ((c = getopt_long (argc, argv, "hvo:m:Ml:s:Vr:nRa:te:x:k:u:w:iAdTf",
 			      long_options, &option_index))) {
 #else /* not HAVE_GETOPT_LONG */
-    switch ((c = getopt (argc, argv, "hvo:m:Ml:s:Vr:nRa:te:x:k:u:w:i"))) {
+    switch ((c = getopt (argc, argv, "hvo:m:Ml:s:Vr:nRa:te:x:k:u:w:iAdTf"))) {
 #endif /* not HAVE_GETOPT_LONG */
+    case 'f': /* fixvolume */
+      fix_vol = TRUE;
+      break;
+    case 'T': /* truncation */
+      truncation = TRUE;
+      break;
+    case 'd': /* day */
+      day = TRUE;
+      break;
+    case 'A': /* annulus */
+      annulus = TRUE;
+      break;
     case 'i': /* histogram */
       histogram = TRUE;
       break;
@@ -545,7 +868,7 @@ int main (int argc, char * argv[])
     case 'u': /* ky */
       L = atof (optarg);
       break;
-#ifndef FTT_2D
+#if (!FTT_2D)
     case 'w': /* kz */
       M = atof (optarg);
       break;
@@ -558,6 +881,10 @@ int main (int argc, char * argv[])
      "Usage: poisson [OPTION] LEVELS CYCLE NRELAX\n"
      "Solve a simple poisson equation on various grids\n"
      "\n"
+     "  -f    --fixvolume   fix volume of cut cells using a piecewise linear approximation\n"
+     "  -T    --truncation  compute truncation error (only for Day problem)\n"
+     "  -d    --day         solves Day problem\n"
+     "  -A    --annulus     solves annulus problem\n"
      "  -s F  --solid=F     read GTS file F describing solid boundaries\n"
      "                      (if F is `-' the standard input is used)\n"
      "  -r L  --refine=L    refines up to level L\n"
@@ -572,12 +899,12 @@ int main (int argc, char * argv[])
      "  -t    --tree        output a GFS representation of the whole cell tree\n"
      "  -e L  --error=L     use an evaluation of the solution with a maximum level L\n"
      "                      as the reference solution for the error computation\n"
-     "  -i    --histogram   outputs list of cell fractions and associated errors"
+     "  -i    --histogram   outputs list of cell fractions and associated errors\n"
      "  -x L  --order=L     run several simulations of varying size (starting at L)\n"
      "                      and outputs the corresponding errors\n"
      "  -k N  --kx=N        set wave number in direction x to N (default is 3)\n"
      "  -u N  --ky=N        set wave number in direction y to N (default is 3)\n"
-#ifndef FTT_2D
+#if (!FTT_2D)
      "  -w N  --kz=N        set wave number in direction z to N (default is 3)\n"
 #endif /* not FTT_2D */
      "  -m M  --method=M    use method M, where M is one of:\n"
@@ -633,7 +960,8 @@ int main (int argc, char * argv[])
   if (levels_ref > levels) {
     reference = solution (levels_ref, levelmax, relative,
 			  solid, nosolid, probability, 
-			  FALSE, method, vcycle, nrelax, 
+			  FALSE, annulus, day, truncation, fix_vol,
+			  method, vcycle, nrelax, 
 			  &total_time, &domain);
     ftt_cell_traverse (reference->root, FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
 		       (FttCellTraverseFunc) gfs_get_from_below_intensive,
@@ -647,13 +975,15 @@ int main (int argc, char * argv[])
 
   for (l = minlevel; l <= levels; l++) {
     if (box)
-      ftt_cell_destroy (box->root, gfs_cell_cleanup);
+      ftt_cell_destroy (box->root, (FttCellCleanupFunc) gfs_cell_cleanup, 
+			NULL);
     box = solution (l, levelmax, relative, solid, nosolid, probability, 
-		       verbose, method, vcycle, nrelax, &total_time, &domain);
+		    verbose, annulus, day, truncation, fix_vol,
+		    method, vcycle, nrelax, &total_time, &domain);
 
-    gfs_residual (domain, FTT_TRAVERSE_LEAFS, -1, gfs_p, gfs_div, gfs_res);
+    gfs_residual (domain, FTT_DIMENSION, FTT_TRAVERSE_LEAFS, -1, gfs_p, gfs_div, gfs_res);
     if (verbose) {
-      norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1);
+      norm = gfs_domain_norm_residual (domain, FTT_TRAVERSE_LEAFS, -1, 1.);
       fprintf (stderr, 
 	       "div n: %d time: %g first: %g second: %g infty: %g w: %g\n",
 	       vcycle, total_time,
@@ -687,13 +1017,45 @@ int main (int argc, char * argv[])
 			   (FttCellTraverseFunc) cell_error_ref, data);
 	fprintf (stderr, "level: %d dpmax: %g amax: %g\n", 
 		 l, dpmax, amax);
+	norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEAFS, -1);
+      }
+      else if (annulus) {
+	norm.bias = 0.;
+	ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			   (FttCellTraverseFunc) cell_annulus_error, 
+			   &norm.bias);
+	norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEAFS, -1);
+	ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			   (FttCellTraverseFunc) cell_annulus_error, 
+			   &norm.bias);
+	norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEAFS, -1);
+      }
+      else if (day) {
+	if (truncation) {
+	  gfs_norm_init (&norm);
+	  ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			     (FttCellTraverseFunc) cell_day_truncation, 
+			     &norm);
+	  gfs_norm_update (&norm);
+	}
+	else {
+	  norm.bias = 0.;
+	  ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			     (FttCellTraverseFunc) cell_day_error, 
+			     &norm.bias);
+	  norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEAFS, -1);
+	  ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			     (FttCellTraverseFunc) cell_day_error, 
+			     &norm.bias);
+	  norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEAFS, -1);
+	}
       }
       else {
 	stats = gfs_stats_variable (box->root, gfs_p, FTT_TRAVERSE_LEAFS, -1);
 	ftt_cell_traverse (box->root, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
 			   (FttCellTraverseFunc) cell_error, &stats.mean);
+	norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEAFS, -1);
       }
-      norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEAFS, -1);
       fprintf (stderr, "total err level: %d first: %g second: %g infty: %g bias: %g w: %g\n",
 	       l, norm.first, norm.second, norm.infty, norm.bias, norm.w);
       norm = gfs_norm_variable (box->root, gfs_dp, FTT_TRAVERSE_LEVEL, 
@@ -761,16 +1123,12 @@ int main (int argc, char * argv[])
     }
   }
   else if (output_tree) {
-    if (olevel >= 0) {
-      ftt_cell_traverse (box->root, FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-			 (FttCellTraverseFunc) gfs_get_from_below_intensive,
-			 gfs_p);
-      ftt_cell_write (box->root, olevel, stdout, 
-		      (FttCellWriteFunc) gfs_cell_write, gfs_p);
-    }
-    else
-      ftt_cell_write (box->root, -1, stdout, 
-		      (FttCellWriteFunc) gfs_cell_write, gfs_p);
+    ftt_cell_traverse (box->root, FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
+		       (FttCellTraverseFunc) gfs_get_from_below_intensive,
+		       gfs_p);
+    if (olevel >= 0)
+      domain->max_depth_write = olevel;
+    gts_graph_write (GTS_GRAPH (domain), stdout);
   }
 
   return 0;
